#
#   This file is part of SST/macroscale:
#                The macroscale architecture simulator from the SST suite.
#   Copyright (c) 2009 Sandia Corporation.
#   This software is distributed under the BSD License.
#   Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#   the U.S. Government retains certain rights in this software.
#   For more information, see the LICENSE file in the top
#   SST/macroscale directory.
#

import re
import sys
import getopt
import sstmac

rsqnodes = [4, 24]

# hostname=nid00010
# numprocs=1
# username=ctvaugh
# startime=1264087508
# fileprefix=dumpi--2010.01.21.08.25.08
# version=0
# subversion=6
# subsubversion=2

def read_metadata(metafile):
    """Grab info out of a dumpi metafile (subject to change)."""
    return sstmac.dumpimeta.construct(metafile)

def build_rsqtorus(meta):
    """Construct a 2D torus that can fit the given RSQ run.
    Should later be fixed to correspond to the actual layout of RSQ
    (dual and quad node positions)"""
    totnodes = rsqnodes[0]*rsqnodes[1]
    torus = sstmac.torus.construct_2D(rsqnodes[0], rsqnodes[1], 1, 1)
    idxing = sstmac.rsqindexing.construct()
    alloc = sstmac.rsqallocation.construct(meta, torus)
    nodes = meta.getnumprocs()
    return [torus, idxing, alloc, nodes, totnodes]

def build_torus(meta):
    """Construct a 2D torus network network"""
    totnodes = rsqnodes[0]*rsqnodes[1]
    torus = sstmac.torus.construct_2D(rsqnodes[0],rsqnodes[1],1,1)
    idxing = sstmac.blockindexing.construct()
    alloc = sstmac.firstavailableallocation.construct()
    nodes = meta.getnumprocs()
    return [torus, idxing, alloc, nodes, totnodes]

def build_xbar(meta):
    """Construct a crossbar network (for optimistic estimates"""
    totnodes = rsqnodes[0]*rsqnodes[1]*4
    crossbar = sstmac.crossbar.construct(totnodes)
    idxing = sstmac.blockindexing.construct()
    alloc = sstmac.firstavailableallocation.construct()
    nodes = meta.getnumprocs()
    return [crossbar, idxing, alloc, nodes, totnodes]

def build_gamma(meta):
    """Construct a gamma network"""
    totnodes = rsqnodes[0]*rsqnodes[1]*4
    gamma = sstmac.gamma.construct(8,3,1)
    idxing = sstmac.blockindexing.construct()
    alloc = sstmac.firstavailableallocation.construct()
    nodes = meta.getnumprocs()
    return [gamma, idxing, alloc, nodes, totnodes]

def build_fattree(meta):
    """Construct a fattree network"""
    totnodes = rsqnodes[0]*rsqnodes[1]*4
    fattree = sstmac.fattree.construct(3,12)
    idxing = sstmac.blockindexing.construct()
    alloc = sstmac.firstavailableallocation.construct()
    nodes = meta.getnumprocs()
    return [fattree, idxing, alloc, nodes, totnodes]

def build_torus3(meta):
    """Construct a 3D torus network network"""
    totnodes = rsqnodes[0]*rsqnodes[1]*4
    torus3 = sstmac.torus.construct_3D(8,8,8,1)
    idxing = sstmac.blockindexing.construct()
    alloc = sstmac.firstavailableallocation.construct()
    nodes = meta.getnumprocs()
    return [torus3, idxing, alloc, nodes, totnodes]

def create(metafile, timescale, bandwidth, latency, cpubw, cpulat,
           router_node_bw, router_router_bw, netsharing, n_rebalance_max,
           nettype, circtype, verbose):
    """Run a tracefile from RSQ through the simulator"""
#    print nodes, " ", metafile, " ", timescale, " ", bandwidth, " ", latency
    if verbose:
        sstmac.debug().to_stderr()
    # The system is almost entirely defined by the metadata
    meta = read_metadata(metafile)
    print "Read a metafile with",meta.getnumprocs(),"processors"
    if nettype == "rsqtorus":
        [net, idxing, alloc, nodes, totnodes] = build_rsqtorus(meta)
    elif nettype == "torus":
        [net, idxing, alloc, nodes, totnodes] = build_torus(meta)
    elif nettype == "crossbar":
        [net, idxing, alloc, nodes, totnodes] = build_xbar(meta)
    elif nettype == "gamma":
        [net, idxing, alloc, nodes, totnodes] = build_gamma(meta)
    elif nettype == "fattree":
        [net, idxing, alloc, nodes, totnodes] = build_fattree(meta)
    elif nettype == "torus3":
        [net, idxing, alloc, nodes, totnodes] = build_torus3(meta)
    else:
        raise Exception("Invalid network type \"" + nettype + "\"")
    # Go.
    print "Built a network with",totnodes,"nodes"
    mgr = sstmac.manager.construct()
    if circtype == "exclusive":
        mgr.set_interconnect(sstmac.circuit.construct(sstmac.timestamp(latency), bandwidth, net))
    elif circtype == "shared":
        flow = sstmac.flower.construct(net, router_node_bw, router_router_bw, netsharing)
        flow.n_rebalance_max = n_rebalance_max
        mgr.set_interconnect(sstmac.sharedcircuit.construct(sstmac.timestamp(latency), bandwidth, net, flow))
        #mgr.set_interconnect(sstmac.sharedcircuit.construct(sstmac.timestamp(latency), bandwidth, net))
    else:
        raise Exception("Invalid circuit type \"" + circtype + "\"")
    mgr.set_eventmanager(sstmac.eventmap.construct())
    mgr.add_nodes(sstmac.simplenode.construct(sstmac.timestamp(cpulat), cpubw),
                  totnodes)
    parser = sstmac.parsedumpi.construct(meta.fileprefix_,timescale)
    job = sstmac.mpilaunch.construct(parser,timescale)
    job.set_scheduler(sstmac.scheduler.construct(81920, 10240000, False))
    print "Adding launcher for",meta.getnumprocs(),"peers."
    mgr.add_launcher(sstmac.timestamp(0), job, nodes,
                     1, idxing, alloc)
    runtime = mgr.run().sec()
    print "# %5s %10s %10s %10s %10s %10s" % ("procs", "timescale", "bandwidth", "latency", "nettype", "runtime")
    print "%7d %10.6e %8.6e %8.6e %10s %8.6f" % (meta.getnumprocs(),timescale,bandwidth,latency,nettype,runtime)


def main(args):
    """Parse arguments and run dumpi experiment."""
    (metafile, timescale, bandwidth, latency, cpubw, cpulat, nettype,verbose) =\
               (None, 1, 2.5e9, 1.3e-6, 1e10, 0, "torus", False)
    (router_router_bw, router_node_bw) = (1e9, 1e9)
    circtype = "exclusive"
    netsharing = sstmac.ORDERED
    n_rebalance_max = -1
    opts = getopt.gnu_getopt(args, "hvi:t:b:l:B:L:N:C",
                             ['help', 'verbose', 'metafile=', 'timescale=',
                              'bandwidth=', 'latency=', 'cpubandwidth=',
                              'cpulatency=', 'router_router_bw=',
                              'netsharing=', 'n_rebalance_max=',
                              'router_node_bw=',
                              'network=', 'circuit='])
    for pair in opts[0]:
        if pair[0] == "-h" or pair[0] == "--help":
            print sys.argv[0], "\\"
            print "    [(--help|-h)]                 \\"
            print "    [(--verbose|-v)]              \\"
            print "    [(--timescale|-t)   <float> ] \\"
            print "    [(--bandwidth|-b)   <int>   ] \\"
            print "    [(--latency|-l)     <float> ] \\"
            print "    [(--cpubandwidth|B) <int>   ] \\"
            print "    [(--cpulatency|L)   <float> ] \\"
            print "    [(--router_router_bw) <float> ] \\"
            print "    [(--router_node_bw) <float>   ] \\"
            print "    [(--netsharing) <string>   ] \\"
            print "    [(--n_rebalance_max) <int>   ] \\"
            print "    [(--network|-N)     <string>] \\"
            print "    [(--circuit|-C)     <string>] \\"
            print "    (--metafile|-i)     <string>"
            print "Note that the --metafile (or -i) argument is not optional."
            print "Valid arguments for --network are \"torus\", \"fatree\", \"rsqtorus\"*, \"gamma\", \"torus3\" and \"crossbar\""
            print "Valid arguments for --circuit are \"exclusive\"* and \"shared\""
            print "Valid arguments for --netsharing are \"ordered\"*, \"overcommit\", and \"throttle_overcommits\""
            return
        if pair[0] == "-v" or pair[0] == "--verbose":
            verbose = True
        if pair[0] == "-n" or pair[0] == "--nodes":
            nodes = int(pair[1])
        if pair[0] == "-i" or pair[0] == "--metafile":
            metafile = pair[1]
        if pair[0] == "-t" or pair[0] == "--timescale":
            timescale = float(pair[1])
        if pair[0] == "-b" or pair[0] == "--bandwidth":
            bandwidth = float(pair[1])
        if pair[0] == "-l" or pair[0] == "--latency":
            latency = float(pair[1])
        if pair[0] == "-B" or pair[0] == "--cpubandwidth":
            cpubw = float(pair[1])
        if pair[0] == "-L" or pair[0] == "--cpulatency":
            cpulat = float(pair[1])
        if pair[0] == "--router_router_bw":
            router_router_bw = float(pair[1])
        if pair[0] == "--router_node_bw":
            router_node_bw = float(pair[1])
        if pair[0] == "--n_rebalance_max":
            n_rebalance_max = int(pair[1])
        if pair[0] == "--netsharing":
            if pair[1] == "ordered":
                netsharing = sstmac.ORDERED
            elif pair[1] == "overcommit":
                netsharing = sstmac.OVERCOMMIT
            elif pair[1] == "throttle_overcommits":
                netsharing = sstmac.THROTTLE_OVERCOMMITS
            else:
                raise Exception("bad --netsharing argument");


        if pair[0] == '-N' or pair[0] == "--network":
            nettype = pair[1]
        if pair[0] == '-C' or pair[0] == "--circuit":
            circtype = pair[1]
#    print "create(",nodes,",",",metafile,",",timescale,",",bandwidth,",",latency,",",nettype,",",verbose,")"
    if not metafile:
        raise Exception("--metafile argument is not optional")
    create(metafile, timescale, bandwidth, latency,
           cpubw, cpulat, router_node_bw, router_router_bw, netsharing, n_rebalance_max, nettype, circtype, verbose)

if __name__ == "__main__":
    if len(sys.argv) > 1:
        main(sys.argv[1:])
    else:
        main(["--help"])
