/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/macrobridge/macrosender.h>
#include <sstmac/macrobridge/macroevent.h>
#include <sstmac/simulator/timestamp.h>
#include <sstmac/util/errors.h>
#include <sstmac/process/context.h>
#include <sstmac/sstmacconfig.h>
#include <exception>
#include <iostream>
#include <Python.h>
#include <assert.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

/// For some odd reason the python interpreter is not always in Python.h
extern "C" int Py_Main(int, char**);

using namespace SST;

namespace sstmac {
  
  using boost::shared_ptr;


  // A static singleton of a manager.
  boost::shared_ptr<macrosender::mgr> macrosender::mgr_;

  // The static stack size used for python (needs to be reasonably large
  // since Python is not a trivial piece of code).
  const size_t macrosender::python_stacksize_ = 1<<24;  // 16 MB stack.

  //
  // Private nested type to forward events.
  //
  class macrosender::mgr : public eventmanager {
    /// Who owns us?
    macrosender *parent_;

    /// Hi.
    mgr(macrosender *parent) : parent_(parent) {}
    
  public:
    /// Hello.
    static shared_ptr<mgr> construct(macrosender *parent) {
      return shared_ptr<mgr>(new mgr(parent));
    }
    
    /// Good bye.
    virtual ~mgr() {}
    
    /// Clear all events and set time back to a zero of your choice.
    /// This call shall not be permitted while the event manager is running.
    virtual void clear(const timestamp& = timestamp(0)) {
      throw illformederror("macrosender::mgr::clear():  Not supported\n");
    }
    
    /// Run the eventmanager.
    /// The eventmanager shall return control when no more messages remain.
    virtual void run() {
      //std::cerr << "macrosender::mgr::run()\n";
      //std::cerr << "macrosender::run:  Switching to main context.\n";
      swap_context(parent_->python_context_, parent_->macro_context_);
      //std::cerr << "macrosender::run:  Back in script context.\n";
    }
    
    /// Get the current eventmanager time.
    /// This call must work regardless of whether the eventmanager is currently
    /// running or not.  After a run, the call should return the time at which
    /// the last event was completed.
    const timestamp& now() const {
      static timestamp ts;
      //std::cerr << "macrosender::mgr::now()\n";
      ts = timestamp::exact_ticks(parent_->getCurrentSimTime());
      return ts;
    }

    /// Set off the given eventhandler at the given time.
    virtual void schedule(const timestamp &start_time,
                          const boost::shared_ptr<eventhandler>& handler) 
    {
      //std::cerr << "macrosender::mgr::schedule(" << start_time << ", "
      //          << handler << ")\n";
      parent_->schedule(start_time, handler);
    }
  };

  //
  // Private method to handle component events.
  //
  void macrosender::handle_events(Event*) {
    throw illformederror("macrosender::handle_events:  Sender is not "
                         "set up to handle events.");
  }

  //
  // Schedule an event.
  //
  void macrosender::schedule(const timestamp &target,
                             const boost::shared_ptr<eventhandler> &handler)
  {
    port_->Send(target.ticks() - getCurrentSimTime(),
                defaultTimeBase,
                new macroevent(handler));
  }

  //
  // Private method to add wiring between components.
  //
  void macrosender::get_wired(const std::string &target) {
    // Templates gone wild.
    typedef Event::Handler<macrosender> handler_t;
    handler_t *handler = new handler_t(this, &macrosender::handle_events);
    port_ = configureLink(target, defaultTimeBase, handler);
    if(port_ == NULL) {
      throw illformederror("macrosender::get_wired:  LinkAdd failed.");
    }
  }

  macrosender::macrosender(ComponentId_t id, Params_t &params) :
    Component(id), python_stack_(new char[python_stacksize_])
  {
    registerTimeBase(sstmac::timestamp::tick_interval_string());

    //std::cerr << "macrosender[" << this << "] constructing.\n"
    //          << "Parameters[" << params.size() << "]\n";
    std::string target = "";
    for(Params_t::const_iterator it = params.begin(); it != params.end(); ++it){
      //std::cerr << "    \"" << it->first << "\"==>\"" << it->second << "\"\n";
    }
    // We have to have a script name.
    Params_t::const_iterator it = params.find("script");
    if(it != params.end())
      script_ = it->second;
    else
      throw valueerror("macrosender: Parameter value \"script\" is undefined.");
    // We need to know the label to wire up to.
    it = params.find("dest");
    if(it != params.end())
      target = it->second;
    else
      throw valueerror("macrosender: Parameter value \"dest\" is undefined.");
    if(target == "") {
      //std::cerr << "macrosender:  dest must not be a blank string.\n";
      throw valueerror("macrosender:  blank target name.");
    }
    this->get_wired(target);
    init_context(macro_context_);
    init_context(python_context_);

    //std::cerr << "macrosender[" << this << "] constructed\n";
  }

  macrosender::~macrosender() {
    //std::cerr << "macrosender[" << this << "] going out of scope\n";
    destroy_context(macro_context_);
    destroy_context(python_context_);
    delete[] python_stack_;
  }

  // This is the name that swig associates with its init function.
  extern "C" {
    void init_sstmac(void);
  }

  //
  // This is the method that runs python.
  //
  void macrosender::python_routine(void *arg) {
    macrosender *self = (macrosender*)arg;
    // Here's where we run our python script
    //std::cerr << "Running python script " << self->script_ << "\n";
    Py_Initialize();
    assert(PyRun_SimpleString("import sys\n") == 0);
    init_sstmac();
    assert(PyRun_SimpleString("sys.path = sys.path + [\"" 
                              SSTMAC_PYTHONDIR "/sst-macroscale\"]\n") == 0);
    assert(PyRun_SimpleString("import sstmac\n") == 0);
    FILE *fp = fopen(self->script_.c_str(), "r");
    if(fp == NULL) {
      //std::cerr << "Failed to open " << self->script_ << ": "
      //          << strerror(errno) << "\n";
      std::stringstream ostr;
      ostr << "macrosender::Setup:  Failed to read script \""
           << self->script_
           << "\"";
      throw ioerror(ostr.str());
    }
    std::vector<char*> argv;
    argv.push_back(strdup(self->script_.c_str()));
    PySys_SetArgv(argv.size(), &argv.front());
    assert(PyRun_SimpleFile(fp, self->script_.c_str()) == 0);
    for (int i=0; i<argv.size(); i++) free(argv[i]);
  }

  //
  // Git goin'
  //
  int macrosender::Setup() {
    std::cerr << "macrosender[" << this << "]::Setup()\n";
    if(mgr_.get() == NULL)
      mgr_ = mgr::construct(this);
    // run the python script in a separate context so we can 
    // return control back to SST/microscale as soon as the script hits
    // manager::run() (which calls eventmanager::run(), which in turn
    // goes to macrosender::mgr::run() -- loads of fun).
    start_context(python_context_, python_stack_, python_stacksize_,
                  python_routine, this, &macro_context_);
    swap_context(macro_context_, python_context_);

    // And now we're done.
    return 0;
  }

  //
  // Whoa!
  //
  int macrosender::Finish() {
    std::cerr << "macrosender[" << this << "]::Finish()\n";
    swap_context(macro_context_, python_context_);
    //std::cerr << "macrosender[" << this << "]::Finish():  Script context done\n";
    return 0;
  }

  //
  // Get a manager.
  shared_ptr<eventmanager> macrosender::manager() {
    assert(mgr_.get() != NULL);
    return mgr_;
  }

} // end of namespace sstmac
