/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// hdtorus.h: Interface for torus networks.
//
// Author: Ali Pinar <apinar@sandia.gov>

#ifndef SSTMAC_NETWORK_HDTORUS_H
#define SSTMAC_NETWORK_HDTORUS_H
#include <sstmac/network/network.h>

namespace sstmac {

  /** Implements a high Dimensional torus network.

      <b>Configuring the 2HDTorus network</b>

      The inputs for this network can specify the following attributes:

      <table border="1">
      <tr><td>Keyword<td>Type<td>Default<td>Description
      <tr><td><tt>n1</tt><td>int<td>4<td>The number of nodes in the first
      dimension.
      <tr><td><tt>n2</tt><td>int<td>4<td>The number of nodes in the second
      dimension.
      <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.
      <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.
      </table>

      <b>Configuring the 3HDTorus network</b>

      The inputs for this network
      can specify the following attributes:

      <table border="1">
      <tr><td>Keyword<td>Type<td>Default<td>Description
      <tr><td><tt>n1</tt><td>int<td>4<td>The number of nodes in the first
      dimension.
      <tr><td><tt>n2</tt><td>int<td>4<td>The number of nodes in the second
      dimension.
      <tr><td><tt>n3</tt><td>int<td>0<td>The number of nodes in the third
      dimension.
      <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.
      <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.
      </table>

      <b>Configuring the 4HDTorus network</b>

      The inputs for this network
      can specify the following attributes:

      <table border="1">      <tr><td>Keyword<td>Type<td>Default<td>Description
      <tr><td><tt>n1</tt><td>int<td>4<td>The number of nodes in the first
      dimension.
      <tr><td><tt>n2</tt><td>int<td>4<td>The number of nodes in the second
      dimension.
      <tr><td><tt>n3</tt><td>int<td>0<td>The number of nodes in the third
      dimension.
       <tr><td><tt>nx</tt><td>int<td>4<td>The number of nodes in the second
      dimension.
      <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.
      <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.
      </table>
  */

  class hdtorus: public network {
  public:
    int width;
    
  private:
    std::string name_;
    bool direct_;
    int nndpersw_;
    std::vector<int>  dimensions_; 
    std::vector <bool>  wrapped_;
    void compute_diffvect(const vertex &v1, const vertex &v2, std::vector<int> &diffvect ) const;
    void compute_vertex_numbers(std::vector <std::vector<int> > &vertices, int P);
    void compute_vertices( std::vector <std::vector<int> > &vertices, std::set<vertex> &vertices_set, int id=-1);
    void compute_switch_edges(std::vector <std::vector<int> > &vertices, std::set<edge> &edges);
    void compute_node_edges(std::vector <std::vector<int> > &vertices, std::set<edge> &edges, int nndpersw);
    bool is_switch(const vertex_index &vi) const {
      if (direct_) return true;
      const vertex &v = g().index_to_vertex(vi);
      if (v[1] == nndpersw_) return true;
      return false;
    }
  protected:
    hdtorus(std::vector<int> dimensionarray, int nndpersw, int widtharg);
            
  public:
    int diameter() const;
    double a() const;
    int b() const;
    int shortest_path_length(const vertex &v1, const vertex &v2, int id=0) const;
    
    void shortest_path(const vertex&v1, const vertex&v2, walk&w) const;
    
    void next_hop(vertex_index loc, vertex_index dst, std::vector<lwedge> &nxt) const;
    
    static boost::shared_ptr<hdtorus> construct(std::vector<int> dimensions, int nndpersw, int widtharg);
    
    static boost::shared_ptr<hdtorus> construct_2D(int n1, int n2, int nndpersw, int widtharg); 
    
    static boost::shared_ptr<hdtorus> construct_3D(int n1, int n2, int n3, int nndpersw, int widtharg); 
    
    static boost::shared_ptr<hdtorus> construct_4D(int n1, int n2, int n3, int n4, int nndpersw, int widtharg); 
    

    virtual ~hdtorus();
    std::string name() const;

    /// Returns the number of nodes link to each switch.
    int nnodeperswitch() const { return nndpersw_; }

    /// Returns the vector giving each dimension of the torus.
    const std::vector<int> &dimensions() const { return dimensions_; }

    /// Return the vector specifying whether or not each dimension is wrapped.
    const std::vector<bool> &wrapped() const { return wrapped_; }
 };

}

#endif
