/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// tbird.cc: Implementation of the tbird network.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#include <sstream>
#include <sstmac/network/tbird.h>

using std::set;
using std::vector;
using std::string;

namespace sstmac {

  static void
  compute_nodes(set<vertex> &vertices)
  {
    vertices.clear();

    vertex v(3);
    // the first index is the tier number (0 for nodes)
    v[0] = 0;
    for (int i=0; i<288; i++) {
      // the first index is the leaf switch number
      v[1] = i;
      for (int j=0; j<16; j++) {
        // the first index is the node number within the leaf switch
        v[2] = j;
        vertices.insert(v);
      }
    }
  }

  static void
  compute_switches(set<vertex> &vertices)
  {
    vertices.clear();

    // switch vertices are (288-port switch chassis, 
    // leaf switch num)

    // switch vertices are
    // (tier, [ tier address ])
    // for tier = 3 (spine of core switches)
    //  tier address = [ core switch num, spine chip number ]
    // for tier = 2 (leaf of core switches)
    //  tier address = [ core switch num, leaf chip number ]
    // for tier = 1 (leaf switches)
    //  tier address = [ leaf switch number, 0 ]

    // core switches
    vertex v(3);
    for (int i=0; i<8; i++) {
      v[1] = i;
      for (int j=0; j<12; j++) {
        v[0] = 3;
        v[2] = j;
        vertices.insert(v);
      }
      for (int j=0; j<24; j++) {
        v[0] = 2;
        v[2] = j;
        vertices.insert(v);
      }
    }

    // leaf switches
    v[0] = 1;
    v[2] = 0;
    for (int i=0; i<288; i++) {
      v[1] = i;
      vertices.insert(v);
    }
  }

  static void
  compute_edges(set<edge> &edges)
  {
    edges.clear();

    // Compute the edges between the nodes and tier 1.
    vertex v0(3), v1(3);
    v0[0] = 0;             // node
    v1[0] = 1; v1[2] = 0;  // leaf
    for (int i=0; i<288; i++) {
      v0[1] = i;
      v1[1] = i;
      for (int j=0; j<16; j++) {
        v0[2] = j;
        edges.insert(edge(v0,v1));
        edges.insert(edge(v1,v0));
      }
    }

    // compute the edges between tier 1 and tier 2
    v0[0] = 1; v0[2] = 0;
    v1[0] = 2;
    // loop through the leaf chips in the core switches
    for (int core_leaf=0; core_leaf<24; core_leaf++) {
      v1[2] = core_leaf;
      // loop through the external ports on the leaf switches
      for (int core_spine=0; core_spine<12; core_spine++) {
        int leaf_switch_num = core_leaf*12 + core_spine;
        v0[1] = leaf_switch_num;
        // loop through the core switches
        for (int core=0; core<8; core++) {
          v1[1] = core;
          edges.insert(edge(v0,v1));
          edges.insert(edge(v1,v0));
        }
      }
    }

    // compute the edges between tier 2 and tier 3
    v0[0] = 2;
    v1[0] = 3;
    for (int core=0; core<8; core++) {
      v0[1] = core;
      v1[1] = core;
      for (int core_spine=0; core_spine<12; core_spine++) {
        v1[2] = core_spine;
        for (int core_leaf=0; core_leaf<24; core_leaf++) {
          v0[2] = core_leaf;
          edges.insert(edge(v0,v1));
          edges.insert(edge(v1,v0));
        }
      }
    }
  }

  // static sstmac::xmlclass::register_class reg("sstmac::tbird",
  //                               sstmac::xmlclass::ctor_template<tbird>);

  // tbird::tbird(TiXmlHandle h)
  // {
  //   init_tbird();
  // }

  tbird::tbird()
  {
    init_tbird();
  }

  void
  tbird::init_tbird()
  {
    compute_nodes(nodes_);

    set<vertex> switches;
    compute_switches(switches);

    set<edge> edges;
    compute_edges(edges);

    set<vertex> all_vertices;
    all_vertices.insert(nodes_.begin(), nodes_.end());
    all_vertices.insert(switches.begin(), switches.end());
    g_.init(all_vertices, edges);
    init();
  }

  double
  tbird::a() const
  {
    return 6;
  }

  int
  tbird::b() const
  {
    return n()/4;
  }

  boost::shared_ptr<tbird> tbird::construct() {
    return boost::shared_ptr<tbird>(new tbird);
  }

  tbird::~tbird() {
  }

  void
  tbird::shortest_path(const vertex &v1, const vertex &v2, walk &w) const
  {
    throw std::runtime_error("tbird::shortest_path: not unique");
  }

  void
  tbird::route(const vertex &v1, const vertex &v2, walk &w) const
  {
    w.clear();

    if (v1[0] != 0 || v2[0] != 0) {
      throw std::runtime_error("tbird::route: can only route between nodes");
    }

    if (v1[1] < 0 || v1[2] < 0
        || v2[1] < 0 || v1[2] < 0
        || v1[1] > 288 || v1[2] > 16
        || v2[1] > 288 || v2[2] > 16) {
      throw std::runtime_error("tbird::route: node invalid");
    }

    if (v1 == v2) {
      return;
    }

    // the tier1 up switch
    vertex sw1u(3);
    sw1u[0] = 1;
    sw1u[1] = v1[1];
    sw1u[2] = 0;

    // the tier1 down switch
    vertex sw1d(3);
    sw1d[0] = 1;
    sw1d[1] = v2[1];
    sw1d[2] = 0;

    // the up link
    w.append(g().generate_lwedge(v1,sw1u));

    // if the target is on the same tier 1 switch
    // send directly to the target
    if (sw1u == sw1d) {
      w.append(g().generate_lwedge(sw1d,v2));
      return;
    }

    // need a couple of pseudorandom numbers for routing

#if 0
    int u_0_7 = (v1[2] + v2[2] + 1)%8;
    // using the following tends to concentrate traffic
    // when running on just the low order nodes
    //int u_0_11 = (v1[1]/12+v2[1]/12)%12;
    // the following helps for small numbers of nodes, but
    // is much worse on large numbers
    //int u_0_11 = ((v1[2]+v2[2] + 1)*3)/8;
    // hybrid approach works well
    int u_0_11 = ((v1[1]/12+v2[1]/12)%12 + ((v1[2]+v2[2] + 1)*3)/8 + 1)%12;
#else
    int node1 = v1[1]*16 + v1[2];
    int node2 = v2[1]*16 + v2[2];
    // these did poorly
    //int u_0_7 = (node1 + node2 + 1)%8;
    //int u_0_11 = (node1 + node2 + 1)%12;
    // make the "random" choices independent
    // but this is cheating -- route can only depend on
    // current network device and destination
    //int u_0_7 = (node1 + 1 + node2 + 1)%8;
    //int u_0_11 = ((node1 + 1 + node2 + 1)/8)%12;
    // like the above, but only depends on node2
    // this actually works better than the above
    int u_0_7 = (node2 + 1)%8;
    int u_0_11 = ((node2 + 1)/8)%12;
#endif

    // pseudorandomly pick a core
    int core = u_0_7;

    // the tier2 up switch
    vertex sw2u(3);
    sw2u[0] = 2;
    sw2u[1] = core;
    sw2u[2] = v1[1]/12;

    // the tier2 down switch
    vertex sw2d(3);
    sw2d[0] = 2;
    sw2d[1] = core;
    sw2d[2] = v2[1]/12;

    w.append(g().generate_lwedge(sw1u,sw2u));

    if (sw2d == sw2u) {
      w.append(g().generate_lwedge(sw2u,sw1d));
      w.append(g().generate_lwedge(sw1d,v2));
      return;
    }

    // the tier3 switch
    vertex sw3(3);
    sw3[0] = 3;
    sw3[1] = core;

    sw3[2] = u_0_11;

    w.append(g().generate_lwedge(sw2u,sw3));
    w.append(g().generate_lwedge(sw3,sw2d));
    w.append(g().generate_lwedge(sw2d,sw1d));
    w.append(g().generate_lwedge(sw1d,v2));
  }

  int
  tbird::diameter() const
  {
    return 8;
  }

  bool
  tbird::vertex_symmetric() const
  {
    return true;
  }

  string
  tbird::name() const
  {
    std::ostringstream ostr;
    ostr << "TBird()";
    return ostr.str();
  }

}
