/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_ACTORMAGIC_H
#define SSTMAC_PROCESS_ACTORMAGIC_H

#include <sstmac/process/actorid.h>
#include <map>
#include <set>
#include <boost/shared_ptr.hpp>

namespace sstmac {

  // Forward declaration of mportant types
  class scheduler;
  class key;

  /**
   * Magic functionality for networked actors.
   */
  class actormagic {
  public:
    /// Global registry type for actorid objects.
    typedef std::map<int, actorid> tagmap_t;

    /// Global registry for named counters.
    typedef std::map<std::string, int> counter_t;

  private:
    /// Private nested type to handle a barrier call.
    class barr {
      /// This key gets used to context switch callers out until everybody arrives
      boost::shared_ptr<key> key_;
      /// This gets initialized by the first entrant and defines whom we expect
      std::set<actorid> group_;
      /// This is the set of participants that have arrived to date.
      std::set<actorid> entered_;
      
    public:
      /// Construct a new barrier handler.
      barr();

      /// Join the barrier.  If this is the first entrant, it also
      /// initializes the target group.
      /// No caller will return from this call until the barrier is complete.
      void join(actorid id, const std::set<actorid> &group,
                const boost::shared_ptr<scheduler>& sched);
    };

    /// The scheduler that we use to context switch calling thread out.
    boost::shared_ptr<scheduler> sched_;

    /// We magically know about all processes in our simulation world.
    /// New members join our world, and departing members ask to be removed.
    std::set<actorid> world_;

    /// The caller specifies an integer identifier for a barrier.
    /// The first entrant to the barrier gets to define the group that is
    /// expected to enter the barrier, everybody else needs to provide a
    /// matching group definition.
    /// When a barrier is complete, all callers are released.
    typedef std::map<int, barr> barrier_t;
    barrier_t barrier_;

    /// Global registry of actorid objects (global to a given parallel run).
    tagmap_t tagmap_;

    /// Global registry of counters.
    counter_t counter_;

    /// Constructor is private -- this is standard setup for
    /// reference-counted objects in the simulator.
    actormagic(const boost::shared_ptr<scheduler>& sched);

  public:
    /// External objects can allocate a new object using this static method.
    static boost::shared_ptr<actormagic> 
    construct(const boost::shared_ptr<scheduler>& sched);

    /// A process can be added to the world.
    /// Each process id will only exist once in the world.
    /// \return true if this process was not previously of this world.
    bool add(const actorid &id);

    /// A process can be removed from this world.
    /// \return true if this process was a member of our world prior to this call.
    bool remove(const actorid &id);

    /// Who is in our world?
    const std::set<actorid>& world() const;

    /// Join a barrier.
    /// This call should only be made from a user-space context managed
    /// by our current scheduler (or the scheduler will get very upset at you).
    void barrier(int tag, const actorid &id, const std::set<actorid> &group);

    /// Access the global tag map.
    tagmap_t& tagmap() { return tagmap_; }
    
    /// Access the global tag map in a const context.
    const tagmap_t tagmap() const { return tagmap_; }

    /// Create or move a global counter.
    void move_global_counter(const std::string &label, int increment);

    /// Read a global counter.
    int read_global_counter(const std::string &label) const;
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_ACTORMAGIC_H
