/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_ACTORSERVER_CLIENT_H
#define SSTMAC_PROCESS_ACTORSERVER_CLIENT_H

#include <sstmac/process/actorserver.h>
#include <sstmac/process/actormbox.h>

namespace sstmac {

  /**
   * The client-side part of a distributed actor-based application.
   * This is where all the real work happens.
   */
  class actorserver::client : public actormbox {
    /// Private message types all derive from here -- this mostly frees us
    /// from switching on dynamic casts (not completely though).
    class internalrequest;
    /// Private message type indicating a request to link a process.
    class linkrequest;
    /// Private message type indicating that a link has been established.
    class linkresponse;
    /// Private message type indicating that we would like to unlink.
    class unlinkrequest;
    /// Private event handler type to wait for link being established or removed.
    class linkhandler;
    /// Private message type indicating a request to monitor a process.
    class monitorrequest;
    /// Private message type indicating that a monitor has been established.
    class monitorresponse;
    /// Private message type indicating that we would like to unmonitor.
    class unmonitorrequest;
    /// Private event handler waiting on a monitor to be established or removed.
    class monitorhandler;
    /// Private message type indicating an error exit on a link.
    class linkexit;
    /// Private message type indicating a monitor signal.
    class monitorexit;
    
    /// This object (needed to provide a reference counted pointer
    /// to the application).
    boost::weak_ptr<client> self_;
    
    /// A unique identifier for link or monitor requests.
    int32_t nextunique_;

    /// The identifier for this client.
    actorid id_;
    
    /// The parent server.
    boost::shared_ptr<actorserver> parent_;

    /// Any blocked clients (normally there will only be one, but we
    /// want to ensure that this object can be safely shared as well).
    typedef boost::shared_ptr<actorpattern> pat_t;
    typedef boost::shared_ptr<key> key_t;
    typedef std::pair<pat_t, key_t> entry_t;
    typedef std::vector<blocked_t> blocked_t;
    blocked_t blocked_;

    /// Messages that have not been matched yet
    /// Although it is not required, we also process messages in arrival-order
    /// to make it easier to grok program execution.
    typedef std::pair< actorid, boost::shared_ptr<actormsg> > delivery_t;
    typedef std::list< delivery_t > unmatched_t;
    unmatched_t unmatched_;

    /// Any other clients linked to this client.
    typedef std::set<actorid> links_t;
    links_t links_;

    /// Pending request for links to be established (normally zero or one).
    typedef std::map< int32_t, boost::shared_ptr<linkhandler> > linkreq_t;
    linkreq_t linkreq_;

    /// Any clients that are monitoring this client.
    typedef std::set<actorid> monitors_t;
    monitors_t monitors_;

    /// Construct a new client object.
    client(actorid id, const boost::shared_ptr<actorserver>& parent);

  public:
    /// Create a new client.
    static boost::shared_ptr<client>
    construct(actorid id, const boost::shared_ptr<actorserver>& parent);

    /// All done.
    virtual ~client();

    /// This method is required by the actorserver.
    void handle(const actorid &sender, boost::shared_ptr<actormsg> payload);

    /* Everything else is required by the actormbox base class */

    /// What's my pid.
    virtual actorpid mypid() const { return id_; }

    /// What nodes are known to the current system?
    virtual const std::vector<nodeid>& nodes() const;

    /// Spawn a process on the given node.  Returns the pid of the new process.
    /// Spawning consists of cloning the input object, assigning it a new pid,
    /// and calling initialize to get the object going.
    /// \param task  The object (thread) that will be spawned.
    /// \param size  How many bytes should we consider this transfer to be?
    /// \param tout  How long are we willing to wait before timing out.
    ///              A timeout of less than zero means we will wait forever.
    /// \return      the pid for the new object, or actorpid::invalid
    ///              if the spawn fails (times out).
    virtual actorpid spawn(const boost::shared_ptr<const actortask>& task,
                         int64_t size, timestamp tout);

    /// Spawn a process on the given node and link it to the current process.
    /// \param task    The task object that will be cloned on the remote node.
    /// \param size    The number of bytes that we (pretend) will be needed
    ///                to send the object across a network.
    /// \param tout    How long we are willing to wait.  Negative values
    ///                indicate that we will never time out.
    /// \return        the handle for the new process or
    ///                actorid::invalid() if the call fails.
    virtual actorpid spawn_link(const boost::shared_ptr<const actortask>& task,
                              int64_t size, timestamp tout);

    /// Spawn a process on the given node and set up a monitor for it.
    /// \param task    The task object that will be cloned on the remote node.
    /// \param size    The number of bytes that we (pretend) will be needed
    ///                to send the object across a network.
    /// \param tout    How long we are willing to wait.  Negative values
    ///                indicate that we will never time out.
    /// \return        the handle for the new process or
    ///                actorid::invalid() if the call fails.
    virtual actorpid spawn_monitor(const boost::shared_ptr<const actortask>& task,
                                 int64_t size, timestamp tout);

    /// Create a link to the given target.
    /// A link is unique -- an attempt to re-do an existing link is ignored.
    /// \param target  The handle for the target we are linking ourselves to.
    /// \return        true if the link was formed or already existed,
    ///                false if the request timed out or host was unreachable.
    virtual bool link(const actorid &target, timestamp tout);

    /// Remove our link to the given target.
    /// \param target  The handle for the target we are unlinking ourselves from.
    /// \return        true if the link was removed or did not exist.
    ///                false if the request timed out or host was unreachable.
    virtual bool unlink(const actorid &target, timestamp tout);

    /// Create a monitor for the given target.
    /// This is quietly ignored if the target is no longer in existence.
    /// A monitor is unique -- you will get exactly one actordownmsg when the
    /// process exits (this differs from the behavior of Erlang).
    /// \param target  The handle for the target we are monitoring.
    /// \return        true if the monitor was formed or already existed,
    ///                false if the request timed out or host was unreachable.
    virtual bool monitor(const actorid &target, timestamp tout);

    /// Remove our monitor for the given target.
    /// \param target  The handle for the target we want to stop monitoring.
    /// \return        true if the monitor was removed or did not exist.
    ///                false if the request timed out or host was unreachable.
    virtual bool unmonitor(const actorid &target, timestamp tout);
    
    /// Send a message to the given pid.
    /// This call always succeeds except for invalid pid.
    /// \param target  The handle for the target we are sending to.
    /// \param msg     The content we want to package with the message.
    /// \param size    The number of bytes we are pretending the message takes.
    /// \throw valueerror if the pid is not valid.
    /// \throw nullerror if sending a NULL message (reserved for recv timeout).
    virtual void send(const actorpid &target, const boost::shared_ptr<const actormsg>& msg,
                      int64_t size);

    /// Test whether there is any message that matches the given pattern.
    /// This has the same effect as calling recv with a zero-length timeout.
    /// \param source  Our desired source (may be actorpid::any).
    /// \param pat     The pattern we are matching (may be actorpattern::any).
    /// \return        true if a matching message is found, false otherwise.
    virtual bool test(const actorpid &source,
                      const boost::shared_ptr<const actorpattern>& pat) const;
    
    /// Block until we get a message matching the given pattern.
    /// \param source  Our desired source (may be actorpid::any).
    /// \param pat     The pattern we are matching (may be actorpattern::any).
    /// \param tout    How long we are willing to wait.  Negative values
    ///                indicate that we will never time out.
    /// \return        std::make_pair(actual_source, actual_payload) on success
    ///                or std::make_pair(actorpid::invalid, NULL) if timed out.
    virtual std::pair< actorid, boost::shared_ptr<actormsg> >
    recv(const actorpid &source,
         const boost::shared_ptr<const actorpattern>& pat,
         timestamp tout);

    /// Indicate that the current process is exiting for the given reason.
    /// If the reason is not actorexit::normal, all monitors will be sent an
    /// exit message and all linked processes will get a copy of the
    /// exit reason.
    /// This call must only be called from the process itself (otherwise
    /// the scheduler will likely throw an exception at you).
    /// This call (effectively) doesn't return -- the calling thread is 
    /// context switched out and removed from the scheduler.
    /// \param reason  The reason that gets forwarded to all monitors and links.
    virtual void exit(const boost::shared_ptr<const actorexit>& reason);
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_ACTORSERVER_CLIENT_H
