/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_APPLAUNCH_H
#define SSTMAC_PROCESS_APPLAUNCH_H

#include <string>
#include <map>
#include <boost/shared_ptr.hpp>
#include <sstmac/simulator/launcher.h>
#include <sstmac/util/nocopy.h>
#include <boost/weak_ptr.hpp>

namespace sstmac {

  // Important types.
  class scheduler;
  class app;
  class taskid;

  /**
   * A launcher for apps 
   */
  class applaunch : public launcher, virtual public nocopy {
    
    /// Private eventhandler type to take care of threads exiting.
    class completion;
    
    /// This object.
    boost::weak_ptr<applaunch> self_;
    
    /// The scheduler that manages thread context.
    boost::shared_ptr<scheduler> scheduler_;
    
    /// Map applications to task id's
    std::map<taskid,boost::shared_ptr<app> > apps_;
    
    /// The environment we are running under.
    /// This gets set when run(...) is called and is freed when
    /// scheduler_exit(...) is called.
    boost::shared_ptr<environment> environment_;
    
    /// Private method that gets called by completion when all threads are done.
    void scheduler_exit(const timestamp &exit_time);
    
  protected:
    /// Protected constructor.
    applaunch();
    
    /// Who am I?
    void set_self(const boost::shared_ptr<applaunch>& self);
    
    /// This is who I am.
    boost::shared_ptr<applaunch> get_self();
    
    /// This should be called at the beginning of any run(...) method,
    /// including those defined by derived classes.
    /// It performs basic setup of the scheduler and environment. 
    void run_setup(const boost::shared_ptr<environment>& my_world);

  public:
    /// Public constructor.
    static boost::shared_ptr<applaunch> construct();

    /// Destructor
    virtual ~applaunch() throw();

    /// Set map of applications which will get added to the scheduler.
    /// Derived launchers, such as mpilaunch, may use a different
    /// approach to add applications.
    void set_applications(std::map<taskid,boost::shared_ptr<app> > apps);
    
    /// Access the scheduler.
    boost::shared_ptr<scheduler> get_scheduler();

    /// Define the scheduler.
    void set_scheduler(const boost::shared_ptr<scheduler>& sched);

    /// Access the environment
    boost::shared_ptr<environment> get_environment();

    /// Define the environment
    void set_environment(const boost::shared_ptr<environment>& env);

    /// Provide description string.
    /// Inherited from launcher.
    virtual const std::string& description() const;

    /// Adds application thread to and runs scheduler
    /// Inherited from launcher 
    virtual void run(const timestamp &current_time,
                     const boost::shared_ptr<environment>& my_world);

    /// Are we done yet?
    /// Inherited from launcher
    bool completed() const;
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_APPLAUNCH_H
