/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_CONTEXT_H
#define SSTMAC_PROCESS_CONTEXT_H

#include <sstmac/sstmacconfig.h>
#include <sstmac/util/errors.h>

#ifndef SSTMAC_USE_PTH
#include <ucontext.h>
#include <sstmac/process/context-util.h>
#else // SSTMAC_USE_PTH
#include <pth.h>
#endif // ! SSTMAC_USE_PTH

#include <errno.h>
#include <string.h>
#include <iostream>

namespace sstmac {

#ifndef SSTMAC_USE_PTH

  /*
   * Functionality to use ucontext for lightweight threads.
   */
  
  /// The context type.
  typedef ucontext_t threadcontext_t;

  /// Initializing a context.
  inline void init_context(threadcontext_t &context) {
    const std::string here("init_context(ucontext_t &context):  ");
    if(getcontext(&context) != 0) {
      throw schedulererror(here + strerror(errno));
    }
  }

  inline void destroy_context(threadcontext_t &) {
  }

  /// Start a new context.
  inline void start_context(threadcontext_t &context,
                            void *stack, size_t stacksize,
                            void(*func)(void*), void *args,
                            threadcontext_t *yield_to)
  {
    funcptr funcp(func);
    voidptr voidp(args);
    context.uc_stack.ss_sp = stack;
    context.uc_stack.ss_size = stacksize;
    init_context(context);
    context.uc_link = yield_to;
    makecontext(&context, (void(*)())(context_springboard), 4,
                funcp.fpair.a, funcp.fpair.b, voidp.vpair.a, voidp.vpair.b);
  }

  /// Swap context.
  inline void swap_context(threadcontext_t &from, threadcontext_t &to) {
    static const std::string here("switch_context(ucontext_t&, ucontext_t&): ");
    if(swapcontext(&from, &to) == -1) {
      throw schedulererror(here + strerror(errno));
    }
  }

#else // SSTMAC_USE_PTH

  /*
   * Functionality to use pth for lightweight threads
   */
  typedef pth_uctx_t threadcontext_t;

  /// Initializing a context.
  inline void init_context(threadcontext_t &context) {
    const std::string here("init_context(pth_uctx_t &context):  ");
    if(pth_uctx_create(&context) != TRUE) {
      throw schedulererror(here + strerror(errno));
    }
  }

  inline void destroy_context(threadcontext_t &context) {
    const std::string here("destroy_context(pth_uctx_t&):  ");
    if(pth_uctx_destroy(context) != TRUE) {
      throw schedulererror(here + strerror(errno));
    }
  }

  /// Start a new context.
  inline void start_context(threadcontext_t &context,
                            void *stack, size_t stacksize,
                            void(*func)(void*), void *args,
                            threadcontext_t *yield_to)
  {
    const std::string here("start_context(pth_uctx_t, ...):  ");
    if(stacksize < (16384))
      throw valueerror(here + "PTH does not accept stacks smaller than 16KB.");
    init_context(context);
    int retval = pth_uctx_make(context, (char*)stack, stacksize, NULL,
                               func, args, (yield_to ? *yield_to : NULL));
    if(retval != TRUE) {
        throw schedulererror(here + strerror(errno));
    }
  }

  /// Swap context.
  inline void swap_context(threadcontext_t &from, threadcontext_t &to) {
    static const std::string here("switch_context(ucontext_t&, ucontext_t&): ");
#ifdef HAVE_PTH_UCTX_SWITCH_IGNORE_SIGMASK
    if(pth_uctx_switch_ignore_sigmask(from, to) != TRUE) {
      throw schedulererror(here + strerror(errno));
    }
#else
    if(pth_uctx_switch(from, to) != TRUE) {
      throw schedulererror(here + strerror(errno));
    }
#endif
  }

#endif // ! SSTMAC_USE_PTH

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_CONTEXT_H
