/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// The implementation of a lock for lightweight threads.

#include <sstmac/process/lock.h>
#include <sstmac/process/key.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/timestamp.h>
#include <sstmac/util/errors.h>
#include <iostream>
#include <assert.h>

namespace sstmac {

  using boost::shared_ptr;

  //
  // Nested unlocking type.
  //
  class lock::unlocker : public eventhandler {
    /// The key we will be releasing.
    shared_ptr<key> key_;
    /// The scheduler that will handle the actual unblocking.
    shared_ptr<scheduler> sched_;
  protected:
    /// Hi.
    unlocker(const shared_ptr<key>& key_in, const shared_ptr<scheduler>& sched_in) :
      key_(key_in), sched_(sched_in)
    {}
  public:
    /// Hi.
    static shared_ptr<unlocker> construct(const shared_ptr<key>& key_in,
                                          const shared_ptr<scheduler>& sched_in)
    {
      return shared_ptr<unlocker>(new unlocker(key_in, sched_in));
    }

    /// Bye.
    virtual ~unlocker() throw() {}

    /// Done.
    void handle(const timestamp&) {
      if(sched_.get() != 0) {
        sched_->unblock(key_);
        sched_ = shared_ptr<scheduler>();
      }
    }
  };

  //
  // Hello.
  //
  lock::lock(const boost::shared_ptr<scheduler>& sched) : sched_(sched), locked_(false)
  {}

  //
  // Wheee!
  //
  shared_ptr<lock> lock::construct(const shared_ptr<scheduler>& s) {
    return shared_ptr<lock>(new lock(s));
  }

  //
  // Goodbye.
  //
  lock::~lock() throw() {
    if(locked_) {
      std::cerr << "lock::~lock:  Lock going out of scope while engaged.";
      assert(false);
    }
  }

  //
  // Engage the lock.
  //
  void lock::grab() {
    if(locked_) {
      keys_.push_back(key::construct());
      sched_.lock()->block(keys_.back());
    }
    // We get here if we were not locked or if we have now been unblocked.
    locked_ = true;
  }

  //
  // Try to grab the lock.
  //
  bool lock::try_grab() {
    if(! locked_) {
      locked_ = true;
      return true;
    }
    else {
      return false;
    }
  }

  //
  // Grab the lock unless the given event comes up first.
  //
  bool lock::timed_grab(const timestamp &maxwait,
                        const boost::shared_ptr<eventmanager>& mgr)
  {
    if(locked_) {
      keys_.push_back(key::construct());
      shared_ptr<eventhandler> evt =
        unlocker::construct(keys_.back(), sched_.lock());
      mgr->schedule(mgr->now()+maxwait, evt);
      sched_.lock()->block(keys_.back());
    }
    // We get here if we were not locked or if we have now been unblocked.
    if(! locked_) {
      locked_ = true;
      return true;
    }
    else {
      return false;
    }
  }

  //
  // Release the lock.
  //
  void lock::release() {
    locked_ = false;
    if(! keys_.empty()) {
      shared_ptr<key> the_key = keys_.front();
      keys_.pop_front();
      sched_.lock()->unblock(the_key);
    }
  }

} // end of namespace sstmac.
