/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpibarrierengine.h>
#include <sstmac/process/mpiqueue.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/util/topologyiterator.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;

  // A utility type for sendrecv
  // I figured out that I can actually end up recursively inside sendrecv,
  // which will then result in an invalid iterator and major screw-ups.
  struct sendrecvop {
    bool sending;
    mpiid partner;
  };

  //
  // Fire off the next set of send and receive operations.
  //
  void mpibarrierengine::sendrecv(const timestamp &now) {
    if(debugger::active())
      debug() << "mpibarrierengine[" << this << "]:: sendrecv(" << now << ")\n";
    if(pending_recvs_) {
      throw illformederror("mpibarrierengine::sendrecvs: Pending receive.");
    }
    std::vector<sendrecvop> ops;
    while(iter_->active()) {
      if(iter_->sending()) {
        sendrecvop op = {true, mpiid(iter_->dest()) };
        ops.push_back(op);
      }
      if(iter_->receiving()) {
        sendrecvop op = {false, mpiid(iter_->source()) };
        ops.push_back(op);
        // We can only safely handle one receive at a time.
        // (a barrier will generally receive zero or one time,
        // followed by zero or more sends).
        iter_->next();
        break;
      }
      iter_->next();
    }
    // Fire off all sends and receives.
    for(size_t it = 0; it < ops.size(); ++it) {
      if(!ops[it].sending) {
        ++pending_recvs_;
        this->start_recv(1, mpitype::mpi_byte, tag_, ops[it].partner);
      }
      else {
        ++pending_sends_;
        this->start_send(1, mpitype::mpi_byte, tag_, ops[it].partner,
                         payload::null());
      }
    }

    // If we have no pending sends or receives, we are done.
    if(pending_sends_ == 0 && pending_recvs_ == 0) {
      mpicollective::complete(now, shared_ptr<mpicollpayload>());
    }
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpibarrierengine::send_complete(const timestamp &now,
                                       int count,
                                       const mpitype &type,
                                       const mpitag &tag,
                                       const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpibarrierengine[" << this << "]::send_complete(" <<now <<", "
              << count << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    if(! pending_recvs_) {
      this->sendrecv(now);
    }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpibarrierengine::recv_complete(const timestamp &now,
                                       int count,
                                       const boost::shared_ptr<const mpistatus>& stat)
  {
    if(debugger::active())
      debug() << "mpibcastengine[" << this << "]::recv_complete(" << now << ", "
              << count << ", " << stat << ")\n";
    --pending_recvs_;
    if(! pending_recvs_) {
      this->sendrecv(now);
    }
  }

  //
  // Copy complete.
  //
  void mpibarrierengine::copy_complete(const timestamp&, int, const mpitype&,
                                       const mpitag&, const mpiid&)
  {
    throw illformederror("mpibarrierengine::copy_complete:  "
                         "MPI barrier doesn't copy data.");
  }

  //
  // Hi.
  //
  mpibarrierengine::mpibarrierengine(const shared_ptr<topologyiterator>& iter,
                                     const shared_ptr<key>& the_key,
                                     const shared_ptr<mpiqueue>& queue,
                                     const mpitag &tag,
                                     const mpicomm &comm) :
    mpicollective(the_key, queue, comm), iter_(iter), tag_(tag),
    pending_sends_(0), pending_recvs_(0)
  {
    if(tag_ == mpitag::any_tag)
      throw valueerror("mpibarrierengine(...)"
                       ":  refusing to send using mpitag::any_tag.");
  }

  //
  // Hi.
  //
  shared_ptr<mpibarrierengine>
  mpibarrierengine::construct(const shared_ptr<topologyiterator>& iter,
                              const shared_ptr<key>& the_key,
                              const shared_ptr<mpiqueue>& queue,
                              const mpitag &tag,
                              const mpicomm &comm)
  {
    shared_ptr<mpibarrierengine> retval
      (new mpibarrierengine(iter, the_key, queue, tag, comm));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpibarrierengine::~mpibarrierengine() throw() {
  }

  //
  // Start this kernel.
  //
  void mpibarrierengine::start(const timestamp &start_time,
                               const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpibarrierengine[" << this << "]::start(" << start_time
              << ", " << my_node << ")\n";
    mpicollective::start(start_time, my_node);
    this->sendrecv(start_time);
  }

} // end of namespace sstmac

