/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpibcastengine.h>
#include <sstmac/process/mpiqueue.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/util/topologyiterator.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;

  // A utility type for sendrecv
  // Although this is very unlikely to happen for bcast, there is the
  // possibility of ending up with recursive calls to sendrecv (and an
  // invalid iterator) if we don't separate iteration from communications
  struct sendrecvop {
    bool sending;
    mpiid partner;
  };

  //
  // Start next cycle of sends and receives.
  //
  void mpibcastengine::sendrecv(const timestamp &now) {
    if(debugger::active())
      debug() << "mpibcastengine[" << this << "]:: sendrecv(" << now << ")\n";
    if(pending_recvs_) {
      throw illformederror("mpibcastengine::sendrecvs: Pending receive.");
    }
    std::vector<sendrecvop> ops;
    while(iter_->active()) {
      if(iter_->sending()) {
        sendrecvop op = { true, mpiid(iter_->dest()) };
        ops.push_back(op);
      }
      if(iter_->receiving()) {
        sendrecvop op = { false, mpiid(iter_->source()) };
        ops.push_back(op);
        // We can only safely handle one receive at a time.
        // (a bcast will generally receive zero or one time,
        // followed by zero or more sends).
        iter_->next();
        break;
      }
      iter_->next();
    }
    // Do any sends/receives.
    for(size_t it = 0; it < ops.size(); ++it) {
      if(! ops[it].sending) {
        ++pending_recvs_;
        this->start_recv(count_, type_, tag_, ops[it].partner);
      }
      else {
        ++pending_sends_;
        this->start_send(count_, type_, tag_, ops[it].partner, content_);
      }
    }
    // If we have no pending sends or receives, we are done.
    if(pending_sends_ == 0 && pending_recvs_ == 0) {
      mpicollective::complete(now, content_);
    }
  }

  //
  // Hi.
  //
  mpibcastengine::mpibcastengine(const shared_ptr<topologyiterator>& iter,
                                 const shared_ptr<key>& thekey,
                                 const shared_ptr<mpiqueue>& queue, int count,
                                 const mpitype &type, const mpiid &root,
                                 const mpitag &tag, const mpicomm &comm,
                                 const boost::shared_ptr<const payload>& content) :
    mpicollective(thekey, queue, comm), iter_(iter),
    count_(count), type_(type), tag_(tag), rank_(comm.rank()),
    pending_sends_(0), pending_recvs_(0)
  {
    if(comm.rank() == root) {
      if(content) content_ = content->clone();
    }
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpibcastengine::send_complete(const timestamp &now,
                                     int count,
                                     const mpitype &type,
                                     const mpitag &tag,
                                     const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpibcastengine[" << this << "]::send_complete(" << now << ", "
              << count << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    if(! pending_recvs_)
      this->sendrecv(now);
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpibcastengine::recv_complete(const timestamp &now, int count,
                                     const boost::shared_ptr<const mpistatus>& stat)
  {
    if(debugger::active())
      debug() << "mpibcastengine[" << this << "]::recv_complete(" << now << ", "
              << ", " << count << stat << ")\n";
    content_ = stat->content();
    this->start_immediate_copy(count, stat->recv_type(),
                               stat->tag(), stat->source());
  }

  //
  // Completion of a copy event.
  //
  void mpibcastengine::copy_complete(const timestamp &now,
                                     int count,
                                     const mpitype &type,
                                     const mpitag &tag,
                                     const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpibcastengine[" << this << "]::copy_complete(" << now
              << ", " << count << ", " << type << ", " << tag << ", "
              << source << ")\n";
    --pending_recvs_;
    this->sendrecv(now);
  }

  //
  // Hello.
  //
  shared_ptr<mpibcastengine>
  mpibcastengine::construct(const shared_ptr<topologyiterator>& iter, 
                            const shared_ptr<key>& thekey, const shared_ptr<mpiqueue>& queue,
                            int count, const mpitype &type, const mpiid &root,
                            const mpitag &tag, const mpicomm &comm,
                            const boost::shared_ptr<const payload>& content)
  {
    shared_ptr<mpibcastengine> retval
      (new mpibcastengine(iter, thekey, queue, count, type,
                          root, tag, comm, content));
    retval->set_self(retval);
    return retval;
  }
  
  //
  // Goodbye.
  //
  mpibcastengine::~mpibcastengine() throw() {
  }

  //
  // Start this kernel.
  //
  void mpibcastengine::start(const timestamp &start_time,
                             const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpibcastengine[" << this << "]::start(" << start_time
              << ", " << my_node << ")\n";
    mpicollective::start(start_time, my_node);
    this->sendrecv(start_time);
  }

} // end of namespace sstmac

