/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpibtreegather.h>
#include <sstmac/process/mpiqueue.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/mpisource.h>
#include <sstmac/process/mpimagic.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/util/btscatteriterator.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <deque>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;

  //
  // Given an absolute rank, return the relative rank.
  //
  int mpibtreegather::relrank(const mpiid &therank) const {
    return ((therank.id - root_.id + comm_.size().id) % comm_.size().id);
  }

  //
  // Given a relative rank, return the absolute rank.
  //
  mpiid mpibtreegather::absrank(int therank) const {
    return mpiid((therank + root_.id) % comm_.size().id);
  }

  //
  // Fire off any and all receives concurrently.
  //
  void mpibtreegather::init_recvs(const timestamp &now) {
    if(debugger::active())
      debug() << "mpibtreegather::init_recvs(" << now << ")\n";
    const int relative_rank = relrank(comm_.rank());
    const int size = comm_.size().id;
    int keymask = 1;
    if(relative_rank == 0) {
      int v = size;
      int shift = 0;
      while(v) {
        ++shift;
        v >>= 1;
      }
      keymask = 1 << shift;
    }
    else {
      while(! (relative_rank & keymask)) {
        keymask <<= 1;
      }
    }
    // Figure out the sources for this node.
    int localsrc = 1;
    std::vector<mpisource> sources;
    while(localsrc < keymask) {
      const int source = relative_rank + localsrc;
      if(source >= size)
        break;
      const int entries = ((source+localsrc <= size) ? localsrc : size-source);
      const int length = entries * recvcount_;
      ++pending_recvs_;
      sources.push_back(mpisource(absrank(source), length));
      localsrc <<= 1;
    }
    if(pending_recvs_) {
      // Fire off all receives at once.
      for(size_t i = 0; i < sources.size(); ++i) {
        this->start_recv(sources[i].count, recvtype_, tag_, sources[i].source);
      }
    }
    else {
      // Leaf node -- no receives, just sends.
      init_send(now);
    }
  }

  //
  // Fire off the send operation (if any).  Can be called from root.
  //
  void mpibtreegather::init_send(const timestamp &now) {
    if(debugger::active())
      debug() << "mpibtreegather::init_send(" << now << ")\n";
    // We should only be here when all receives are done.
    if(pending_recvs_) {
      throw illformederror("mpibtreegather::init_send:  Pending receives.");
    }
    const int relative_rank = relrank(comm_.rank());
    const int size = comm_.size().id;
    if(relative_rank != 0) {
      // We need to do a send.
      int keymask = 1;
      while(! (relative_rank & keymask))
        keymask <<= 1;
      const int localdest = relative_rank - keymask;
      const int entries = ((relative_rank+keymask) <= size ?
                           keymask : size-relative_rank);
      const int length = entries * sendcount_;
      ++pending_sends_;
      this->start_send(length, sendtype_, tag_, absrank(localdest));
    }
    else if(pending_sends_ <= 0 && pending_recvs_ <= 0) {
      // Complete
      shared_ptr<const payload> content;
      if((comm_.rank() == root_)) {
        if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
          throw illformederror("mpibtreegather::init_send:  Root is done "
                               "but not all payload has been gathered.");
        content = magic_->get_collective_payload(typeid(*this), tag_, comm_);
      }
      mpicollective::complete(now, content);
    }
  }

  //
  // Hi.
  //
  mpibtreegather::mpibtreegather(const shared_ptr<key>& thekey,
                                 const shared_ptr<mpiqueue>& queue,
                                 int sendcount, const mpitype &sendtype,
                                 int recvcount, const mpitype &recvtype,
                                 const mpiid &root, const mpitag &tag,
                                 const mpicomm &comm,
                                 const shared_ptr<const payload>& content,
                                 const shared_ptr<mpimagic>& magic) :
    mpicollective(thekey, queue, comm),
    sendcount_(sendcount), sendtype_(sendtype),
    recvcount_(recvcount), recvtype_(recvtype),
    root_(root), tag_(tag), comm_(comm), magic_(magic),
    pending_sends_(0), pending_recvs_(0)
  {
    bool isroot = comm_.rank() == root_;
    // Clone the content if it is non-null.
    magic_->add_collective_payload(typeid(*this), tag_, comm_,
                                   (content ? content->clone() : content),
                                   isroot);
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpibtreegather::send_complete(const timestamp &now, int count,
                                     const mpitype &type, const mpitag &tag,
                                     const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpibtreegather::send_complete(" << now << ", " << count
              << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    if(! pending_sends_) {
      shared_ptr<const payload> content;
      if((comm_.rank() == root_)) {
        if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
          throw illformederror("mpibtreegather::send_complete:  Root is done "
                               "but not all payload has been gathered.");
        content = magic_->get_collective_payload(typeid(*this), tag_, comm_);
      }
      mpicollective::complete(now, content);
    }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpibtreegather::recv_complete(const timestamp &now, int count,
                                     const shared_ptr<const mpistatus>& stat)
  {
    if(debugger::active())
      debug() << "mpibtreegather::recv_ complete(" << now << ", " << count
              << ", " << stat << ")\n";
    this->start_funneled_copy(count, stat->recv_type(),
                              stat->tag(), stat->source());
  }

  //
  /// Callback for the copier to signal completion.
  //
  void mpibtreegather::copy_complete(const timestamp &now, int count,
                                     const mpitype &type, const mpitag &tag,
                                     const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpibtreegather::copy_done(" << now << ", " << count << ", "
              << type << ", " << tag << ", " << source << ")\n";
    --pending_recvs_;
    if(! pending_recvs_) {
      this->init_send(now);
    }
  }

  //
  // Hello.
  //
  shared_ptr<mpibtreegather>
  mpibtreegather::construct(const shared_ptr<key>& thekey,
                            const shared_ptr<mpiqueue>& queue,
                            int sendcount, const mpitype &sendtype,
                            int recvcount, const mpitype &recvtype,
                            const mpiid &root, const mpitag &tag,
                            const mpicomm &comm,
                            const shared_ptr<const payload>& content,
                            const shared_ptr<mpimagic>& magic)
  {
    shared_ptr<mpibtreegather> retval
      (new mpibtreegather(thekey, queue, sendcount, sendtype, recvcount,
                          recvtype, root, tag, comm, content, magic));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpibtreegather::~mpibtreegather() throw() {
  }

  //
  // Start this kernel.
  //
  void mpibtreegather::start(const timestamp &now,
                             const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpibtreegather[" << this << "]::start(" << now
              << ", " << my_node << ")\n";
    mpicollective::start(now, my_node);
    this->init_recvs(now);
  }


} // end of namespace sstmac

