/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPICOLLPAYLOAD_H
#define SSTMAC_PROCESS_MPICOLLPAYLOAD_H

#include <sstmac/process/payload.h>
#include <sstmac/util/errors.h>
#include <vector>

namespace sstmac {

  /**
   * Keep a mapping from some form of sortable entities to message payloads.
   * Intended to make it easier to manage various collective messaging ops.
   */
  class mpicollpayload : public payload {
  public:
    /// The data storage type
    typedef std::vector< boost::shared_ptr<const payload> > content_t;

  private:
    /// The actual data map.
    content_t content_;

    /// Hello.
    mpicollpayload(size_t target_size) :
      content_(target_size)
    {}
    
    /// Private copy constructor.
    mpicollpayload(const mpicollpayload &other) :
      payload(other), content_(other.content_)
    {}

  public:
    /// Get a new object.
    static boost::shared_ptr<mpicollpayload> construct(size_t target_size) {
      return boost::shared_ptr<mpicollpayload>(new mpicollpayload(target_size));
    }

    /// Goodbye.
    virtual ~mpicollpayload() throw() {}

    /// Clone the map.
    boost::shared_ptr<mpicollpayload> realclone() const {
      return boost::shared_ptr<mpicollpayload>(new mpicollpayload(*this));
    }

    /// Clone the map.
    boost::shared_ptr<payload> clone() const {
      return realclone();
    }

    /// Access the content in a const context
    /// (we don't allow non-const access).
    const content_t& value() const { return content_; }

    /// Access a value or die trying.
    boost::shared_ptr<const payload> get(mpiid id) const {
      if((id.id < 0) || (size_t(id.id) >= content_.size()))
        throw valueerror("mpicollpayload::get(id):  Rank out of range.");
      return content_[id.id];
    }

    /// Insert a value (convenience function).
    void set(mpiid key, const boost::shared_ptr<const payload>& value) {
      if((key.id < 0) || (size_t(key.id) >= content_.size()))
        throw valueerror("mpicollpayload::set(key, value)  Rank out of range.");
      content_[key.id] = value;
    }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPICOLLPAYLOAD_H
