/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpihalvingreducescatter.h>
#include <sstmac/process/mpicollpayload.h>
#include <sstmac/process/mpimagic.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::dynamic_pointer_cast;

  //
  // A private method to perform data exchanges.
  //
  void mpihalvingreducescatter::sendrecv(const timestamp &now) {
    if(pending_recvs_) // We can't send anything until our receive is done.
      return;
    if(! peers_.empty()) {
      commstep curr = peers_.front();
      peers_.pop_front();
      // Incrementing sends and receives separately - just so we don't
      // end up entering sendrecv too early.
      if(curr.sendbytes > 0) ++pending_sends_;
      if(curr.recvbytes > 0) ++pending_recvs_;
      if(curr.sendbytes > 0)
        this->start_send(curr.sendbytes, type_, tag_, curr.peer);
      if(curr.recvbytes > 0)
        this->start_recv(curr.recvbytes, type_, tag_, curr.peer);
    }
    else if(pending_sends_ == 0 && pending_recvs_ == 0) {
      // We're done.
      if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
        throw illformederror("mpihalvingreducescatter::sendrecv: Transfer done "
                               "but not all payload has been gathered.");
      shared_ptr<const payload> content =
        magic_->get_collective_payload(typeid(*this), tag_, comm_);
      mpicollective::complete(now, content);
    }
  }

  //
  // Hello.
  //
  mpihalvingreducescatter::
  mpihalvingreducescatter(const shared_ptr<key>& thekey,
                          const shared_ptr<mpiqueue>& queue,
                          const std::vector<int> &recvcnts,
                          const mpitype &type, const mpiop &op,
                          const mpitag &tag, const mpicomm &comm,
                          const shared_ptr<const payload>& content,
                          const shared_ptr<mpimagic>& magic) :
    mpicollective(thekey, queue, comm),
    recvcnts_(recvcnts), type_(type), op_(op), tag_(tag), comm_(comm),
    magic_(magic), pending_sends_(0), pending_recvs_(0)
  {
    // Clone the content if it is non-null
    magic_->add_collective_payload(typeid(*this), tag, comm,
                                   (content ? content->clone() : content),
                                   true);
    if(recvcnts_.size() != size_t(comm.size().id))
      throw valueerror("mpihalvingreducescatter(...): Number of peers "
                       "differs from the number of receive counts.");
    std::vector<int>::const_iterator it, end = recvcnts.end();
    for(it = recvcnts.begin(); it != end; ++it)
      if(*it < 0)
        throw valueerror("mpihalvingreducescatter(...):  "
                         "Found negative receive count.");
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpihalvingreducescatter::send_complete(const timestamp &now,
                                             int count,
                                             const mpitype &type,
                                             const mpitag &tag,
                                             const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpihalvingreducescatter::send_complete(" << now << ", "
              << count << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    this->sendrecv(now);
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpihalvingreducescatter::recv_complete(const timestamp &now,
                                              int count,
                                              const shared_ptr<const mpistatus>& stat)
  {
    if(debugger::active())
      debug() << "mpihalvingreducescatter::recv_complete(" << now << ", "
              << count << ", " << stat <<")\n";
    this->start_funneled_copy(count, stat->recv_type(),
                              stat->tag(), stat->source());
  }

  //
  // Callback to signal copy completion.
  //
  void mpihalvingreducescatter::copy_complete(const timestamp &now,
                                              int count,
                                              const mpitype &type,
                                              const mpitag &tag,
                                              const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpihalvingreducescatter::copy_complete(" << now << ", "
              << count << ", " << type << ", " << tag << ", " << source<< ")\n";
    --pending_recvs_;
    this->sendrecv(now);
  }

  //
  // Hello.
  //
  shared_ptr<mpihalvingreducescatter>
  mpihalvingreducescatter::construct(const shared_ptr<key>& thekey,
                                     const shared_ptr<mpiqueue>& queue,
                                     const std::vector<int> &cnts,
                                     const mpitype &type, const mpiop &op,
                                     const mpitag &tag, const mpicomm &comm,
                                     const shared_ptr<const payload>& content,
                                     const shared_ptr<mpimagic>& magic)
  {
    shared_ptr<mpihalvingreducescatter> retval
      (new mpihalvingreducescatter(thekey, queue, cnts, type,
                                   op, tag, comm, content, magic));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpihalvingreducescatter::~mpihalvingreducescatter() throw() {
  }

  //
  // Utility routine to find communication partner given a pivot.
  //
  inline int partner(int rank, int pivot) {
    if(rank >= (pivot<<1))
      throw valueerror("mpihalvingreducescatter::start::partner: "
                       "rank out of valid range.");
    return (rank < pivot ? rank + pivot : rank - pivot);      
  }
  //
  // Utility routine to find the power-of-two rank for a given index.
  //
  inline int to_pof2(const mpiid &rank, int extras) {
    if(rank.id >= 2*extras) return rank.id-extras;
      else return (rank.id%2 ? -1 : rank.id/2);
  }
  //
  // Utility routine to convert a power-of-two rank to an MPI rank.
  //
  inline mpiid from_pof2(int rank, int extras) {
    if(rank < 0)
      throw valueerror("mpihalfringreducescatter::start::from_pof2: "
                       "Negative rank.");
    return mpiid(rank > extras ? rank+extras : rank*2);
  }
  //
  // Utility routine to find the greatest power of two not greater than value
  //
  inline int msb(int value) {
    if(value < 0)
      throw valueerror("mpihalvingreducescatter::start::msb: Negative value");
    int rv = 1;
    while(rv <= value)
      rv <<= 1;
    return (rv>>1);
  }
  
  //
  // Get busy child.
  //
  void mpihalvingreducescatter::start(const timestamp &now,
                                      const shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpihalvingreducescatter::start(" << now << ", "
              << my_node << ")\n";
    mpicollective::start(now, my_node);
    //
    // Figure out our entire pattern of sends and receives.
    const int size = msb(comm_.size().id);
    const int extras = comm_.size().id - size;
    const int rank = to_pof2(comm_.rank(), extras);
    // Set up mappings of how much we are sending/receiving.
    std::vector<int> sizes;  sizes.reserve(size);
    int total_elements = 0;
    for(size_t it = 0; it < recvcnts_.size(); ++it) {
      total_elements += recvcnts_.at(it);
      if(to_pof2(mpiid(it), extras) < 0)
        sizes.back() += recvcnts_[it];
      else
        sizes.push_back(recvcnts_[it]);
    }
    //
    // First, check whether we need to collapse this node.
    if(rank < extras) {
      if(rank < 0) {
        // This node needs to pre-send all it's data down.
        peers_.push_back(commstep(mpiid(comm_.rank().id-1), total_elements, 0));
      }
      else {
        peers_.push_back(commstep(mpiid(comm_.rank().id+1), 0, total_elements));
      }
    }
    //
    // Next, set up all the normal (power-of-two) data exchanges.
    if(rank >= 0) {
      int pivot = size;
      while(pivot > 1) {
        const int cap = pivot;
        pivot >>= 1;
        // Figure out where we're sending
        const int relrank = rank % cap;
        const int shift = rank - relrank;
        const int pal = partner(relrank, pivot) + shift;
        // Figure out what we're sending
        int lowframe[]  = { shift, pivot+shift };
        int highframe[] = { pivot+shift, cap+shift };
        int *recvframe = (relrank < pivot ? &lowframe[0]  : &highframe[0]);
        int *sendframe = (relrank < pivot ? &highframe[0] : &lowframe[0]);
        int sendamt = 0, recvamt = 0;
        for(int it = recvframe[0]; it < recvframe[1]; ++it)
          sendamt += sizes.at(it);
        for(int it = sendframe[0]; it < sendframe[1]; ++it)
          recvamt += sizes.at(it);
        // and add it to the queue of send/receive pairs.
        peers_.push_back(commstep(from_pof2(pal,extras), sendamt, recvamt));
      }
    }
    //
    // Finally, check whether we need to send back any data.
    if(rank < extras) {
      if(rank < 0) {
        // We need to get our data back from below.
        peers_.push_back(commstep(mpiid(comm_.rank().id-1), 
                                 0, recvcnts_.at(comm_.rank().id)));
      }
      else {
        peers_.push_back(commstep(mpiid(comm_.rank().id+1),
                                 recvcnts_.at(comm_.rank().id+1), 0));
      }
    }
    // and now we can go.
    this->sendrecv(now);
  }

} // end of namespace sstmac
