/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPILAUNCH_H
#define SSTMAC_PROCESS_MPILAUNCH_H

#include <sstmac/process/applaunch.h>

namespace sstmac {

  // Important types.
  class mpistrategy;
  class mpicommfactory;
  class scheduler;
  class mpiapp;

  /**
   * A launcher that runs a cooperatively scheduled group of threads
   */
  class mpilaunch : public applaunch {

    /// The mpi message strategy.
    boost::shared_ptr<mpistrategy> strategy_;

    /// The communicator factory shared by my tasks.
    boost::shared_ptr<mpicommfactory> factory_;

    /// The template object that gets cloned for each mpi peer instance.
    boost::shared_ptr<mpiapp> peer_template_;

    /// Should compute_api's get initialized?
    bool do_compute_;

    /// What is the target world size.
    int mpirun_np_;

  protected:
    /// Define the object that you want to run.
    /// This object gets cloned (using mpiapp::clone()) to arrive at
    /// the target number of tasks (as passed in to launcher::run).
    mpilaunch(const boost::shared_ptr<mpiapp>& peertype,bool use_compute_api,
              int mpirun_np);

  public:
    /// Build me an object.
    /// \param ptype            an object type that gets cloned on each node.
    /// \param use_compute_api  flag indicating whether to use the compute api
    /// \param mpirun_np        if > 0, limit MPI world size to mpirun_np nodes.
    static boost::shared_ptr<mpilaunch>
    construct(const boost::shared_ptr<mpiapp>& ptype, bool use_compute_api=false,
              int mpirun_np = 0);

    /// Goodbye.
    virtual ~mpilaunch() throw();

    /// What are we?
    const std::string& description() const;

    /// Access (with modification rights) the mpi message strategy.
    boost::shared_ptr<mpistrategy> get_mpistrategy();

    // Define the mpi message strategy.
    void set_mpistrategy(const boost::shared_ptr<mpistrategy>& strat);

    /// Access the communicator factory.
    boost::shared_ptr<mpicommfactory> get_mpicommfactory();

    /// Define the communicator factory type.
    void set_mpicommfactory(const boost::shared_ptr<mpicommfactory>& fac);

    /// Fire this object off.
    void run(const timestamp &current_time,
             const boost::shared_ptr<environment>& my_world);
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_MPILAUNCH_H
