/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPIMAGIC_H
#define SSTMAC_PROCESS_MPIMAGIC_H

#include <sstmac/process/mpicomm.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/util/nocopy.h>
#include <typeinfo>
#include <vector>
#include <map>
#include <set>

namespace sstmac {

  // Important types.
  class key;
  class payload;
  class scheduler;
  class mpicollpayload;
  
  /**
   * A magic short-circuit route for MPI communications.
   * Not yet capable of distributed computing.
   */
  class mpimagic : virtual public nocopy {
    /// A private nested type to keep track of barrier keys.
    struct barrierkey {
      mpitag ttag; mpicomm tcomm;
      barrierkey(mpitag tt, mpicomm cc);
      bool operator<(const barrierkey &other) const;
    };
    /// A private nested type to keep track of individual barrier calls.
    class barrier;
    /// The key used to block a given barrier call (this may have to change).
    std::map< barrierkey, boost::shared_ptr<barrier> > barrier_;

    /// A private key type for finding matches for collective ops.
    struct payloadkey {
      const std::type_info *tinfo; mpitag ttag; mpicomm tcomm;
      payloadkey(const std::type_info &inf, mpitag tt, mpicomm cc);
      bool operator<(const payloadkey &other) const;
    };
    /// A private nested type to keep track of payload for collective mpi ops.
    class payloaddata;
    /// Outstanding payload collections.
    typedef std::map< payloadkey, boost::shared_ptr<payloaddata> > payload_t;
    payload_t payload_;

  protected:
    /// Hello.
    mpimagic();

  public:
    /// Hi.
    static boost::shared_ptr<mpimagic>
    construct();

    /// Goodbye.
    virtual ~mpimagic() throw();

    /// Magic synchronization barrier.
    /// Callers are blocked until everybody has entered the barrier.
    void barrier(mpitag tag, mpicomm comm, const boost::shared_ptr<scheduler>& sched);

    /// Magic storage for payload in collective operations.
    /// Add payload for the current node.  Each node is only allowed
    /// to add its payload once.
    /// See also get_collective_payload(...) for retrieving payload data.
    /// \param caller   The C++ typeid for the collective implementation.
    /// \param tag      The mpi tag provided for the collective operation.
    /// \param comm     The communicator being used for the collective op.
    /// \param data     The payload being sent (shared) by this node.
    /// \param consumer Flag to track whether this node will be fetching
    ///                 payload or not (true for root in all-to-one calls
    ///                 and for all nodes in all-to-all calls).
    /// \throw valueerror if the caller/tag/comm combination has already
    ///                defined a collective payload.
    void add_collective_payload(const std::type_info &caller,
                                mpitag tag, mpicomm comm,
                                const boost::shared_ptr<const payload>& data,
                                bool consumer);

    /// Retrieve magically stored payload.
    /// As each node removes its payload, the data is removed from this
    /// container (so each node can only retrieve its payload once).
    /// \param caller  The C++ typeid for the collective implementation.
    /// \param tag     The mpi tag provided for the collective operation.
    /// \param comm    The communicator that was used for the collective op.
    /// \throw valueerror if the caller/tag/comm combination is not defined
    ///         for a collective payload.
    /// \throw illformederror if any node in the communicator has not provided
    ///          payload for the collective operation.
    boost::shared_ptr<const mpicollpayload>
    get_collective_payload(const std::type_info &caller,
                           mpitag tag, mpicomm comm);

    /// Test whether payload is in place for this node.
    /// \return true if the caller/tag/comm combination is recognized and
    ///              the comm.rank() has associated payload.
    bool have_collective_payload(const std::type_info &caller,
                                 mpitag tag, mpicomm comm) const;
    
    /// Test whether payload is in place for all nodes in the communicator.
    /// \return true if the caller/tag/comm combination is recognized and
    ///              all ranks in the communicator have associated payload.
    bool have_all_collective_payload(const std::type_info &caller,
                                     mpitag tag, mpicomm comm) const;
  };
  
} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_MPIMAGIC_H
