/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpimessage.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/payload.h>
#include <sstmac/util/alloc.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>
#include <vector>
#include <stdlib.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  // The mpimessage objects are very short-lived, so we don't feel like
  // continually allocating and deallocating those bastards.
  namespace messagepool {
    static std::vector<mpimessage*> *chunks = 0;

    static mpimessage* alloc_chunk() {
      if(! chunks) {
        chunks = new std::vector<mpimessage*>;
        chunks->reserve(1024);
      }
      if(! chunks->empty()) {
        mpimessage *memptr = chunks->back();
        chunks->pop_back();
        return memptr;
      }
      else {
        mpimessage *memptr = sstmac::allocate_aligned<mpimessage>();
        return memptr;
      }
    }

    static void free_chunk(mpimessage *chunk) {
      chunk->~mpimessage();
      chunks->push_back(chunk);
    }

  } // end of namespace messagepool.

  //
  // Hello.
  //
  mpimessage::mpimessage(int64_t envelope, int64_t mintrans,
                         int count, const mpitype &type, const mpiid &source,
                         const mpiid &dest, const mpitag &tag,
                         const mpicommid &commid, int seqnum,
                         const messageid &msgid, bool ssend, category cat,
                         const taskid &source_task, const taskid &dest_task,
                         const shared_ptr<const payload>& content) :
    envelope_(envelope), mintrans_(mintrans),
    count_(count), type_(type), source_(source), dest_(dest), tag_(tag),
    commid_(commid), seqnum_(seqnum), msgid_(msgid), ssend_(ssend), cat_(cat),
    source_task_(source_task), dest_task_(dest_task), content_(content)
  {
    if(envelope_ < 0) envelope_ = 0;
    if(mintrans_ < 0) mintrans_ = 0;
  }

  //
  // Construct a new message.
  //
  shared_ptr<mpimessage>
  mpimessage::construct(int64_t envelope, int64_t mintrans,
                        int count, const mpitype &type, const mpiid &source,
                        const mpiid &dest, const mpitag &tag,
                        const mpicommid &commid, int seqnum,
                        const messageid &msgid, bool ssend, category cat,
                        const taskid &source_task, const taskid &dest_task,
                        const shared_ptr<const payload>& content)
  {
    if(debugger::active())
      debug() << "mpimessage::construct(" << envelope << ", " << mintrans
              << ", " << count << ", " << type << ", "
              << source << ", " << dest << ", " << tag << ", " << commid
              << ", " << seqnum << ", " << ssend << ", " << cat
              << ", " << source_task << ", " << dest_task 
              << ", " << content << ")\n";
    shared_ptr<mpimessage> retval
      (new (messagepool::alloc_chunk())
       mpimessage(envelope, mintrans, count, type, source, dest, tag, commid,
                  seqnum, msgid, ssend, cat, source_task, dest_task, content),
       messagepool::free_chunk);
    return retval;
  }

  //
  // Goodbye.
  //
  mpimessage::~mpimessage() throw() {
  }

  //
  // Parent method to get the complemetary part of this message.
  //
  shared_ptr<networkdata> mpimessage::complement() const {
    shared_ptr<mpimessage> other
      (new mpimessage(envelope_, mintrans_, count_, type_, source_,
                      dest_, tag_, commid_, seqnum_, msgid_, ssend_,
                      cat_, source_task_, dest_task_, content_));
    return other;
  }

  //
  // Create a status object.
  //
  shared_ptr<mpistatus> mpimessage::build_status() const {
    shared_ptr<mpistatus> stat = mpistatus::construct();
    stat->set_commid(this->commid());
    stat->set_source(this->source());
    stat->set_dest(this->dest());
    stat->set_tag(this->tag());
    stat->set_send_type(this->type());
    stat->set_content(this->content());
    return stat;
  }

} // end of namespace sstmac
