/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpiqueue-recvrequest.h>
#include <sstmac/process/mpirecv.h>
#include <sstmac/process/mpimessage.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Hello.
  //
  mpiqueue::recvrequest::
  recvrequest(const shared_ptr<mpiqueue>& queue, int count, const mpitype &type,
              const mpiid &source, const mpitag &tag, const mpicomm &comm,
              mpimessage::category cat, const shared_ptr<mpirecv>& completion) :
    queue_(queue), acked_(false), source_(source), tag_(tag), comm_(comm),
    cat_(cat), count_(count), type_(type), completion_(completion)
  {}

  //
  // Who am I?
  //
  void mpiqueue::recvrequest::set_self(const shared_ptr<recvrequest>& self) {
    if(self.get() != this) {
      throw valueerror("mpiqueue::recvrequest:  'self' pointer does not point "
                       "to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<mpiqueue::recvrequest> mpiqueue::recvrequest::get_self() {
    return self_.lock();
  }

  //
  // Hello.
  //
  shared_ptr<mpiqueue::recvrequest>
  mpiqueue::recvrequest::construct(const shared_ptr<mpiqueue>& queue, int count,
                                   const mpitype &type, const mpiid &source,
                                   const mpitag &tag, const mpicomm &comm,
                                   mpimessage::category cat,
                                   const shared_ptr<mpirecv>& completion)
  {
    shared_ptr<recvrequest> retval(new recvrequest(queue, count, type, source,
                                                   tag, comm, cat, completion));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpiqueue::recvrequest::~recvrequest() {
  }

  //
  // Do we match the given message?
  //
//   bool mpiqueue::recvrequest::matches(boost::shared_ptr<mpimessage> msg) const {
//     bool fit = ((cat_ == msg->cat()) && (comm_.id() == msg->commid()) &&
//                 source_.matches(msg->source()) && tag_.matches(msg->tag()));
//     if(fit) {
//       // Just a little error checking.
//       if(type_ != msg->type() || count_ != abs(msg->count())) {
//         if(debugger::active()) debug() << "WARNING:  mpiqueue::recvrequest::match:  Matching message "
//                 << "but not matching type and count.  Expected " << count_
//                 << " elements of type " << type_ << " but received " 
//                 << msg->count() << " elements of type " << msg->type() << "\n";
//       }
//     }
//     return fit;
//   }

  //
  // We be done.
  //
  void mpiqueue::recvrequest::handle(const timestamp &current_time,
                                     const boost::shared_ptr<mpimessage>& mess)
  {
    // Figure out whether this is a full message or just a handshake.
    if(mess->count() < 0) {
      // Ack that we received the handshake.
      queue_.lock()->send_ack(mess, self_.lock());
      acked_ = true;
      // We're done for now -- the queue will deliver the message later.
    }
    else {
      // We get here if this was a ready-send message or if
      // we already acknowledged the handshake and the message just arrived.
      if(mess->ssend()) {
        if(! acked_) {
          // We can ack, but we will not be receiving a response.
          queue_.lock()->send_ack(mess, shared_ptr<recvrequest>());
        }
      }
      completion_->set_message(mess);
      completion_->handle(current_time);
    }
  }

} // end of namespace sstmac
