/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpiqueue-sendrequest.h>
#include <sstmac/process/mpiserver.h>
#include <sstmac/process/mpisend.h>
#include <sstmac/process/mpimessage.h>
#include <sstmac/process/payload.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Hello.
  //
  mpiqueue::sendrequest::sendrequest(const shared_ptr<mpiqueue>& queue,
                                     int64_t handsize, int64_t envelope,
                                     int64_t mintrans, int count,
                                     const mpitype &type, const mpiid &dest,
                                     const mpitag& tag, const mpicomm &comm,
                                     bool rsend, bool ssend, 
                                     mpimessage::category cat, 
                                     int seqno, const messageid &id,
                                     const shared_ptr<mpisend>& completion,
                                     const shared_ptr<const payload>& content) :
    queue_(queue), handsize_(handsize), envelope_(envelope),
    mintrans_(mintrans),count_(count), type_(type),
    dest_(dest), tag_(tag), comm_(comm), rsend_(rsend), ssend_(ssend),
    cat_(cat), seqno_(seqno), id_(id), completion_(completion),
    content_(content), pending_ack_(false), data_sent_(false)
  {}

  //
  // Who am I?
  //
  void mpiqueue::sendrequest::set_self(const shared_ptr<sendrequest>& self) {
    if(self.get() != this) {
      throw valueerror("mpiqueue::sendrequest::set_self:  'self' pointer does "
                       "not point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<mpiqueue::sendrequest> mpiqueue::sendrequest::get_self() {
    return self_.lock();
  }

  //
  // Hi there.
  //
  shared_ptr<mpiqueue::sendrequest>
  mpiqueue::sendrequest::construct(const boost::shared_ptr<mpiqueue>& queue,
                                   int64_t handsize, int64_t envelope,
                                   int64_t mintrans, int count,
                                   const mpitype &type, const mpiid &dest,
                                   const mpitag& tag, const mpicomm &comm,
                                   bool rsend, bool ssend, 
                                   mpimessage::category message_type,
                                   int seqno, const messageid &id,
                                   const shared_ptr<mpisend>& completion,
                                   const shared_ptr<const payload>& content)
  {
    if(debugger::active())
      debug() << "mpiqueue::sendrequest::construct("
              << queue << ", " << handsize << ", " 
              << mintrans << ", " << envelope << ", "
              << count << ", " << type
              << ", " << dest << ", " << tag << ", " << comm << ", " << rsend
              << ", " << ssend << ", " << message_type << ", " << seqno
              << ", " << completion << ", " << content << ")\n";
    shared_ptr<sendrequest> retval
      (new sendrequest(queue, handsize, envelope, mintrans, count, type, dest,
                       tag, comm, rsend, ssend, message_type, seqno, id,
                       completion, content));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpiqueue::sendrequest::~sendrequest() throw() {
  }

  //
  // Go.
  //
  void mpiqueue::sendrequest::start(const boost::shared_ptr<mpiserver>& lib) {
    if(debugger::active())
      debug() << "mpiqueue::sendrequest[" << this<< "]::start(" << lib << ")\n";
    if(lib == 0) {
      throw nullerror("mpiqueue::sendrequest::start:  null mpiserver pointer.");
    }
    shared_ptr<mpimessage> mess;
    // Decide whether we are sending a full payload or a handshake.
    int bytes = count_ * type_.size;
    if(debugger::active()) debug() << "rsend_==" << rsend_ << ", bytes==" << bytes
            << ", handsize_==" << handsize_ << "\n";
    if(rsend_ || (bytes <= handsize_)) {
      if(debugger::active())
        debug() << "mpiqueue::sendrequest::start:  Ready-send.\n";
      // Fire the whole sucker off.
      const taskid pt = comm_.peer_task(dest_);
      mess = mpimessage::construct(envelope_, mintrans_, count_, type_,
                                   comm_.rank(), dest_, tag_,
                                   comm_.id(), seqno_, id_, ssend_, cat_,
                                   comm_.my_task(), pt, content_);
      data_sent_ = true;
      // If this is not an ssend operation, send the completion to our listener.
      if(! ssend_) {
        lib->send(mess, completion_);
      }
      else {
        //  small ssend operation - send data but don't signal completion
        // until ack comes back.
        if(debugger::active())
          debug() << "mpiqueue::sendrequest[" << this << "]::start:  "
                  << "Sending small ssend message (ignoring completion).\n";
        pending_ack_ = true;
        queue_.lock()->request_ack(mess, this->get_self());
        if(debugger::active()) debug() << "mpiqueue::sendrequest::start:  IGNOREDEVENT 1\n";
        lib->send(mess, eventhandler::ignore);
      }
    }
    else {
      if(debugger::active())
        debug() << "mpiqueue::sendrequest[" << this << "]::start:  "
                << "Initiating handshake (ignoring send completion).\n";
      mess = mpimessage::construct(envelope_, mintrans_, -count_, type_,
                                   comm_.rank(), dest_, tag_,
                                   comm_.id(), seqno_, id_, ssend_, cat_,
                                   comm_.my_task(), comm_.peer_task(dest_),
                                   content_);
      // Request to be notified of a (future) incoming ack.
      pending_ack_ = true;
      queue_.lock()->request_ack(mess, this->get_self());
      // and fire off a handshake packet (negative size indicates a handshake).
        if(debugger::active()) debug() << "mpiqueue::sendrequest::start:  IGNOREDEVENT 2\n";
      lib->send(mess, eventhandler::ignore);
    }
  }

  //
  // We just received an ack.
  //
  void mpiqueue::sendrequest::ack(const timestamp &current_time,
                                  const boost::shared_ptr<mpimessage>& response,
                                  const boost::shared_ptr<mpiserver>& lib)
  {
    if(debugger::active())
      debug() << "mpiqueue::sendrequest[" << this << "]::ack("
              << current_time << ", " << response << ", " << lib << ")\n";
    if(! pending_ack_) {
      throw illformederror("mpiqueue::sendrequest::ack:  This request "
                           "is not waiting for an ack.");
    }
    if(response->seqnum() != seqno_) {
      throw illformederror("mpiqueue::sendrequest::ack:  Received a "
                           "non-matching ack.");
    }
    if(! data_sent_) {
      // Package up the rest of our stuff and send it off.
      // The receiver for the payload has already signaled that it's present,
      // so we can route the completion for the send directly to our listener.
      if(debugger::active())
        debug() << "mpiqueue::sendrequest[" << this << "]::ack:  "
                << "Sending data.  Completion routed to " << completion_<< "\n";
      shared_ptr<mpimessage> mess =
        mpimessage::construct(envelope_, mintrans_, count_, type_,
                              comm_.rank(), dest_, tag_,
                              comm_.id(), seqno_, id_, false,
                              mpimessage::data, comm_.my_task(),
                              comm_.peer_task(dest_), content_);
      lib->send(mess, completion_);
    }
    else {
      // We got our ack for an ssend operation.  Our work here is done.
      if(debugger::active())
        debug() << "mpiqueue::sendrequest[" << this << "]::ack:  "
                << "Ack for ssend operation.  Signaling completion\n";
      completion_->handle(current_time);
    }
  }

} // end of namespace sstmac.
