/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpiqueue.h>
#include <sstmac/process/mpiqueue-recvrequest.h>
#include <sstmac/process/mpiqueue-sendrequest.h>
#include <sstmac/process/mpiqueue-proberequest.h>
#include <sstmac/process/mpistatus.h>
#include <sstmac/process/payload.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/simulator/networkinterface.h>
#include <sstmac/process/mpiserver.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>
#include <stdint.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  inline void debug_message(const shared_ptr<mpimessage> &message) {
    if(debugger::active())
      debug() << message << ": bytes=" 
              << message->bytes() << " source_task="
              << message->source_task() << ", dest_task="
              << message->dest_task() << ", count="
              << message->count() << ", type=" << message->type() 
              << ", source=" << message->source() << ", dest="
              << message->dest() << ", tag=" << message->tag() << ", commid="
              << message->commid() << ", seqnum=" << message->seqnum()
              << ", id=" << message->id() << ", ssend=" << message->ssend()
              << ", cat=" << message->cat() << "\n";
  }

  //
  // Hi there.
  //
  mpiqueue::mpiqueue(const shared_ptr<mpiserver>& server,
                     const shared_ptr<nodeinterface>& node,
                     int64_t handsize,
                     int64_t envelope,
                     int64_t mintrans) :
    server_(server), node_(node), handsize_(handsize),
    envelope_(envelope), mintrans_(mintrans), next_id_(0)
  {
    if(server == 0) {
      throw nullerror("mpiqueue(...):  null server pointer.");
    }
    if(node == 0) {
      throw nullerror("mpiqueue(...):  null node pointer.");
    }
    // need to debug dangling messages.
    next_id_ = messageid(node->get_taskid().id * int64_t(4000000000LL));
  }

  //
  // Who am I?
  //
  void mpiqueue::set_self(const boost::shared_ptr<mpiqueue>& self) {
    if(self.get() != this) {
      throw valueerror("mpiqueue::set_self:  'self' pointer does not "
                       "point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
//   shared_ptr<mpiqueue> mpiqueue::get_self() {
//     return self_.lock();
//   }

  //
  // Hello.
  //
  shared_ptr<mpiqueue> mpiqueue::construct(const shared_ptr<mpiserver>& server,
                                           const shared_ptr<nodeinterface>& node,
                                           int64_t handsize,
                                           int64_t envelope,
                                           int64_t mintrans)
  {
    shared_ptr<mpiqueue> retval
      (new mpiqueue(server, node, handsize, envelope, mintrans));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpiqueue::~mpiqueue() throw() {
  }

  //
  // Send data.
  //
  void mpiqueue::send(int count, const mpitype &type, const mpiid &dest,
                      const mpitag &tag, const mpicomm &comm,
                      bool rsend, bool ssend,
                      mpimessage::category message_type,
                      const shared_ptr<mpisend>& completion,
                      const shared_ptr<const payload>& content)
  {
    if(debugger::active()) debug() << "mpiqueue[" << this << "]::send(" << count << ", " << type
            << ", " << dest << ", " << tag << ", " << comm << ", " << rsend
            << ", " << ssend <<", " << message_type <<", " << completion<<")\n";
    if(completion == 0) {
      throw nullerror("mpiqueue::send:  Null completion event pointer.");
    }
    taskid tid = comm.peers().at(dest.id);
    shared_ptr<sendrequest> req =
      sendrequest::construct(this->get_self(), handsize_, envelope_,
                             mintrans_, count, type, dest, 
                             tag, comm, rsend, ssend, message_type,
                             next_outbound_[tid]++, next_id_++, completion,
                             content);
    if(server_ == 0) {
      throw illformederror("mpiqueue::send:  null server pointer.");
    }
    req->start(server_);
  }

  //
  // Receive data.
  //
  void mpiqueue::recv(int count, const mpitype &type, const mpiid &source,
                      const mpitag &tag, const mpicomm &comm,
                      mpimessage::category message_type,
                      const shared_ptr<mpirecv>& completion)
  {
    if(debugger::active()) debug() << "mpiqueue[" << this << "]::recv(" << count << ", " << type
            << ", " << source << ", " << tag << ", " << comm << ", "
            << message_type << ", " << completion << ")\n";
    if(completion == 0) {
      throw nullerror("mpiqueue::recv:  Null completion event pointer.");
    }
    shared_ptr<recvrequest> req =
      recvrequest::construct(this->get_self(), count, type, source, tag,
                             comm, message_type, completion);
    // Figure out whether we already have a matching mesage.
    need_recv_t::iterator it, end = need_recv_.end();
    for(it = need_recv_.begin(); it != end; ++it) {
      if(req->matches(*it)) {
        if(debugger::active())
          debug() << "mpiqueue::recv:  Matching message found.\n";
        // We're good to go.
        shared_ptr<mpimessage> mess = *it;
        need_recv_.erase(it);
        req->handle(node_->now(), mess);
        return;
      }
    }
    // If we get here, we still need to wait for the message.
    if(debugger::active())
      debug() << "mpiqueue::recv:  Matching message not found\n";
    need_message_.push_back(req);
  }

  //
  // Ask for a notification when a message with the given signature arrives.
  //
  void mpiqueue::probe(const mpicomm &comm, const mpiid &source,
                       const mpitag &tag,
                       const boost::shared_ptr<mpiprobe>& completion)
  {
    if(debugger::active()) debug() << "mpiqueue[" << this << "]::probe(" << comm << ", " << source << ", " << tag
            << ", " << completion << ")\n";
    if(completion == 0) {
      throw nullerror("mpiqueue::probe:  Null completion event pointer.");
    }
    shared_ptr<proberequest> req =
      proberequest::construct(comm, source, tag, completion);
    // Figure out whether we already have a matching message.
    need_recv_t::iterator it, end = need_recv_.end();
    for(it = need_recv_.begin(); it != end; ++it) {
      if((*it)->cat() == mpimessage::user && req->matches(*it)) {
        // We're good to go.
        req->complete(node_->now(), *it);
        return;
      }
    }
    // If we get here, we still need to wait for the message.
    probelist_.push_back(req);    
  }

  //  
  // Immediate-mode probe for a message with the given signature.
  //
  shared_ptr<mpistatus> mpiqueue::iprobe(const mpicomm &comm,
                                         const mpiid &source,
                                         const mpitag &tag)
  {
    if(debugger::active()) debug() << "mpiqueue[" << this << "]::iprobe(" << comm << ", " << source << ", "
            << tag << ")\n";
    shared_ptr<proberequest> req =
      proberequest::construct(comm, source, tag, shared_ptr<mpiprobe>());
    need_recv_t::iterator it, end = need_recv_.end();
    for(it = need_recv_.begin(); it != end; ++it) {
      if((*it)->cat() == mpimessage::user && req->matches(*it)) {
        // This is it.
        return (*it)->build_status();
      }
    }
    // If we got here, there was no match
    return shared_ptr<mpistatus>();
  }

  //
  // Cancel all pending operations associated with the given event handler.
  //
  void mpiqueue::cancel(const shared_ptr<eventhandler>& ev) {
    if(debugger::active()) debug() << "mpiqueue[" << this << "]::cancel(" << ev << ")\n";
    throw unimplementederror("mpiqueue::cancel:  Not implemented yet.");
  }

  //
  // Used when a receive operation sends an ack.
  //
  void mpiqueue::send_ack(const shared_ptr<mpimessage>& message,
                          const shared_ptr<recvrequest>& operation)
  {
    if(debugger::active()) debug() << "mpiqueue[" << this << "]::send_ack(" << message << ", " << operation << ")\n";
    debug_message(message);
    // Create a response message of size 0 with source and destination switched.
    // This will have to be cleaned up later.
    shared_ptr<mpimessage> ack = mpimessage::construct
      (envelope_, mintrans_, 0, message->type(), message->dest(),
       message->source(), message->tag(), message->commid(), message->seqnum(),
       message->id(), false, mpimessage::ack, message->dest_task(),
       message->source_task(), payload::null());
    // OK,
    if(operation != 0) {
      // If this is non-null, the sender expects a response to its ack.
      ack_sent_[message->id()] = operation;
    }
    if(debugger::active()) debug() << "mpiqueue::send_ack:  Using ignoredevent\n";
    server_->send(ack, eventhandler::ignore);
  }

  //
  // Used when a send operation requests an ack for the given message.
  //
  void mpiqueue::request_ack(const boost::shared_ptr<mpimessage>& message,
                             const boost::shared_ptr<sendrequest>& operation)
  {
    if(debugger::active())
      debug() << "mpiqueue[" << this << "]::request_ack("
              << message << ", " << operation <<")\n"
              << "  :  Message id: " << message->id() << "\n";
    debug_message(message);
    ack_needed_[message->id()] = operation;
  }

  //
  // CALLBACK used by mpiserver when this object has a message.
  //
  void mpiqueue::incoming_message(const timestamp &current_time,
                                  const boost::shared_ptr<mpimessage>& message)
  {
    static bool first_warning = true;
    if(debugger::active())
      debug() << "mpiqueue[" << this << "]::incoming_message("
              << current_time << ", " << message << ")\n";
    //debug_message(message);
    // Acks are handled specially.
    if(message->cat() == mpimessage::ack) {
      // Incoming ack.
      const messageid &id = message->id();
      if(debugger::active())
        debug() << "  :  Incoming ack from task " << message->source_task()
                << " to task " << message->dest_task()
                << " with id " << id << "\n";
      if(debugger::active())
        debug() << " : Finding ack for " << message->id() << "\n";
      ack_needed_t::iterator it = ack_needed_.find(message->id());
      if(it != ack_needed_.end()) {
        // We have a winner.
        if(debugger::active())
          debug() << "mpiqueue[" << this << "]::incoming_message:  "
                  << "Ack with matching request.\n";
        shared_ptr<sendrequest> req = it->second;
        ack_needed_.erase(it);
        req->ack(current_time, message, server_);
      }
      else {
        // No match for this ack (this should not happen -- should be an error).
        if(debugger::active())
          debug() << "mpiqueue[" << this << "]::incoming_message:  ERROR:  "
                  << "Ack with no match.\n";
        throw illformederror("pmiqueue::incoming_message:  Ack with no match.");
      }
      return;
    }
    else if(message->cat() == mpimessage::data) {
      // Completion of an earlier handshake -- we better have a receiver ready.
      if(debugger::active()) debug() << "mpiqueue[" << this << "]::incoming_message:  Completion of a message.\n";
      ack_sent_t::iterator it = ack_sent_.find(message->id());
      if(it == ack_sent_.end()) {
        // Uh-oh.
        if(debugger::active()) debug() << "Fatal error on message.\n";
        debug_message(message);
        throw illformederror("mpiqueue::incoming_message:  Data message "
                             "without a matching ack.");
      }
      shared_ptr<recvrequest> recver = it->second;
      ack_sent_.erase(it);
      recver->handle(current_time, message);
      return;
    }
    else {
      // All other messages are tested to find a pending receiver.
      // FIXME:  Handle rearranging out-of-order messages.
      const taskid &tid = message->source_task();
      if(message->seqnum() != next_inbound_[tid]) {
        if(first_warning) {
          first_warning = false;
          std::cerr << "WARNING:  mpiqueue[" << this << "]::incoming_message:  "
                    << "Messages arriving out-of-order, but message "
                    << "reordering is not implemented yet.  Received message "
                    << message->seqnum() << " when waiting for "
                    << next_inbound_[tid] << "\n";
        }
      }
      next_inbound_[tid] = message->seqnum()+1;
      // OK, we can continue.
      bool found_match = false;
      need_message_t::iterator it, end = need_message_.end();
      for(it = need_message_.begin(); it != end; ++it) {
        if((*it)->matches(message)) {
          if(debugger::active()) debug() << "mpiqueue[" << this << "]::incoming_message:  Receiver is ready.\n";
          found_match = true;
          shared_ptr<recvrequest> req = *it;
          need_message_.erase(it);
          req->handle(current_time, message);
          return;
        }
      }
      // We get here if no match was found.
      // Messages that don't have a respondent are added to the list
      if(debugger::active()) debug() << "mpiqueue[" << this << "]::incoming_message:  Receiver not ready.\n";
      need_recv_.push_back(message);
      // And, finally, mpimessage::user messages are tested for outstanding probes
      probelist_t::iterator pit, pend = probelist_.end();
      for(pit = probelist_.begin(); pit != pend; ++pit) {
        if((*pit)->matches(message)) {
          // Match.
          shared_ptr<proberequest> preq = *pit;
          probelist_.erase(pit);
          preq->complete(current_time, message);
        }
      }
    }
  }

} // end of namespace sstmac
