/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/mpirecvengine.h>
#include <sstmac/process/mpiqueue.h>
#include <sstmac/util/alloc.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <vector>
#include <stdlib.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  // Yet another class of short-lived objects.  Let's reduce the load
  // on the memory manager.
  namespace mrpool {
    static std::vector<mpirecvengine*> *chunks = 0;

    static mpirecvengine* alloc_chunk() {
      if(! chunks) {
        chunks = new std::vector<mpirecvengine*>;
        chunks->reserve(1024);
      }
      if(! chunks->empty()) {
        mpirecvengine *memptr = chunks->back();
        chunks->pop_back();
        return memptr;
      }
      else {
        mpirecvengine *memptr = sstmac::allocate_aligned<mpirecvengine>();
        return memptr;
      }
    }

    static void free_chunk(mpirecvengine *chunk) {
      chunk->~mpirecvengine();
      chunks->push_back(chunk);
    }

  } // end of namespace mrpool.

  //
  // Constructor.
  //
  mpirecvengine::mpirecvengine(const boost::shared_ptr<key>& thekey,
                               const shared_ptr<mpiqueue>& queue,
                               int count, const mpitype &type,
                               const mpiid &source, const mpitag &tag,
                               const mpicomm &comm, mpimessage::category cat) :
    mpikernel(thekey),
    queue_(queue), count_(count), type_(type), source_(source), tag_(tag),
    comm_(comm), cat_(cat)
  {
    if(debugger::active())
      debug() << "mpirecvengine[" << this << "]::mpirecvengine(" << thekey
              << ", " << queue << ", " << count << ", " << type << ", "
              << source << ", " << tag << ", " << comm << ", " << cat << ")\n";
  }

  //
  // Who am I?
  //
  void mpirecvengine::set_self(const shared_ptr<mpirecvengine>& self) {
    if(self.get() != this) {
      throw valueerror("mpirecvengine::set_self:  'self' pointer does not "
                       "point to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  shared_ptr<mpirecvengine> mpirecvengine::get_self() {
    return self_.lock();
  }

  //
  // Build me a sharable object.
  //
  shared_ptr<mpirecvengine>
  mpirecvengine::construct(const boost::shared_ptr<key>& thekey,
                           const shared_ptr<mpiqueue>& queue, int count,
                           const mpitype &type, const mpiid &source,
                           const mpitag &tag, const mpicomm &comm,
                           mpimessage::category cat)
  {
    shared_ptr<mpirecvengine> retval
      (new(mrpool::alloc_chunk()) mpirecvengine(thekey, queue, count, type,
                                                source, tag, comm, cat),
       mrpool::free_chunk);
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpirecvengine::~mpirecvengine() throw() {
  }

  //
  // This is the method from the kernel.
  //
  void mpirecvengine::start(const timestamp & start_time,
                            const boost::shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpirecvengine::start(" << start_time << ", "
              << my_node << ")\n";
    queue_->recv(count_, type_, source_, tag_, comm_, cat_, this->get_self());
  }

  //
  // Grab the completion message.
  //
  void mpirecvengine::handle(const timestamp &current_time) {
    if(debugger::active())
      debug() << "mpisendengine[" << this 
              << "]::handle(" << current_time << ")\n";
    mpikernel::complete(current_time, mpirecv::get_message()->build_status());
  }

} // end of namespace sstmac
