/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPISTRATEGIES_H
#define SSTMAC_PROCESS_MPISTRATEGIES_H

#include <sstmac/simulator/timestamp.h>
#include <sstmac/process/mpicomm.h>
#include <sstmac/process/mpitag.h>
#include <sstmac/process/mpiid.h>
#include <sstmac/process/mpitype.h>
#include <sstmac/process/mpiop.h>
#include <sstmac/process/mpicollpayload.h>
#include <boost/shared_ptr.hpp>
#include <vector>
#include <stdint.h>

namespace sstmac {

  // Forward declaration of important types:
  class mpiqueue;
  class mpikernel;
  class mpimagic;
  class key;

  //
  /// Base strategy for sending messages.
  //
  class mpisendstrategy {
  public:
    /// Virtual destructor is critical here.
    virtual ~mpisendstrategy() throw() {}
    /// Post a send request to the given node and associate the
    /// completion of this request with the given request handle.
    /// All sends are non-blocking; if return receipts are required the
    /// completion of the request must be associated with the arrival
    /// of the return receipt.
    ///
    /// Like all strategy methods, this method must be reentrant.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          int count, const mpitype &type, const mpiid &dest,
          const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  /// Base strategy for receiving messages.
  //
  class mpirecvstrategy {
  public:
    /// Must be virtual
    virtual ~mpirecvstrategy() throw() {}
    /// Post a receive request from the given node to be matched by the
    /// given request handle.  All receives are non-blocking; it is up
    /// to the strategy to handle return receipts or other synchronization.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          int count, const mpitype &type, const mpiid &source,
          const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  /// Base strategy for broadcast operations.
  //
  class mpibcaststrategy {
  public:
    /// Must be virtual
    virtual ~mpibcaststrategy() throw() {}
    /// Broadcast from root to all the mpiids in the given list.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          int count, const mpitype &type, const mpiid &root,
          const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
      };

  //
  /// Base strategy for reduce operations.
  //
  class mpireducestrategy {
  public:
    /// Must be virtual.
    virtual ~mpireducestrategy() throw() {}
    /// Reduce between root mpiid and all other participants.
    /// Associates the completion of the reduction operation with the
    /// given request handle.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          int count, const mpitype &type, const mpiop &op,
          const mpiid &root, const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  /// Base strategy for allreduce operations.
  //
  class mpiallreducestrategy {
  public:
    /// Must be virtual.
    virtual ~mpiallreducestrategy() throw() {}
    /// Reduce between all nodes.
    /// Associates the completion of the reduction operation with the
    /// given request handle.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          int count, const mpitype &type, const mpiop &op,
          const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  // Base strategy for gather operations.
  //
  class mpigatherstrategy {
  public:
    /// Must be virtual.
    virtual ~mpigatherstrategy() throw() {}
    /// Perform a gather operation
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey, const boost::shared_ptr<mpiqueue>& queue,
          int sendcount, const mpitype &sendtype,
          int recvcount, const mpitype &recvtype,
          const mpiid &root, const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  /// Base strategy for gatherv operations.
  //
  class mpigathervstrategy {
  public:
    /// Must be virtual.
    virtual ~mpigathervstrategy() throw() {}
    /// Gather data to root node from all other participants.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey, const boost::shared_ptr<mpiqueue>& queue,
          int sendcount, const mpitype &sendtype,
          const std::vector<int> &recvcounts, const mpitype &recvtype, 
          const mpiid &root, const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  // Base strategy for allgather operations.
  //
  class mpiallgatherstrategy {
  public:
    /// Must be virtual.
    virtual ~mpiallgatherstrategy() throw() {}
    /// Gather data to all nodes from all participants.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey, const boost::shared_ptr<mpiqueue>& queue,
          int sendcount, const mpitype &sendtype,
          int recvcount, const mpitype &recvtype,
          const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  // Base strategy for allgatherv operations.
  //
  class mpiallgathervstrategy {
  public:
    /// Must be virtual.
    virtual ~mpiallgathervstrategy() throw() {}
    /// Gather data to all nodes from all participants.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey, const boost::shared_ptr<mpiqueue>& queue,
          int sendcount, const mpitype &sendtype,
          const std::vector<int> &recvcount, const mpitype &recvtype,
          const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  // Base strategy for all-to-all communication with fixed data lengths.
  //
  class mpialltoallstrategy {
  public:
    /// Must be virtual.
    virtual ~mpialltoallstrategy() throw() {}
    /// Gather data from all nodes to all nodes.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          int sendcnt, const mpitype &sendtype,
          int recvcnt, const mpitype &recvtype,
          const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  // Base strategy for all-to-all communication with variable data lengths.
  //
  class mpialltoallvstrategy {
  public:
    /// Must be virtual.
    virtual ~mpialltoallvstrategy() throw() {}
    /// Gather data from all nodes to all nodes.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          const std::vector<int> &sendcnts, const mpitype &sendtype,
          const std::vector<int> &recvcnts, const mpitype &recvtype,
          const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  // Base strategy for scatter operations.
  //
  class mpiscatterstrategy {
  public:
    /// Must be virtual.
    virtual ~mpiscatterstrategy() throw() {}
    /// Perform a scatter operation
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          int sendcount, const mpitype &sendtype,
          int recvcount, const mpitype &recvtype,
          const mpiid &root, const mpitag &tag,
          const mpicomm &comm,
          const boost::shared_ptr<const mpicollpayload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  // Base strategy for scatterv operations.
  //
  class mpiscattervstrategy {
  public:
    /// Must be virtual.
    virtual ~mpiscattervstrategy() throw() {}
    /// Scatter data from root node to all other participants.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          const std::vector<int> &sendcounts, const mpitype &sendtype,
          int recvcount, const mpitype &recvtype, 
          const mpiid &root, const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const mpicollpayload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  // Base strategy for fused reduce-scatter operations.
  //
  class mpireduce_scatterstrategy {
  public:
    /// Must be virtual.
    virtual ~mpireduce_scatterstrategy() throw() {}
    /// Perform a fused reduce-scatter operation
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          const std::vector<int> &recvcnts, const mpitype &type,
          const mpiop &op, const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  class mpibarrierstrategy {
  public:
    /// Must be virtual.
    virtual ~mpibarrierstrategy() throw() {}
    /// Synchronize nodes.
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

  //
  // Base strategy for scan operations.
  //
  class mpiscanstrategy {
  public:
    /// Must be virtual.
    virtual ~mpiscanstrategy() throw() {}
    /// Build a kernel to perform a scan operation
    virtual boost::shared_ptr<mpikernel>
    build(const boost::shared_ptr<key>& thekey,
          const boost::shared_ptr<mpiqueue>& queue,
          int count, const mpitype &type, const mpiop &op,
          const mpitag &tag, const mpicomm &comm,
          const boost::shared_ptr<const payload>& content,
          const boost::shared_ptr<mpimagic>& magic) const = 0;
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_MPISTRATEGIES_H
