/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_MPISTRATEGY_H
#define SSTMAC_PROCESS_MPISTRATEGY_H

#include <sstmac/process/mpistrategies.h>
#include <boost/shared_ptr.hpp>

namespace sstmac {

  /**
   * A full ensemble of strategies for an MPI communication environment.
   * All the individual strategies have (hopefully reasonable) defaults.
   *
   * All operations are preseumed to be non-blocking.  The choice of whether
   * (or when) to block is left to the caller.
   */
  class mpistrategy {
  private:
    /// The strategy objects for non-blocking sends.
    boost::shared_ptr<mpisendstrategy> send_, bsend_, ssend_, rsend_;
    /// The strategy object for blocking receives.
    boost::shared_ptr<mpirecvstrategy> recv_;
    /// The strategy object for broadcasts.
    boost::shared_ptr<mpibcaststrategy> bcast_;
    /// The strategy object for reduce operations.
    boost::shared_ptr<mpireducestrategy> reduce_;
    /// The strategy object for allreduce.  Should this be fused with reduce?
    boost::shared_ptr<mpiallreducestrategy> allreduce_;
    /// The strategy object for gather operations.
    boost::shared_ptr<mpigatherstrategy> gather_;
    /// The strategy object for gatherv operations.
    boost::shared_ptr<mpigathervstrategy> gatherv_;
    /// The strategy object for allgather operations.
    boost::shared_ptr<mpiallgatherstrategy> allgather_;
    /// The strategy object for allgatherv operations.
    boost::shared_ptr<mpiallgathervstrategy> allgatherv_;
    /// The strategy object for alltoall operations.
    boost::shared_ptr<mpialltoallstrategy> alltoall_;
    /// The strategy object for alltoallv operations.
    boost::shared_ptr<mpialltoallvstrategy> alltoallv_;
    /// The strategy object for scatter operations.
    boost::shared_ptr<mpiscatterstrategy> scatter_;
    /// The strategy object for scatterv operations.
    boost::shared_ptr<mpiscattervstrategy> scatterv_;
    /// The strategy object for fused reduce_scatter operations.
    boost::shared_ptr<mpireduce_scatterstrategy> reduce_scatter_;
    // The strategy object for synchronization barriers.
    boost::shared_ptr<mpibarrierstrategy> barrier_;
    /// The strategy object for scan operations.
    boost::shared_ptr<mpiscanstrategy> scan_;

  protected:
    /// Default construction.  Builds default implementations for
    /// all the individual strategy types.
    mpistrategy();
    
  public:
    /// Build me an object.
    static boost::shared_ptr<mpistrategy> construct();

    /// The strategy objects for sends.
    /// The set_* methods throw mpistrategy_error if the strategy is NULL.
    boost::shared_ptr<mpisendstrategy> send() { return send_; }
    boost::shared_ptr<const mpisendstrategy> send() const { return send_; }
    void set_send(const boost::shared_ptr<mpisendstrategy>& st);
    boost::shared_ptr<mpisendstrategy> bsend() { return bsend_; }
    boost::shared_ptr<const mpisendstrategy> bsend() const { return bsend_; }
    void set_bsend(const boost::shared_ptr<mpisendstrategy>&);
    boost::shared_ptr<mpisendstrategy> ssend() { return ssend_; }
    boost::shared_ptr<const mpisendstrategy> ssend() const { return ssend_; }
    void set_ssend(const boost::shared_ptr<mpisendstrategy>& st);
    boost::shared_ptr<mpisendstrategy> rsend() { return rsend_; }
    boost::shared_ptr<const mpisendstrategy> rsend() const { return rsend_; }
    void set_rsend(const boost::shared_ptr<mpisendstrategy>& st);

    /// The strategy object for receives.
    boost::shared_ptr<mpirecvstrategy> recv() { return recv_; }
    boost::shared_ptr<const mpirecvstrategy> recv() const { return recv_; }
    void set_recv(const boost::shared_ptr<mpirecvstrategy>& st);

    /// The strategy object for broadcasts.
    boost::shared_ptr<mpibcaststrategy> bcast() { return bcast_; }
    boost::shared_ptr<const mpibcaststrategy> bcast() const { return bcast_; }
    void set_bcast(const boost::shared_ptr<mpibcaststrategy>& st);

    /// The strategy object for reduce operations.
    boost::shared_ptr<mpireducestrategy> reduce() { return reduce_; }
    boost::shared_ptr<const mpireducestrategy> reduce() const {
      return reduce_;
    }
    void set_reduce(const boost::shared_ptr<mpireducestrategy>& st);

    /// The strategy object for allreduce.
    boost::shared_ptr<mpiallreducestrategy> allreduce() { return allreduce_; }
    boost::shared_ptr<const mpiallreducestrategy> allreduce() const {
      return allreduce_;
    }
    void set_allreduce(const boost::shared_ptr<mpiallreducestrategy>& st);

    /// The strategy object for gather operations.
    boost::shared_ptr<mpigatherstrategy> gather() { return gather_; }
    boost::shared_ptr<const mpigatherstrategy> gather() const {return gather_;}
    void set_gather(const boost::shared_ptr<mpigatherstrategy>& st);

    /// The strategy object for gatherv operations.
    boost::shared_ptr<mpigathervstrategy> gatherv() { return gatherv_; }
    boost::shared_ptr<const mpigathervstrategy> gatherv() const {
      return gatherv_;
    }
    void set_gatherv(const boost::shared_ptr<mpigathervstrategy>& st);

    /// The strategy object for allgather operations.
    boost::shared_ptr<mpiallgatherstrategy> allgather() { return allgather_; }
    boost::shared_ptr<const mpiallgatherstrategy> allgather() const {
      return allgather_;
    }
    void set_allgather(const boost::shared_ptr<mpiallgatherstrategy>& st);

    /// The strategy object for allgatherv operations.
    boost::shared_ptr<mpiallgathervstrategy> allgatherv() {return allgatherv_;}
    boost::shared_ptr<const mpiallgathervstrategy> allgatherv() const {
      return allgatherv_;
    }
    void set_allgatherv(const boost::shared_ptr<mpiallgathervstrategy>& st);

    /// The strategy object for alltoall operations.
    boost::shared_ptr<mpialltoallstrategy> alltoall() { return alltoall_; }
    boost::shared_ptr<const mpialltoallstrategy> alltoall() const {
      return alltoall_;
    }
    void set_alltoall(const boost::shared_ptr<mpialltoallstrategy>& st);

    /// The strategy object for alltoallv operations.
    boost::shared_ptr<mpialltoallvstrategy> alltoallv() { return alltoallv_; }
    boost::shared_ptr<const mpialltoallvstrategy> alltoallv() const {
      return alltoallv_;
    }
    void set_alltoallv(const boost::shared_ptr<mpialltoallvstrategy>& st);

    /// The strategy object for scatter operations.
    boost::shared_ptr<mpiscatterstrategy> scatter() {
      return scatter_;
    }
    boost::shared_ptr<const mpiscatterstrategy> scatter() const {
      return scatter_;
    }
    void set_scatter(const boost::shared_ptr<mpiscatterstrategy>& st);

    /// The strategy object for scatterv operations.
    boost::shared_ptr<mpiscattervstrategy> scatterv() {
      return scatterv_;
    }
    boost::shared_ptr<const mpiscattervstrategy> scatterv() const {
      return scatterv_;
    }
    void set_scatterv(const boost::shared_ptr<mpiscattervstrategy>& st);

    /// The strategy object for fused reduce_scatter operations.
    boost::shared_ptr<mpireduce_scatterstrategy> reduce_scatter() {
      return reduce_scatter_;
    }
    boost::shared_ptr<const mpireduce_scatterstrategy> reduce_scatter() const {
      return reduce_scatter_;
    }
    void set_reduce_scatter(const boost::shared_ptr<mpireduce_scatterstrategy>& st);

    /// The strategy object for synchronization barriers.
    boost::shared_ptr<mpibarrierstrategy> barrier() { return barrier_; }
    boost::shared_ptr<const mpibarrierstrategy> barrier() const {
      return barrier_;
    }
    void set_barrier(const boost::shared_ptr<mpibarrierstrategy>& st);

    /// Thestrategy object for scan operations.
    boost::shared_ptr<mpiscanstrategy> scan() { return scan_; }
    boost::shared_ptr<const mpiscanstrategy> scan() const { return scan_; }
    void set_scan(const boost::shared_ptr<mpiscanstrategy>& st);
  };
} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_MPISTRATEGY_H
