/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_PAYLOADMAP_H
#define SSTMAC_PROCESS_PAYLOADMAP_H

#include <sstmac/process/payload.h>
#include <sstmac/util/errors.h>
#include <map>

namespace sstmac {

  /**
   * Keep a mapping from some form of sortable entities to message payloads.
   * Intended to make it easier to manage various collective messaging ops.
   */
  template <typename Key, typename Value>
  class payloadmap : public payload {
  public:
    /// The data storage type
    typedef typename std::map<Key, Value> map_t;

  private:
    /// The actual data map.
    map_t map_;

    /// Hello.
    payloadmap() {
    }
    
    /// Private copy constructor.
    payloadmap(const payloadmap<Key, Value> &other) :
      map_(other.map_)
    {}

  public:
    /// Get a new object.
    static typename boost::shared_ptr< payloadmap<Key, Value> >
    construct() {
      return typename
        boost::shared_ptr< payloadmap<Key, Value> >(new payloadmap);
    }

    /// Goodbye.
    virtual ~payloadmap() throw() {}

    /// Clone the map.
    boost::shared_ptr<payload> clone() const {
      return typename
        boost::shared_ptr< payloadmap<Key, Value> >(new payloadmap(*this));
    }

    /// Access the map.
    map_t& value() { return map_; }

    /// Access the map in a const context.
    const map_t& value() const { return map_; }

    /// Insert a value (convenience function).
    void insert(const Key &key, const Value &value,
                bool allow_redefinition = true)
    {
      if((! allow_redefinition) && (map_.find(key) != map_.end()))
        throw valueerror("payloadmap::insert: Redefining keys is disabled.");
      map_[key] = value;
    }

    /// Merge all results from another payload map into this one.
    void merge(const typename payloadmap<Key, Value> &other,
               bool allow_redefinition = true)
    {
      if(! other)
        throw nullerror("payloadmap::merge: Null payload map.");
      if(allow_redefinition) {
        map.insert(other->value().begin(), other->value().end());
      }
      else {
        map_t::const_iterator it, end = other->value().end();
        for(it = other->value().begin(); it != end; ++it) {
          if(map_.find(it->first) == map_.end())
            map_.insert(*it);
          else
            throw valueerror("payloadmap::merge: Redefining keys is disabled.");
        }
      }
    }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_PAYLOADMAP_H
