/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/processrequest.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/process/key.h>
#include <sstmac/util/alloc.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>
#include <vector>
#include <stdlib.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;

  // The processrequest objects are very short-lived, so let's not
  // continually allocate and dealloce those poor suckers.
  namespace prpool {
    static std::vector<processrequest*> *chunks = 0;

    static processrequest* alloc_chunk() {
      if(! chunks) {
        chunks = new std::vector<processrequest*>;
        chunks->reserve(1024);
      }
      if(! chunks->empty()) {
        processrequest *memptr = chunks->back();
        chunks->pop_back();
        return memptr;
      }
      else {
        processrequest *memptr = allocate_aligned<processrequest>();
        return memptr;
      }
    }

    static void free_chunk(processrequest *chunk) {
      chunk->~processrequest();
      chunks->push_back(chunk);
    }

  } // end of namespace prpool.

  //
  // Hello.
  //
  processrequest::processrequest(const shared_ptr<key>& thekey) :
    key_(thekey), final_time_(-1)
  {}

  //
  // Build me a request.
  //
  shared_ptr<processrequest> processrequest::construct(const shared_ptr<key>& thekey) {
    shared_ptr<processrequest> retval
      (new(prpool::alloc_chunk()) processrequest(thekey), prpool::free_chunk);
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  processrequest::~processrequest() throw() {
    if(! this->is_complete()) {
      std::cerr << "WARNING:  ~processrequest():  request still incomplete "
                << "(blocked on key " << key_ << ")\n";
    }
  }

  //
  // Block the caller thread until this request is complete.
  //
  const timestamp& processrequest::wait(const shared_ptr<scheduler>& sched) {
    if(sched == 0) {
      throw nullerror("processrequest::wait:  Null scheduler pointer.");
    }
    if(this->is_complete()) {
      throw illformederror("processrequest::wait:  Request is complete.");
    }
    sched_.insert(sched);
    if(debugger::active()) debug() << "processrequest[" << this << "] placing a block.\n";
    sched->block(key_);
    // We get here when we're unblocked.
    if(debugger::active())
      debug() << "processrequest[" << this << "]:  Block released\n";
    return final_time_;
  }

  //
  // Grab the completion request.
  //
  void processrequest::complete(const timestamp &current_time,
                                const shared_ptr<const status>& stat)
  {
    if(debugger::active()) 
      debug() << "processrequest[" << this << "]::complete(" 
              << current_time << ", " << stat << ")\n";
    request::complete(current_time, stat);
    final_time_ = current_time;
    if(! sched_.empty()) {
      schedulers_t::iterator it, end = sched_.end();
      for(schedulers_t::iterator it = sched_.begin(); it != end; ++it) {
        (*it)->unblock(key_);
      }
      sched_.clear();
    }
  }

} // end of namespace sstmac
