/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_SIMPLEMAILSERVER_MBOX_H
#define SSTMAC_PROCESS_SIMPLEMAILSERVER_MBOX_H

#include <sstmac/process/simplemailserver.h>
#include <sstmac/process/simplembox.h>
#include <utility>
#include <deque>
#include <list>

namespace sstmac {

  // Forward declaration of important types.
  class key;
  
  /**
   * A concrete implementation of a simplembox.
   */
  class simplemailserver::mbox : public simplembox {
  public:
    typedef simplembox::tagmap_t tagmap_t;

  private:
    /// The process identifier for this object.
    actorid id_;

    /// The parent object that "owns" this mbox.
    boost::shared_ptr<simplemailserver> parent_;

    /// Track whether we are in between init and finalize.
    bool active_;

    /// Blocked receive calls are held here.
    typedef std::pair< actorid, boost::shared_ptr<const actorpattern> > actorp_t;
    typedef std::pair< actorp_t, boost::shared_ptr<key> > pend_t;
    typedef std::list<pend_t> blocked_t;
    blocked_t blocked_;

    /// Un-handled received messages are held here.
    typedef std::pair< actorid, boost::shared_ptr<actormsg> > msgpair_t;
    typedef std::list<msgpair_t> unhandled_t;
    unhandled_t unhandled_;

    /// Constructor.
    mbox(const actorid &id, const boost::shared_ptr<simplemailserver>& parent);

  public:
    /// Get us a new mbox.
    static boost::shared_ptr<mbox>
    construct(const actorid &id,
              const boost::shared_ptr<simplemailserver>& parent);

    /// Done.
    virtual ~mbox();

    /// Allow the parent to send us a message to handle.
    void handle(const actorid &sender, const boost::shared_ptr<actormsg>& payload);

    /* Methods from here on down are part of the simplembox interface */

    /// What's my pid.
    virtual actorid mypid() const;

    /// What nodes are known to the current system?
    virtual const std::vector<nodeid>& nodes() const;

    /// All peer identifiers (magically, we know about everybody).
    virtual const std::set<actorid>& world() const;

    /// Register the given tag mapping in the registry.
    /// A key cannot be redefined unless it is first erased.
    virtual void set_tag(int key, actorid value);

    /// Test whether the given key is defined.
    virtual bool has_tag(int key) const;

    /// Access the value of the given key.
    virtual actorid get_tag(int key) const;

    /// Access the tag mapping in a const context.
    virtual const tagmap_t& tag_map() const;

    /// Create or update the given global counter.
    /// In this simple model, the global counter is stored by "cheating"
    /// (magic bypass).
    virtual void move_global_counter(const std::string &label,
                                     int increment);

    /// Read the local state of the given global counter.
    /// This would (in a real system) be assumed to be a purely
    /// local operation.
    /// \throw valueerror if the given global counter is not defined.
    virtual int read_global_counter(const std::string &label) const;

    /// Initialize.  This method is collective and "magic" -- it ensures
    /// that all nodes have joined the party before anybody starts.
    virtual timestamp init();

    /// Finalize.  This is a magic operation.
    virtual timestamp finalize();

    /// Spawn the given task on the given node and return its pid.
    /// This can only be called between 'init()' and 'finalize()'.
    virtual actorid spawn(nodeid target,
                        const boost::shared_ptr<simplemboxapp>& entity,
                        int64_t communication_size);

    /// Send a message to the given pid.
    /// This call always succeeds except for invalid pid.
    /// \throw valueerror if the pid is not valid.
    /// \throw nullerror if sending a NULL message (reserved for recv timeout).
    virtual void send(const actorid &target,
                      const boost::shared_ptr<const actormsg>& msg,
                      int64_t size);

    /// Test whether there is any message that matches the given pattern.
    /// This has the same effect as calling recv with a zero-length timeout.
    /// \param pat     The pattern we are matching (may be actorpattern::any).
    virtual bool test(const actorid &source,
                      const boost::shared_ptr<const actorpattern>& pat) const;
    
    /// Block until we get a message matching the given pattern.
    /// \param source  Our desired source (may be actorid::any).
    /// \param pat     The pattern we are matching (may be actorpattern::any).
    /// \return        The received object (sorry about the dynamic casts)
    ///                or std::make_pair(actorid::invalid, NULL) if timed out.
    virtual std::pair< actorid, boost::shared_ptr<actormsg> >
    recv(const actorid &source, const boost::shared_ptr<const actorpattern>& pat);

    /// Pretend to compute for the given amount of time.
    virtual void compute(timestamp cputime_per_thread, int threads = 1);

    /// Pretend to be idle (block) for the given amount of time.
    virtual void sleep(timestamp cputime);

    /// Pretend to copy the given number of bytes.
    virtual void copy(int64_t bytes);

    /// Get current time.
    virtual timestamp now() const;
  };

} // end of namespace sstmac

#endif // ! SSTMAC_PROCESS_SIMPLEMAILSERVER_MBOX_H
