/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/simplemailserver-mbox.h>
#include <sstmac/process/simplemailserver-data.h>
#include <sstmac/process/actormagic.h>
#include <sstmac/simulator/tasklist.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/simulator/networkinterface.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  using boost::shared_ptr;
  using boost::weak_ptr;
  using boost::dynamic_pointer_cast;

  //
  // Private method to send data.
  //
  void simplemailserver::send(const actorid &source, const actorid &target,
                              const boost::shared_ptr<const actormsg>& msg,
                              int64_t size)
  {
    if(! (source.is_real() && target.is_real()))
      throw valueerror("simplemailserver::send:  Not a valid source or target");
    if(valid_nodes_.find(target.node()) == valid_nodes_.end())
      throw valueerror("simplemailserver::send:  Not a valid target node.");
    if(! msg)
      throw nullerror("simplemailserver::send:  Unwilling to send null content");
    shared_ptr<data> dd = data::construct(source, target, size, msg->actorclone());
    if(target.node() != my_node_id_)
      net_->send(nodeif_->now(), target.node(), dd, eventhandler::ignore);
    else
      this->incoming_message(timestamp(0), my_node_id_, dd);
  }
  
  //
  // Private constructor.
  //
  simplemailserver::simplemailserver(nodeid nid,
                                     const shared_ptr<actormagic>& magic,
                                     const shared_ptr<tasklist>& tlist,
                                     const shared_ptr<scheduler>& sched) :
    sched_(sched), magic_(magic), my_node_id_(nid), next_child_(1)
  {
    for(taskid id = taskid(0); id < tlist->size(); ++id)
      valid_nodes_.insert(tlist->node(id));
    nodes_.reserve(valid_nodes_.size());
    nodes_.insert(nodes_.begin(), valid_nodes_.begin(), valid_nodes_.end());
    if(valid_nodes_.find(nid) == valid_nodes_.end())
      throw illformederror("simplemailserver:  'My' node id is not in the "
                           "list of valid nodes.");
  }

  //
  // Get us a new, heap-allocated object.
  //
  shared_ptr<simplemailserver>
  simplemailserver::construct(nodeid nid,
                              const shared_ptr<actormagic>& magic,
                              const shared_ptr<tasklist>& tlist,
                              const shared_ptr<scheduler>& sched)
  {
    shared_ptr<simplemailserver>
      rv(new simplemailserver(nid, magic, tlist, sched));
    rv->self_ = rv;
    return rv;
  }

  //
  // Clean up.
  //
  simplemailserver::~simplemailserver() {
  }

  //
  // Create a new mbox interface that can talk to this object.
  //
  shared_ptr<simplembox> simplemailserver::new_mbox() {
    actorid new_id(my_node_id_, next_child_++);
    shared_ptr<mbox> mb = mbox::construct(new_id, self_.lock());
    mboxes_[new_id] = mb;
    return mb;
  }

  /* Interface required by the 'server' base type. */

  //
  // Define a network interface for this server.
  //
  void simplemailserver::initialize(const boost::shared_ptr<nodeinterface>& my_node,
                                    const boost::shared_ptr<networkinterface>& my_net)
  {
    nodeif_ = my_node;
    net_ = my_net;
  }

  //
  // Complete a receive.
  //
  void simplemailserver::incoming_message(const timestamp&,
                                          const nodeid&,
                                          const boost::shared_ptr<networkdata>& load)
  {
    shared_ptr<data> msg = dynamic_pointer_cast<data>(load);
    if(! msg)
      throw valueerror("actorserver::incoming_message:  Did not receive a valid "
                       "actorserver::data payload");
    actorid dest = msg->dest();
    mboxes_t::iterator it = mboxes_.find(dest);
    if(it != mboxes_.end()) {
      // We have this client
      shared_ptr<mbox> dest = it->second.lock();
      if(dest) {
        // and the client hasn't been garbage collected yet.
        dest->handle(msg->source(), msg->payload());
      }
    }

  }

} // end of namespace sstmac
