/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/status.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  // Pointer types.
  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Hello.
  //
  status::status() :
    start_(-1), stop_(-1), sent_(0), recvd_(0)
  {}

  //
  // Copy constructor.
  //
  status::status(const status &s) :
    start_(s.start_), stop_(s.stop_), sent_(s.sent_), recvd_(s.recvd_)
  {}

  //
  // Hello.
  //
  status::status(const timestamp &start_time, const timestamp &stop_time,
                 int64_t bytes_sent, int64_t bytes_recvd) :
    start_(start_time), stop_(stop_time), sent_(bytes_sent), recvd_(bytes_recvd)
  {
    if(start_ > stop_) {
      throw valueerror("status():  Start time is after stop time.");
    }
    if(sent_ < 0 || recvd_ < 0) {
      throw valueerror("status():  Negative byte count for sent or recvd.");
    }
  }

  //
  // Who am I?
  //
  void status::set_self(const boost::shared_ptr<status>& self) {
    if(self.get() != this) {
      throw valueerror("status::set_self:  'self' pointer does not point "
                       "to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  boost::shared_ptr<status> status::get_self() {
    return self_.lock();
  }

  //
  // Call default constructor.
  //
  boost::shared_ptr<status> status::construct() {
    boost::shared_ptr<status> retval(new status);
    retval->set_self(retval);
    return retval;
  }

  //
  // Make a new object.
  //
  boost::shared_ptr<status> status::construct(const timestamp &start_time,
                                              const timestamp &stop_time,
                                              int64_t bytes_sent,
                                              int64_t bytes_recvd)
  {
    boost::shared_ptr<status> retval(new status(start_time, stop_time,
                                                bytes_sent, bytes_recvd));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  status::~status() throw() {
  }

  //
  // Access start time.
  //
  const timestamp& status::start_time() const {
    return start_;
  }

  //
  // Set start time.
  //
  void status::set_start_time(const timestamp &t) {
    start_ = t;
  }

  //
  // Access completion time.
  //
  const timestamp& status::stop_time() const {
    return stop_;
  }

  //
  // Set completion time.
  //
  void status::set_stop_time(const timestamp &t) {
    if(t < start_) {
      throw valueerror("status::set_stop_time:  Can stop before start time.");
    }
    stop_ = t;
  }

  //
  // Access bytes sent.
  //
  int64_t status::bytes_sent() const {
    return sent_;
  }

  //
  // Increment bytes set.
  //
  void status::add_bytes_sent(int64_t bytes) {
    sent_ += bytes;
  }

  //
  // Access bytes received.
  //
  int64_t status::bytes_received() const {
    return recvd_;
  }

  //
  // Increment bytes received.
  //
  void status::add_bytes_received(int64_t bytes) {
    recvd_ += bytes;
  }

} // end of namespace sstmac
