/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESS_STATUS_H
#define SSTMAC_PROCESS_STATUS_H

#include <sstmac/simulator/timestamp.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <stdint.h>

namespace sstmac {

  /**
   * The status report associated with the completion of a kernel.
   * Reports details of when the kernel was run and its volume of
   * communications.
   *
   * Derived types may opt to ornament this with more detail.
   */
  class status {
    /// This object.
    boost::weak_ptr<status> self_;
    /// My timestamps.
    timestamp start_, stop_;
    /// The amount of bytes I communicated.
    int64_t sent_, recvd_;

  protected:
    /// Hello.
    status();

    /// Hello.
    status(const timestamp &start_time, const timestamp &stop_time,
           int64_t bytes_sent, int64_t bytes_recvd);

    /// Copy constructor.  Does not copy self_.
    status(const status &s);

    /// Who am I?
    void set_self(const boost::shared_ptr<status>& self);

    /// This is who I am.
    boost::shared_ptr<status> get_self();

  public:
    /// Default constructor.
    static boost::shared_ptr<status> construct();

    /// Make a new object.
    static boost::shared_ptr<status> construct(const timestamp &start_time,
                                               const timestamp &stop_time,
                                               int64_t bytes_sent,
                                               int64_t bytes_recvd);

    /// Goodbye.
    virtual ~status() throw();

    /// Access start time.
    const timestamp& start_time() const;

    /// Set start time.
    void set_start_time(const timestamp &t);

    /// Access completion time.
    const timestamp& stop_time() const;

    /// Set completion time.
    /// This time must not be before start_time.
    void set_stop_time(const timestamp &t);
    
    /// Access bytes sent.
    int64_t bytes_sent() const;

    /// Increment the number of bytes sent.
    void add_bytes_sent(int64_t bytes);

    /// Access bytes received.
    int64_t bytes_received() const;

    /// Increment the number of bytes received.
    void add_bytes_received(int64_t bytes);
  };

} // end of namespace sstmac.

#endif // ! SSTMAC_PROCESS_STATUS_H
