/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/processor/computekernel.h>
#include <sstmac/processor/instructions.h>
#include <sstmac/util/errors.h>
#include <sstmac/process/processrequest.h>
#include <sstmac/process/status.h>

namespace sstmac {

  // Constructor.
  computekernel::computekernel(const boost::shared_ptr<key>& thekey,
                               const boost::shared_ptr<instructions>& inst):
    key_(thekey), instructions_(inst), when_done_(processrequest::construct(thekey))
  { }

  void computekernel::set_self(const boost::shared_ptr<computekernel>& self) {
    if(self.get() != this) {
      throw valueerror("computerkernel::set_self:  'self' pointer does not refer "
                       "to 'this' object.");
    }
    self_ = self;
  }

  boost::shared_ptr<computekernel> computekernel::construct(const boost::shared_ptr<key>& thekey, 
                                                            const boost::shared_ptr<instructions>& inst) {
    boost::shared_ptr<computekernel> retval(new computekernel(thekey, inst));
    retval->set_self(retval);
    return retval;
  }

  computekernel::~computekernel() throw() { }

  //
  // This is who I am.
  //
  boost::shared_ptr<computekernel> computekernel::get_self()
  {
    return self_;
  }


  // Get a request for this kernel.
  boost::shared_ptr<processrequest> computekernel::get_request()
  {
    return when_done_;
  }

  // Start the kernel
  void computekernel::start(const timestamp &start_time,
                const boost::shared_ptr< nodeinterface >& my_node)
  {
    my_node->compute(start_time,instructions_,this->get_self());
  }

  // Completion event.
  void computekernel::handle(const timestamp &current_time) {
    this->complete(current_time);
  }

  // Mark this kernel as complete.
  // Notifies the request object prior to calling kernel::complete.
  void computekernel::complete(const timestamp &current_time)
  { 
    boost::shared_ptr<status> stat(status::construct());
    stat->set_stop_time(current_time);
    when_done_->complete(current_time,stat);
    kernel::complete(current_time);
  }


} // end of namespace sstmac
