/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_PROCESSOR_COMPUTEKERNEL_H
#define SSTMAC_PROCESSOR_COMPUTEKERNEL_H

#include <sstmac/simulator/kernel.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/nodeinterface.h>

namespace sstmac {

  // Forward declaration of important types.
  class status;
  class processrequest;
  class key;
  class nodeinterface;
  class instructions;

  /**
   * A kernel for computation.  Adds a way to get a request
   * handle for this kernel -- the request shall be notified when the
   * local chain of operations prescribed by this kernel is complete.
   *
   * The request is responsible for blocking and unblocking
   * individual tasks (mpiprogram? objects) in collaboration with the
   * scheduler; the computekernel doesn't have to worry about that stuff.
   */
  class computekernel : public kernel, public eventhandler {

    /// This request will be notified when we're done.
    boost::shared_ptr<processrequest> when_done_;

    boost::shared_ptr<computekernel> self_;
    boost::shared_ptr<key> key_;
    boost::shared_ptr<instructions> instructions_;

  private:

    void set_self( const boost::shared_ptr<computekernel>& );

    /// Constructor.
    computekernel(const boost::shared_ptr<key>& thekey, const boost::shared_ptr<instructions>& inst);

    /// Return self
    boost::shared_ptr<computekernel> get_self();

  public:
    /// Goodbye
    virtual ~computekernel() throw();

    static boost::shared_ptr<computekernel> 
    construct(const boost::shared_ptr<key>&, const boost::shared_ptr<instructions>& );

    /// Respond to an event call.
    void start (const timestamp &start_time, 
                const boost::shared_ptr< nodeinterface >& my_node);

    /// Get a request for this kernel.
    boost::shared_ptr<processrequest> get_request();

    /// Event handler.
    void handle(const timestamp &now);

  protected:
    // Mark this kernel as complete.
    // Notifies the request object prior to calling kernel::complete.
    void complete(const timestamp &current_time);

  };


} // end of namespace sstmac

#endif // ! SSTMAC_PROCESSOR_COMPUTEKERNEL_H
