/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/processor/procscheduler-loadmonitor.h>
#include <sstmac/processor/instructions.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/util/debug.h>
#include <iostream>

namespace sstmac {

  /// Protected constructor
  procscheduler::loadmonitor::
  loadmonitor(int n_core) : iter_(0), n_core_(n_core)
  { 
  }

  /// Public constructor
  boost::shared_ptr<procscheduler::loadmonitor>
  procscheduler::loadmonitor::construct(int n_core)
  {
    return boost::shared_ptr<loadmonitor>
      (new loadmonitor(n_core));
  }

  // Goodbye.
  procscheduler::loadmonitor::~loadmonitor() throw()
  {
  }

  /// Add instructions to monitor
  void 
  procscheduler::loadmonitor::
  add_instructions( int64_t compute_id,
                    const timestamp &start_time,
                    const timestamp &first_end_time )
  {
    start_time_[compute_id] = start_time;
    end_time_[compute_id] = first_end_time; 
    target_time_[compute_id] = first_end_time - start_time;

    if( debugger::active() ) {
      debug() << "loadmonitor: adding instructions object: id=" 
              << compute_id  << std::endl
              << "             start_time=" << start_time << std::endl
              << "             first_end_time=" << first_end_time << std::endl
              << "             target_time=" << first_end_time - start_time
              << std::endl;
    }
  }

  void
  procscheduler::loadmonitor::
  cleanup()
  {

    timestamp now = manager_->now();
    timestamp earliest = now;

    // find earliest start time for active instructions
    std::map<int64_t,timestamp>::iterator iter = start_time_.begin();
    for( iter; iter != start_time_.end();  ++iter ) {
      if( iter->second <= now && 
          end_time_[iter->first] >= now &&
          iter->second < earliest )
        earliest = iter->second;
    }
 
    // remove instructions with end times earlier than earliest start time
    std::vector<int64_t> invalid;
    iter = end_time_.begin();
    for( iter; iter != end_time_.end();  ++iter )
      if( iter->second <= earliest )
        invalid.push_back(iter->first);
    std::vector<int64_t>::iterator viter = invalid.begin();
    for( viter; viter != invalid.end(); ++viter) {
      start_time_.erase(*viter);
      end_time_.erase(*viter);
      target_time_.erase(*viter);
    }
  }

  /// Check if instructions are complete
  bool 
  procscheduler::loadmonitor:: 
  is_complete(int64_t compute_id)
  {
    // get segment list
    if( debugger::active() ) 
      debug() << "loadmonitor: creating segment list" << std::endl;
    std::vector<segment> segments;
    make_segment_list(segments, compute_id);
    print_segment_list(segments);

    // accumulate time
    timestamp start = start_time_[compute_id];
    timestamp end = end_time_[compute_id];
    timestamp t_acc = accumulate_time(segments,start,end,0);

    // find missing time
    timestamp t_miss = target_time_[compute_id] - t_acc; 

    if( debugger::active() ) {
      debug() << "loadmonitor: is_complete id: " << compute_id << std::endl
              << "             acc_time=" << t_acc << std::endl;
    }

    // done if accumulated time >= target 
    if( t_miss <= timestamp(0) ) {
      if( debugger::active() )
        debug() << "             instructions completed\n";
      return true;
    }
 
    if( debugger::active() ) 
      debug() << "             instructions not completed\n";
    return false;
  }

  /// Compute next possible end time
  timestamp
  procscheduler::loadmonitor:: 
  next_end_time(int64_t compute_id)
  {
    // get segment list
    std::vector<segment> segments;
    make_segment_list(segments, compute_id);

    // accumulate time
    timestamp start = start_time_[compute_id];
    timestamp end = end_time_[compute_id];
    timestamp t_acc = accumulate_time(segments,start,end,0);

    // find missing time
    timestamp t_miss = target_time_[compute_id] - t_acc;

    // get iterator to current end time
    std::vector<segment>::iterator iter = segments.begin();
    for(iter; iter != segments.end(); ++iter)
       if( (*iter).start == end ) break;

    if( debugger::active() ) {
      debug() << "loadmonitor: find next possible end time for id="
              << compute_id << std::endl
              << "             with current end time t_acc=" 
              << t_acc << std::endl
              << "                                   t_miss=" 
              << t_miss << std::endl;
    }

    // find next end time
    timestamp next_end;
    if( iter != segments.end() ) {
      for(iter; iter != segments.end(); ++iter) {
        timestamp t_add = accumulate_time(segments,end,(*iter).end,1);
        if( debugger::active() ) {
          debug() << "             adding another segment t_acc="
                  << t_acc + t_add << std::endl;
        }
        if(t_add > t_miss) {
          t_add = accumulate_time(segments,end,(*iter).start,1);
          t_miss -= t_add;

          // determine load factor
          double load_factor = (double) n_core_;
          load_factor /= (double) ((*iter).active + 1);
          if( load_factor > 1 ) load_factor=1;

          next_end = (*iter).start + t_miss / load_factor; 
          if( debugger::active() ) {
            debug() << "             too much, back up" << std::endl
                    << "             at end of last segment t_acc="
                    << t_acc + t_add << std::endl;
          }
          break;
        }
      }
    }
    else {
      next_end = end + t_miss;
    }
    end_time_[compute_id] = next_end;

    if( debugger::active() )
      debug() << "             next_end=" << next_end << std::endl;

    return next_end;
  }

  void
  procscheduler::loadmonitor::
  make_segment_list(std::vector<segment> &segments, int64_t compute_id) {

    cleanup();

    // make list of times which divide segments
    std::set<timestamp> times;
    std::map<int64_t,timestamp>::iterator iter = start_time_.begin();
    for( iter; iter != start_time_.end();  ++iter ) {
      timestamp starti = iter->second;
      if (starti >= start_time_[compute_id])
        times.insert(starti);
    }
    iter = end_time_.begin();
    for( iter; iter != end_time_.end();  ++iter ) {
      timestamp endi = iter->second;
      if (endi > start_time_[compute_id])
        times.insert(endi);
    }

    // make segment list
    std::set<timestamp>::iterator iter3 = times.begin();
    for(int i=0; i < times.size()-1; ++i ) {

      // start and end times
      timestamp begin=(*iter3);
      ++iter3;
      timestamp end=(*iter3);

      // number of actives
      int actives=0;
      iter = start_time_.begin();
      for(iter; iter != start_time_.end(); ++iter)
        if( iter->second < end &&
            end_time_[iter->first] > begin )
          ++actives;
       segment seg = { begin, end, actives };
       segments.push_back(seg);
    }
 
  }

  void
  procscheduler::loadmonitor::
  print_segment_list(std::vector<segment> &segments) {
    std::vector<segment>::iterator iter = segments.begin();
    int i=0;
    for(iter; iter != segments.end(); ++iter) {
      if( debugger::active() ) {
        debug() << "             Segment " << i << " begin: " << (*iter).start 
                << " end: " << (*iter).end << " actives: " 
                << (*iter).active << std::endl;
      }
      ++i;
    }
  }

  timestamp
  procscheduler::loadmonitor::
  accumulate_time(const std::vector<segment> &segments, 
                  const timestamp &start,
                  const timestamp &end,
                  const int &extra_load)
  {

    if( start >= end ) return timestamp(0);

    // get iterator to beginning segment
    std::vector<segment>::const_iterator iter = segments.begin();
    for(iter; iter != segments.end(); ++iter)
       if( (*iter).start == start ) break;

    // accumulate time over remaining segments
    timestamp acc_time(0);
    for(iter; iter != segments.end(); ++iter) {

      // determine load factor
      double load_factor = (double) n_core_;
      load_factor /= (double) ((*iter).active + extra_load);
      if( load_factor > 1 ) load_factor=1;

      acc_time += ( (*iter).end - (*iter).start ) * load_factor; 
      if( (*iter).end == end )
        break;
    }
      
    return acc_time;
  }

} // end of namespace sstmac.
