/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/processor/procscheduler.h>
#include <sstmac/processor/instructions.h>
#include <sstmac/processor/processor.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/processor/procscheduler-loadmonitor.h>
#include <sstmac/processor/procscheduler-loadhandler.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>
#include <boost/shared_ptr.hpp>
#include <iostream>

namespace sstmac {

  int64_t procscheduler::next_id_=0;

  //
  // Constructor
  //
  procscheduler::procscheduler(int n_core) 
  {
    monitor_ = loadmonitor::construct(n_core);
  }

  //
  // Who am I?
  //
  void procscheduler::set_self(const boost::shared_ptr<procscheduler>& self) {
    if(self.get() != this) {
      throw valueerror("procscheduler::set_self:  'self' pointer does not refer to 'this' object."); 
    }
    self_ = self;
  }

  //
  // Construct a new scheduler
  //
  boost::shared_ptr<procscheduler> 
  procscheduler::construct(int n_core) 
  {
    boost::shared_ptr<procscheduler> retval(new procscheduler(n_core));
    retval->set_self(retval);
    return retval;
  }

  void procscheduler::compute(const timestamp &start_time,
                              const boost::shared_ptr<instructions>& inst,
                              const boost::shared_ptr<eventhandler>& completion )
  {
    if(debugger::active())
      debug() << "procscheduler::compute(" << inst << ")\n";

    timestamp comp_time = processor_->compute(inst);
    monitor_->add_instructions(next_id_,start_time,start_time+comp_time);
    mgr_->schedule( start_time + comp_time, 
                     loadhandler::construct(next_id_,monitor_,mgr_,
                                            completion) );
    ++next_id_;
  }

  /// Define a manager.
  void procscheduler::set_manager(const boost::shared_ptr<eventmanager>& mgr) {
    mgr_ = mgr;
    monitor_->set_manager(mgr);
  }


} // end of namespace sstmac
