/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <boost/shared_ptr.hpp>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>
#include <sstmac/processor/simpleprocessor.h>
#include <sstmac/processor/eventdata.h>
#include <sstmac/processor/simpleprocessor-config.h>
#include <sstmac/processor/instructions.h>
#include <iostream>

namespace sstmac {

  //
  // Constructor
  //
  simpleprocessor::simpleprocessor( 
    const boost::shared_ptr<simpleprocessor_config>& cfg) :
    config_(cfg)
  {
    flop_label_ = config_->flop_label();
    memop_label_ = config_->memop_label();
    memop_local_label_ = config_->memop_local_label();
    memop_rand_label_ = config_->memop_rand_label();
    tflop_ = config_->tflop();
    tmem_ = config_->tmemop();
    tmemloc_ = config_->tmemop_local();
    tmemrnd_ = config_->tmemop_rand();

    if(debugger::active()) {
      debug() << "SimpleProcessor: initializing\n";
      debug() << "  average flop service time:   " << tflop_ << " ns\n";
      debug() << "  average memory service time: " << tmem_ << " ns\n";
    }
  }

  //
  // Who am I?
  //
  void simpleprocessor::set_self(const boost::shared_ptr<simpleprocessor>& self) {
    if(self.get() != this) {
      throw valueerror("simpleprocessor::set_self:  'self' pointer does not "
                       "refer to 'this' object.");
    }
    self_ = self;
  }

  //
  // Construct a new processor with the given average service times
  //
  boost::shared_ptr<simpleprocessor>
  simpleprocessor::construct(const boost::shared_ptr<simpleprocessor_config>& cfg) {
    boost::shared_ptr<simpleprocessor> 
      retval(new simpleprocessor(cfg));
    retval->set_self(retval);
    return retval;
  }

  boost::shared_ptr<processor> simpleprocessor::clone() const
  {
    if(debugger::active()) debug() << "simpleprocessor::clone()\n";
    boost::shared_ptr<simpleprocessor> 
      retval(new simpleprocessor(config_));
    return retval;
  }

  timestamp simpleprocessor::compute(const boost::shared_ptr<instructions>& inst)
  {
    if(debugger::active()) debug() << "simpleprocessor::compute(" 
                                   << inst << ")\n";
    boost::shared_ptr<eventdata> data = 
      boost::dynamic_pointer_cast<eventdata>( inst );
    
    // compute execution time in seconds
    timestamp time, zero;
    if(tflop_ > zero) {
      if(flop_label_.size() == 0)
        throw valueerror("missing flop data label");
      uint64_t nop = data->get_event(flop_label_);
      time += nop * tflop_;
    }
    if(tmem_ > zero) {
      if(memop_label_.size() == 0)
        throw valueerror("missing memory op data label");
      uint64_t nop = data->get_event(memop_label_);
      time += nop * tmem_;
    }
    if(tmemloc_ > zero) { 
      if(memop_local_label_.size() == 0)
        throw valueerror("missing local memory op data label");
      uint64_t nop = data->get_event(memop_local_label_);
      time += nop * tmemloc_;
    }
    if(tmemrnd_ > zero) {
      if(memop_rand_label_.size() == 0)
        throw valueerror("missing random memory op data label");
      uint64_t nop = data->get_event(memop_rand_label_);
      time += nop * tmemrnd_;
    }

    if(debugger::active())
      debug() << "SimpleProcessor: execution time: " << time.sec() * 1e6 << "us\n"; 

    return time;
  }

} // end of namespace sstmac
