/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/bandwidth.h>
#include <sstmac/util/errors.h>
#include <iostream>
#include <limits>
#include <math.h>
#include <boost/lexical_cast.hpp>

namespace sstmac {

  //
  // The static mapping from string names to unit labels.
  //
  bandwidth::unitmap_t *bandwidth::unitmap_ = NULL;

  //
  // Initialize the static unit map.
  //
  void bandwidth::init_unitmap() {
    if(unitmap_ == NULL) {
      unitmap_ = new unitmap_t;
#define SSTMAC_ADDUNIT(VAL) (*unitmap_)[#VAL] = VAL
      SSTMAC_ADDUNIT(bps);
      SSTMAC_ADDUNIT(Kbps);     SSTMAC_ADDUNIT(Kibps);
      SSTMAC_ADDUNIT(Mbps);     SSTMAC_ADDUNIT(Mibps);
      SSTMAC_ADDUNIT(Gbps);     SSTMAC_ADDUNIT(Gibps);
      SSTMAC_ADDUNIT(Tbps);     SSTMAC_ADDUNIT(Tibps);
      SSTMAC_ADDUNIT(Pbps);     SSTMAC_ADDUNIT(Pibps);
      SSTMAC_ADDUNIT(Ebps);     SSTMAC_ADDUNIT(Eibps);
      SSTMAC_ADDUNIT(Bps);
      SSTMAC_ADDUNIT(KBps);     SSTMAC_ADDUNIT(KiBps);
      SSTMAC_ADDUNIT(MBps);     SSTMAC_ADDUNIT(MiBps);
      SSTMAC_ADDUNIT(GBps);     SSTMAC_ADDUNIT(GiBps);
      SSTMAC_ADDUNIT(TBps);     SSTMAC_ADDUNIT(TiBps);
      SSTMAC_ADDUNIT(PBps);     SSTMAC_ADDUNIT(PiBps);
      SSTMAC_ADDUNIT(EBps);     SSTMAC_ADDUNIT(EiBps);
#undef SSTMAC_ADDUNIT
    }
  }

  //
  // Given two bandwidth values, compute the greatest common divisor.
  //
  bandwidth::bps_t bandwidth::gcd(const bps_t &a, const bps_t &b) const {
    if(b != 0) return gcd(b, a%b);
    else return a;
  }

  //
  // Convert unit enum to multiplier.
  //
  inline bandwidth::bps_t conversion_factor(bandwidth::units unit) {
    static bandwidth::bps_t *conv = NULL;
    if(conv == NULL) {
      conv = new bandwidth::bps_t[bandwidth::EiBps+1];
      for(int i = 0; i <= int(bandwidth::EiBps); ++i) {
        conv[i] = 0;
      }
#ifndef SSTMAC_USE_GMPXX
      conv[bandwidth::bps]     = 1ULL;
      conv[bandwidth::Kbps]    = 1000ULL;
      conv[bandwidth::Kibps]   = 1024ULL;
      conv[bandwidth::Mbps]    = 1000000ULL;
      conv[bandwidth::Mibps]   = 1048576ULL;
      conv[bandwidth::Gbps]    = 1000000000ULL;
      conv[bandwidth::Gibps]   = 1073741824ULL;
      conv[bandwidth::Tbps]    = 1000000000000ULL;
      conv[bandwidth::Tibps]   = 1099511627776ULL;
      conv[bandwidth::Pbps]    = 1000000000000000ULL;
      conv[bandwidth::Pibps]   = 1125899906842624ULL;
      conv[bandwidth::Ebps]    = 1000000000000000000ULL;
      conv[bandwidth::Eibps]   = 1152921504606846976ULL;
      conv[bandwidth::Bps]     = 8ULL;
      conv[bandwidth::KBps]    = 8000ULL;
      conv[bandwidth::KiBps]   = 8192ULL;
      conv[bandwidth::MBps]    = 8000000ULL;
      conv[bandwidth::MiBps]   = 8388608ULL;
      conv[bandwidth::GBps]    = 8000000000ULL;
      conv[bandwidth::GiBps]   = 8589934592ULL;
      conv[bandwidth::TBps]    = 8000000000000ULL;
      conv[bandwidth::TiBps]   = 8796093022208ULL;
      conv[bandwidth::PBps]    = 8000000000000000ULL;
      conv[bandwidth::PiBps]   = 9007199254740992ULL;
      conv[bandwidth::EBps]    = 8000000000000000000ULL;
      conv[bandwidth::EiBps]   = 9223372036854775808ULL;
#else
      conv[bandwidth::bps]     = "1";
      conv[bandwidth::Kbps]    = "1000";
      conv[bandwidth::Kibps]   = "1024";
      conv[bandwidth::Mbps]    = "1000000";
      conv[bandwidth::Mibps]   = "1048576";
      conv[bandwidth::Gbps]    = "1000000000";
      conv[bandwidth::Gibps]   = "1073741824";
      conv[bandwidth::Tbps]    = "1000000000000";
      conv[bandwidth::Tibps]   = "1099511627776";
      conv[bandwidth::Pbps]    = "1000000000000000";
      conv[bandwidth::Pibps]   = "1125899906842624";
      conv[bandwidth::Ebps]    = "1000000000000000000";
      conv[bandwidth::Eibps]   = "1152921504606846976";
      conv[bandwidth::Bps]     = "8";
      conv[bandwidth::KBps]    = "8000";
      conv[bandwidth::KiBps]   = "8192";
      conv[bandwidth::MBps]    = "8000000";
      conv[bandwidth::MiBps]   = "8388608";
      conv[bandwidth::GBps]    = "8000000000";
      conv[bandwidth::GiBps]   = "8589934592";
      conv[bandwidth::TBps]    = "8000000000000";
      conv[bandwidth::TiBps]   = "8796093022208";
      conv[bandwidth::PBps]    = "8000000000000000";
      conv[bandwidth::PiBps]   = "9007199254740992";
      conv[bandwidth::EBps]    = "8000000000000000000";
      conv[bandwidth::EiBps]   = "9223372036854775808";
#endif
    };
    if(unit < 0 || unit > bandwidth::EiBps)
      throw valueerror("bandwidth::units: Invalid unit (out of valid range).");
    bandwidth::bps_t value = conv[unit];
    if(value == 0)
      throw valueerror("bandwidth::units: Invalid unit "
                       "(uninitialized conversion).");
    return value;
  }

  //
  // Private initialization method used by constructors.
  //
  template <typename T>
  void bandwidth::set_bw(T bw, units unit) {
#ifndef SSTMAC_USE_GMPXX
    static const bps_t minbw = std::numeric_limits<bps_t>::min();
    static const bps_t maxbw = std::numeric_limits<bps_t>::max();
    double scaled = nearbyint(double(bw) * conversion_factor(unit));
    if(scaled < minbw || scaled > maxbw) {
      throw valueerror(std::string("bandwidth(...):  Requested bandwidth ") +
                       boost::lexical_cast<std::string>(scaled) + 
                       " bps is outside the valid range of " +
                       boost::lexical_cast<std::string>(minbw) + " to " +
                       boost::lexical_cast<std::string>(maxbw) +
                       " bits per second (inclusive).");
    }
    bits_per_second_ = bps_t(scaled);
#else
    bits_per_second_ = bps_t(bw) * conversion_factor(unit);
#endif
    // divide 1e12 psec/sec by timestamp::tick_interval() psec/tick
    // to arrive at ticks/sec.
    bps_t ticks_per_sec = bps_t(1e12) / timestamp::tick_interval();
    bps_t scaler = gcd(bits_per_second_, ticks_per_sec);
    bits_per_tick_.numerator   = bits_per_second_ / scaler;
    bits_per_tick_.denominator = ticks_per_sec / scaler;
  }

  //
  // Default constructor.
  //
  bandwidth::bandwidth() {
    set_bw(0, bps);
  }

  //
  // Build a bandwidth object with exactly the given bits per second.
  //
  bandwidth::bandwidth(const bps_t &bitspersecond) {
    set_bw(bitspersecond, bps);
  }

  //
  // Build a new bandwidth object.
  //
  bandwidth::bandwidth(double bw, units unit) {
    set_bw(bw, unit);
  }

  //
  // Alternate convenience way of defining the units (as a string).
  //
  bandwidth::bandwidth(double bw, const std::string &unit) {
    if(unitmap_ == NULL) {
      init_unitmap();
    }
    unitmap_t::const_iterator it = unitmap_->find(unit);
    if(it == unitmap_->end())
      throw valueerror(std::string("bandwidth(...):  Invalid unit string \"") +
                       unit + "\"");
    set_bw(bw, it->second);
  }

  //
  // Define a bandwidth based on a number of bits transferred
  // and the time it takes to transfer them.
  //
  bandwidth::bandwidth(const data &dt, const timestamp &ts) {
    // conversion factor is 1e12 psec/sec / tick_interval psec/tick = ticks/sec
    static const bps_t conv(bps_t(1e12) / timestamp::tick_interval());
    if(ts.ticks() == 0)
      throw valueerror("bandwidth(...):  Transfer time must not be zero.");
    // bandwidth is now   bits * ticks/sec / ticks = bits/sec
    set_bw((dt.get_bits() * conv) / ts.ticks(), bps);
  }

  //
  // Return the bandwidth using the given units.
  //
  double bandwidth::get_bandwidth(units unit) const {
#ifndef SSTMAC_USE_GMPXX
    return (bits_per_second_ / double(unit));
#else
    return (bits_per_second_.get_d() / double(unit));
#endif
  }

  //
  // A convenience function to specify units using a string value.
  //
  double bandwidth::get_bandwidth(const std::string &unit) const {
    if(unitmap_ == NULL) {
      init_unitmap();
    }
    unitmap_t::const_iterator it = unitmap_->find(unit);
    if(it == unitmap_->end()) {
      throw valueerror(std::string("bandwidth::get_bandwidth: Invalid unit \"")+
                       unit + "\"");
    }
    return get_bandwidth(it->second);
  }

  //
  // Given a quantity of data to transfer, return a timestamp
  // corresponding to how long the transfer would take.
  //
  timestamp bandwidth::transfer_time(data quantity) const {
#ifndef SSTMAC_USE_GMPXX
    data::data_t bits = quantity.get_bits();
    // do a pure integer multiplication if we can
    bps_t maxpure = (std::numeric_limits<bps_t>::max() /
                     bits_per_tick_.denominator);
    if(bits <= maxpure) {
      // Hooray!
      return timestamp::exact_ticks((bits * bits_per_tick_.denominator) /
                                    bits_per_tick_.numerator);
    }
    else {
      // Try another renormalization.
      bps_t scaler = gcd(bits, bits_per_tick_.numerator);
      bps_t reduced = bits / scaler;
      if(reduced <= maxpure) {
        // That worked.
        return timestamp::exact_ticks((reduced * bits_per_tick_.denominator) /
                                      (bits_per_tick_.numerator / scaler));
      }
      else {
        // Oh well, we might as well fall back on floating point.
        double ticks = ((double(bits) * bits_per_tick_.denominator) /
                        double(bits_per_tick_.numerator));
        return timestamp::exact_ticks(ticks);
      }
    }
    // we should never get here.
    throw illformederror("bandwidth::transfer_time:  Invalid fall-through.");
#else
    // Sometimes arbitrary width integers are just awesome.
    return timestamp::exact_ticks
      ((quantity.get_bits() * bits_per_tick_.denominator) /
       bits_per_tick_.numerator);
#endif
  }

  //
  // Report how much data would be transferred at this bandwidth.
  //
  data bandwidth::transferred_data(const timestamp &ts) const {
#ifndef SSTMAC_USE_GMPXX
    timestamp::tick_t ticks = ts.ticks();
    // do a pure integer multiplication if we can
    timestamp::tick_t maxpure = (std::numeric_limits<bps_t>::max() /
                     bits_per_tick_.numerator);
    if(ticks <= maxpure) {
      // Hooray!
      data::data_t bits_transferred((ticks * bits_per_tick_.numerator) /
                                    bits_per_tick_.denominator);
      return data(bits_transferred, data::bits);
    }
    else {
      // Try another renormalization.
      bps_t scaler = gcd(ticks, bits_per_tick_.denominator);
      timestamp::tick_t reduced = ticks / scaler;
      if(reduced <= maxpure) {
        // That worked.
        data::data_t bits_transferred((reduced * bits_per_tick_.numerator) /
                                      (bits_per_tick_.denominator / scaler));
        return data(bits_transferred, data::bits);
      }
      else {
        // Oh well, we might as well fall back on floating point.
        double bits_trans = ((double(ticks) * bits_per_tick_.numerator) /
                             double(bits_per_tick_.denominator));
        return data(bits_trans, data::bits);
      }
    }
    // we should never get here.
    throw illformederror("bandwidth::transferred_data:  Invalid fall-through.");
#else
    // Here, again, arbitrary width integers make things a lot simpler.
    return data((ts.ticks() * bits_per_tick_.numerator) /
                bits_per_tick_.denominator);
#endif
  }

  //
  // Increment bandwidth in place.
  //
  bandwidth& bandwidth::operator+=(const bandwidth &bw) {
    this->set_bw(this->bits_per_second_ + bw.bits_per_second_, bps);
    return *this;
  }

  //
  // Decrement bandwidth in place.
  //
  bandwidth& bandwidth::operator-=(const bandwidth &bw) {
    this->set_bw(this->bits_per_second_ - bw.bits_per_second_, bps);
    return *this;
  }

  //
  // Scale bandwidth in place.
  //
  bandwidth& bandwidth::operator*=(double scaling) {
    this->set_bw(this->bits_per_second_ * scaling, bps);
    return *this;
  }

  //
  // Divide bandwidth in place.
  //
  bandwidth& bandwidth::operator/=(double scaling) {
    this->set_bw(this->bits_per_second_ / scaling, bps);
    return *this;
  }

  //
  // Increase bandwidth.
  //
  bandwidth operator+(const bandwidth &a, const bandwidth &b) {
    const bandwidth::bps_t abps = a.bits_per_second();
    const bandwidth::bps_t bbps = b.bits_per_second();
    bandwidth::bps_t sum = abps + bbps;
#ifndef SSTMAC_USE_GMPXX
    // Detect overflow -- only tested against two's complement.
    if((sum < abps) || (sum < bbps)) {
      throw valueerror
        (std::string("operator+(const bandwidth&, const bandwidth&) :  ") +
         "Adding " + boost::lexical_cast<std::string>(a) + " and " +
         boost::lexical_cast<std::string>(b) + " results in overflow.");
    }
#endif
    return bandwidth(sum);
  }

  //
  // Decrease bandwidth.
  //
  bandwidth operator-(const bandwidth &a, const bandwidth &b) {
    const bandwidth::bps_t abps = a.bits_per_second();
    const bandwidth::bps_t bbps = b.bits_per_second();
    bandwidth::bps_t del = abps - bbps;
#ifndef SSTMAC_USE_GMPXX
    // Detect overflow -- only tested against two's complement.
    if((del > abps) || (del > bbps)) {
      throw valueerror
        (std::string("operator+(const bandwidth&, const bandwidth&) :  ") +
         "Subtracting " + boost::lexical_cast<std::string>(a) + " and " +
         boost::lexical_cast<std::string>(b) + " results in overflow.");
    }
#endif
    return bandwidth(del);
  }

  //
  // Scale bandwidth.
  //
  bandwidth operator*(const bandwidth &t, double scaling) {
#ifndef SSTMAC_USE_GMPXX
    // Let's chicken out and just go with floating point ops.
    double scaled = nearbyint(t.bits_per_second() * scaling);
    if(scaled < double(std::numeric_limits<bandwidth::bps_t>::min()) ||
       scaled > double(std::numeric_limits<bandwidth::bps_t>::max())) {
      throw valueerror(std::string("operator*(const bandwidth&, double):  ") +
                       "Multiplying " + boost::lexical_cast<std::string>(t)  +
                       " with " + boost::lexical_cast<std::string>(scaling)  +
                       " results in value outside allowable bandwidth range.");
    }
#else
    // Note that gmpxx uses very stupid type promotion rules.
    bandwidth::bps_t scaled = t.bits_per_second().get_d() * scaling;
#endif
    return bandwidth(scaled);
  }

  //
  // Scale bandwidth.
  //
  bandwidth operator*(double scaling, const bandwidth &t) {
    return (t * scaling);
  }

  //
  // Scale bandwidth.
  //
  bandwidth operator/(const bandwidth &t, double scaling) {
#ifndef SSTMAC_USE_GMPXX
    double scaled = nearbyint(t.bits_per_second() / scaling);
    if(scaled < double(std::numeric_limits<bandwidth::bps_t>::min()) ||
       scaled > double(std::numeric_limits<bandwidth::bps_t>::max())) {
      throw valueerror(std::string("operator*(const bandwidth&, double):  ") +
                       "Dividing " + boost::lexical_cast<std::string>(t) +
                       " with " + boost::lexical_cast<std::string>(scaling) +
                       " results in value outside allowable bandwidth range.");
    }
#else
    // Bypass stupid promotion rules in gmpxx
    bandwidth::bps_t scaled = t.bits_per_second().get_d() / scaling;
#endif
    return bandwidth(scaled);
  }

  //
  // Very, very convenient.
  //
  std::ostream& operator<<(std::ostream &os, const bandwidth &t) {
    os << "bandwidth(" << t.bits_per_second() << "bps)";
    return os;
  }


} // end of namespace sstmac.
