/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_FLOWER_H
#define SSTMAC_SIMULATOR_FLOWER_H

#include <sstmac/simulator/timestamp.h>
#include <sstmac/network/network.h>
#include <vector>
#include <map>
#include <set>
#include <string>
#include <boost/shared_ptr.hpp>

namespace sstmac {

  typedef int edge_id_t;
  typedef int flow_id_t;
  
  /**
   * Abstract base class that represents network traffic as a set of
   * concurrent flows. Different specializations schedule the flows in
   * different ways. A flower object is used by sharedcircuit to
   * determine when network traffic commences and completes.
   */
  class flower {
  public:
    bool debug;

  private:
    boost::shared_ptr<sstmac::network> net_;

  protected:
    flower(const boost::shared_ptr<sstmac::network> &net): debug(false), net_(net) {}

  public:
    virtual ~flower() throw();

    /// Start a flow through the graph. The edge identifiers are indices into 
    /// the edge_bw_ array.
    virtual flow_id_t start_flow(timestamp tstart, const std::vector<edge_id_t> &edges, long nbyte) = 0;

    /// Stop a flow.
    virtual void stop_flow(timestamp t, flow_id_t f) = 0;

    /// The next time that any flow will stop. It is an error to call this if
    /// there is no next stopping flow.
    virtual timestamp next_stopping_time() = 0;

    /// The identifier of the next flow that will stop. A return value of zero
    /// indicates that there is no next stopping flow.
    virtual flow_id_t next_stopping_flow() = 0;

    /// Accesses the network used to construct this object.
    const boost::shared_ptr<sstmac::network> &net() const { return net_; }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_FLOWER_H
