/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/fullnetwork.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/simulator/constantlatency.h>
#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>
#include <iostream>

namespace sstmac {

  using boost::shared_ptr;

  //
  // Construct a fullnetwork object with the given properties
  //
  fullnetwork::fullnetwork(const shared_ptr<latencymodel> &latency,
                           bandwidth bw, int maxsize) :
    latency_(latency), bandwidth_(bw), size_(maxsize)
  {
    if(debugger::active()) debug() << "fullnetwork(" << latency << ", " << bw
                                   << ", " << maxsize << ")\n";
    if(! latency)
      throw nullerror("fullnetwork:  NULL latency model pointer.");
    if(bandwidth_.bits_per_second() <= 0)
      throw valueerror("fullnetwork:  Bandwidth must be greater than zero.");
  }

  //
  // Find out who I am.
  //
  void fullnetwork::set_self(const shared_ptr<fullnetwork> &fn) {
    if(fn.get() != this) {
      throw valueerror("fullnetwork::set_self:  'self' pointer does not point "
                       "to 'this' object.");
    }
    self_ = fn;
  }

  //
  // Let you know who I am.
  //
  shared_ptr<fullnetwork> fullnetwork::get_self() {
    return self_.lock();
  }

  //
  // Construct,
  //
  shared_ptr<fullnetwork> fullnetwork::construct(const timestamp &latency,
                                                 double bandw,
                                                 int maxsize)
  {
    if(latency < timestamp(0)) {
      throw valueerror("fullnetwork:  Latency must not be less than zero.");
    }
    shared_ptr<constantlatency> lat = constantlatency::construct(latency);
    bandwidth bw(bandw, bandwidth::Bps);
    return construct(lat, bw, maxsize);
  }

  //
  // Construct.
  //
  shared_ptr<fullnetwork>
  fullnetwork::construct(const shared_ptr<latencymodel> &lat,
                         bandwidth bw, int maxsize)
  {
    shared_ptr<fullnetwork> retval(new fullnetwork(lat, bw, maxsize));
    retval->set_self(retval);
    return retval;
  }


  //
  // Exit stage left.
  //
  fullnetwork::~fullnetwork() throw() {
  }

  //
  // The maximum number of nodes possible in this simulator.
  //
  int fullnetwork::slots() const {
    if(debugger::active()) debug() << "fullnetwork::slots()\n";
    return size_;
  }

  //
  // Define the eventmanager.
  //
  void fullnetwork::set_eventmanager(const shared_ptr<eventmanager> &mgr) {
    if(mgr == 0) {
      throw nullerror("fullnetwork::set_eventmanager:  Null manager pointer.");
    }
    eventmanager_ = mgr;
  }

  //
  // Send the given message at the given time.
  //
  void fullnetwork::send(const timestamp &time,
                         int64_t bytes,
                         nodeid sender,
                         nodeid recver,
                         const shared_ptr<eventhandler> &sendhandler,
                         const shared_ptr<eventhandler> &recvhandler)
  {
    if(debugger::active()) debug() << "fullnetwork::send(" << time << ", " << bytes << ", "
            << sender << ", " << recver << ", " << sendhandler << ", "
            << recvhandler << ")\n";
    if(sendhandler == 0) {
      throw nullerror("fullnetwork::send:  null sendhandler pointer.");
    }
    if(recvhandler == 0) {
      throw nullerror("fullnetwork::send:  null recvhandler pointer.");
    }
    // Somewhat erroneously, assume that the send completes the moment it is
    // initiated (should probably be at a later time -- it must take some amount
    // of time to put data on the wire).
    eventmanager_->schedule(time, sendhandler);
    // The receive operation is congestion-free.
    data payload(bytes, data::Bytes);
    timestamp lat = latency_->latency(payload);
    timestamp recvtime = (time + lat + payload/bandwidth_);
    eventmanager_->schedule(recvtime, recvhandler);
  }

} // end of namespace sstmac 
