/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/networkcall.h>
#include <sstmac/simulator/interconnect.h>
#include <sstmac/simulator/nodelist.h>
#include <sstmac/simulator/networkdata.h>
#include <sstmac/process/recvhandler.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>

namespace sstmac {

  using boost::shared_ptr;

  //
  // Create a new passthrough for event scheduling.
  //
  networkcall::networkcall(const shared_ptr<interconnect> &net,
                           const shared_ptr<nodelist> &nodes,
                           const nodeid &id) :
    interconnect_(net), nodelist_(nodes), id_(id)
  {}

  //
  // Hi there.
  //
  shared_ptr<networkcall>
  networkcall::construct(const shared_ptr<interconnect> &net,
                         const shared_ptr<nodelist> &nodes,
                         const nodeid &id)
  {
    shared_ptr<networkcall> retval(new networkcall(net, nodes, id));
    return retval;
  }

  //
  // Goodbye.
  //
  networkcall::~networkcall() throw() {
  }

  //
  // Schedule a send.
  //
  void networkcall::send(const timestamp &start_time,
                         const ppid &the_process,
                         const nodeid &the_recver,
                         const shared_ptr<networkdata> &the_payload,
                         const shared_ptr<eventhandler> &the_sendevent)
  {
    if(debugger::active())
      debug() << "networkcall[" << this << "]::send(" << start_time << ", "
              << the_process << ", " << the_recver << ", " << the_payload
              << ", " << the_sendevent << ")\n";
    if(the_payload == 0) {
      throw nullerror("networkcall::send:  null networkdata pointer.");
    }
    // Get the node pointer for the receiver (needed for callback operations).
    shared_ptr<node> recvnode = nodelist_.lock()->get(the_recver);
    // Set up the event associated with completion of the receive
    shared_ptr<eventhandler> the_recvevent =
      recvhandler::construct(recvnode, the_process, id_, the_payload);
    // Fire off the actual send operation
    interconnect_->send(start_time, the_payload->bytes(), id_, the_recver,
                        the_sendevent, the_recvevent);
  }

} // end of namespace sstmac.
