/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_NETWORKCALL_H
#define SSTMAC_SIMULATOR_NETWORKCALL_H

#include <sstmac/simulator/timestamp.h>
#include <sstmac/simulator/nodeid.h>
#include <sstmac/simulator/ppid.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>

namespace sstmac {

  // Important types.
  class nodelist;
  class interconnect;
  class networkdata;
  class eventhandler;

  /**
   * Submit requests from a node to a network.
   */
  class networkcall {
  private:
    /// The network I am communicating with.
    boost::shared_ptr<interconnect> interconnect_;

    /// The (shared) mapping from nodeid values onto node pointers.
    /// The node is permitted to have a reference-counted pointer to
    /// this object, so we need to keep this as a weak reference.
    boost::weak_ptr<const nodelist> nodelist_;

    /// The nodeid for my node (to communicate with the network).
    nodeid id_;

  protected:
    /// Construct a new event passthrough.
    networkcall(const boost::shared_ptr<interconnect> &net,
                const boost::shared_ptr<nodelist> &nodes,
                const nodeid &id);
    
  public:
    /// Hello.
    static boost::shared_ptr<networkcall>
    construct(const boost::shared_ptr<interconnect> &net,
              const boost::shared_ptr<nodelist> &nodes,
              const nodeid &id);

    /// Goodbye.
    virtual ~networkcall() throw();

    /// Schedule a send.
    /// The sender node will be notified via eventhandler::complete
    /// when the send operation is complete.
    /// The receiver node will be given a copy of the networkevent
    /// via node::complete_send.  It is up to the node to find a
    /// suitable recipient for an incoming message.
    virtual void send(const timestamp &start_time,
                      const ppid &the_process,
                      const nodeid &the_receiver,
                      const boost::shared_ptr<networkdata> &the_payload,
                      const boost::shared_ptr<eventhandler> &the_sendevent);
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_NETWORKCALL_H
