/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_NODE_H
#define SSTMAC_SIMULATOR_NODE_H

#include <sstmac/simulator/timestamp.h>
#include <sstmac/simulator/ppid.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/simulator/nodeid.h>
#include <sstmac/processor/processor.h>
#include <boost/shared_ptr.hpp>

namespace sstmac {

  // Forward declarations of relevant types.
  class server;
  class networkcall;
  class eventmanager;
  class eventhandler;
  class networkdata;
  class instructions;
  class nodeinterface;
  class networkinterface;

  /**
   * An interface for "physical" nodes -- may eventually be decorated
   * with additional functionality (OS/cpu modeling or whatever).
   */
  class node {

  public:
    /// Goodbye.
    virtual ~node() {}

    /// Get a new node of the same type as the current one.
    virtual boost::shared_ptr<node> clone() const = 0;

    /// Assign a node index and event call interface for this node.
    /// \throw nullerror if ev is null.
    virtual void initialize(nodeid my_id,
                            const boost::shared_ptr<networkcall> &net,
                            const boost::shared_ptr<eventmanager> &mgr) = 0;

    /// Access the event manager.
    virtual boost::shared_ptr<eventmanager> get_eventmanager() const = 0;

    /// Current time.  To be pulled from the eventmanager.
    virtual const timestamp& now() const = 0;

    /// Get a nodeinterface bridging the given ppid and taskid to this node.
    virtual boost::shared_ptr<nodeinterface>
    get_nodeinterface(const ppid &the_ppid, const taskid &the_taskid) = 0;

    /// Test whether the given ppid has defined a server on this node.
    virtual bool has_server(const ppid &caller) const = 0;

    /// Set a network event server for the given ppid.
    /// Each ppid can only have one network server on a node.
    /// The node must call server::initialize for the provided server.
    virtual void set_server(ppid caller, const boost::shared_ptr<server> &) = 0;

    /// Get the network serever for the given ppid.
    virtual boost::shared_ptr<server> get_server(const ppid &caller) = 0;

    /// Test whether this node has a processor model defined.
    virtual bool has_processor() const = 0;

    /// Set a processor.
    virtual void set_processor(const boost::shared_ptr<processor> &) = 0;

    /// Get the processor.
    virtual const boost::shared_ptr<processor> get_processor() const = 0;

    /// Set a simple trigger (wake-up call).
    /// \param start_time is the time at which the trigger shall be fulfilled.
    /// \param completion is the eventhandler that gets signaled at start_time.
    virtual void schedule(const timestamp &start_time,
                          ppid caller,
                          const boost::shared_ptr<eventhandler> &completion) = 0;

    /// Pretend to compute for the given number of cpu seconds.
    virtual void compute(const timestamp &start_time,
                         ppid caller,
                         const timestamp &cputime,
                         const boost::shared_ptr<eventhandler> &completion) = 0;

    /// Pretend to compute for the given number of cpu seconds.
    virtual void compute(const timestamp &start_time,
                         ppid caller,
                         const timestamp &cputime_per_thread,
                         int threads,
                         const boost::shared_ptr<eventhandler> &completion) = 0;

    /// I am not sure yet how best to handle this one :)
    /// Eventually, this is where we would bridge over to SST/microscale
    /// or other processor models.
    /// \param start_time is the time when the computation should start.
    /// \param instr is the instruction object that will be run.
    /// \param completion is the eventhandler that gets signaled when
    ///        the instructions have been run.
    virtual void compute(const timestamp &start_time,
                         ppid caller,
                         const boost::shared_ptr<instructions> &instr,
                         const boost::shared_ptr<eventhandler> &completion) = 0;

    /// Pretend to copy the given number of bytes between two address spaces.
    /// Used by messaging models to assess cost of non-network 
    /// data transfers.
    /// Completion shall be signaled using the
    /// delegate::kernel()->complete_copy call.
    /// \param start_time time when copy is initiated.
    /// \param bytes number of bytes to copy.
    /// \param completion is the eventhandler signaled upon completion.
    /// \throw valueerror if bytes < 0.
    virtual void copy(const timestamp &start_time,
                      ppid caller,
                      int64_t bytes,
                      const boost::shared_ptr<eventhandler> &completion) = 0;

    /// Incoming network transfer.
    /// Originates from a network simulator.
    virtual void incoming_message(const timestamp &start_time,
                                  ppid caller,
                                  nodeid sender,
                                  const boost::shared_ptr<networkdata> &load)=0;

  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_NODE_H
