/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_ORIGINAL_FLOWER_H
#define SSTMAC_SIMULATOR_ORIGINAL_FLOWER_H

#include <sstmac/simulator/flower.h>
#include <sstmac/simulator/netsharing.h>
#include <sstmac/simulator/timestamp.h>
#include <sstmac/network/network.h>
#include <vector>
#include <map>
#include <set>
#include <string>
#include <boost/shared_ptr.hpp>

namespace sstmac {
  
  /**
   * This is the original flower and is deprecated.
   *
   * This code is for testing only and will disappear in an upcoming
   * release. Please use another flower specialization, such as
   * ordered_flower.
   *
   */
  class original_flower: public flower {
  public:
    long n_rebalance_max; // used by ORDERED (-1 means everything)
  private:

    netsharing algorithm_;
    flow_id_t next_flow_id_;
    flow_id_t stopping_flow_;
    timestamp stopping_time_;

    double max_bw_;
    std::vector<double> edge_bw_;
    double edge_beta(size_t i) const { return 1.0/edge_bw_.at(i); }
    double edge_bw(size_t i) const { return edge_bw_.at(i); }
    /// current time for flow data
    timestamp t_;
    //
    typedef std::map<flow_id_t, std::vector<edge_id_t> > flow_edges_t;
    flow_edges_t flow_edges_;
    //
    typedef std::map<flow_id_t, double> flow_bw_t;
    flow_bw_t flow_bw_;
    //
    typedef std::set<flow_id_t> flow_set_t;
    typedef std::vector<flow_set_t> edge_flows_t;
    edge_flows_t edge_flows_;
    /// bytes remaining in flow
    typedef std::map<flow_id_t,double> flow_nbyte_t;
    flow_nbyte_t flow_nbyte_; 

    void advance_time(timestamp);
    void rebalance_flow_start(flow_id_t f);
    void rebalance_flow_stop(flow_id_t oldf);
    void rebalance_flow_overcommit(flow_id_t f);
    void rebalance_flow_throttle_overcommits(flow_id_t f);
    void rebalance_flow_ordered(flow_id_t f, double target_bw);
    void rebalance_flow_start_overcommit(flow_id_t f);
    void rebalance_flow_start_throttle_overcommits(flow_id_t f);
    void rebalance_flow_start_ordered(flow_id_t f);
    void rebalance_flow_stop_overcommit(flow_id_t oldf);
    void rebalance_flow_stop_throttle_overcommits(flow_id_t oldf);
    void rebalance_flow_stop_ordered(flow_id_t oldf);
    double edge_bw_utilization(edge_id_t e);
    double edge_bw_equal_and_higher_priority_utilization(edge_id_t e, flow_id_t f);
    void find_next_stop();

  protected:
    original_flower(const boost::shared_ptr<sstmac::network> &net,
           double node_router_bw, double router_router_bw,
           netsharing algo);

  public:
    static boost::shared_ptr<original_flower>
      construct(const boost::shared_ptr<sstmac::network> &net,
                double node_router_bw, double router_router_bw,
                netsharing algo = ORDERED);

    virtual ~original_flower() throw() {}

    /// Start a flow through the graph. The edge identifiers are indices into 
    /// the edge_bw_ array.
    flow_id_t start_flow(timestamp tstart, const std::vector<edge_id_t> &edges, long nbyte);
    /// Stop a flow.
    void stop_flow(timestamp t, flow_id_t f);
    /// The next time that any flow will stop. It is an error to call this if
    /// there is no next stopping flow.
    timestamp next_stopping_time();
    /// The identifier of the next flow that will stop. A return value of zero
    /// indicates that there is no next stopping flow.
    flow_id_t next_stopping_flow();

    /// The current net sharing algorithm.
    netsharing sharing() const { return algorithm_; }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_ORIGINAL_FLOWER_H
