/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/overcommit_flower.h>
#include <sstmac/util/errors.h>

#include <boost/lexical_cast.hpp>
#include <iterator>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <stdlib.h>

using boost::shared_ptr;

namespace sstmac {

  shared_ptr<overcommit_flower>
  overcommit_flower::construct(const boost::shared_ptr<sstmac::network> &net,
                               double node_router_bw, double router_router_bw)
  {
    shared_ptr<overcommit_flower> self(new overcommit_flower(net,
                                       node_router_bw, router_router_bw));
    return self;
  }

  overcommit_flower::overcommit_flower(const boost::shared_ptr<sstmac::network> &net,
                                       double node_router_bw, double router_router_bw):
    flower(net),
    node_router_bw_(node_router_bw),
    router_router_bw_(router_router_bw),
    fid_(1)
  {
  }

  overcommit_flower::~overcommit_flower() throw()
  {
  }

  void
  overcommit_flower::require_one_flow()
  {
    if (stop_time_.begin() == stop_time_.end()) {
      throw illformederror("overcommit_flower: no flows available but at least one needed");
    }
  }

  flow_id_t
  overcommit_flower::start_flow(timestamp tstart,
                                const std::vector<edge_id_t> &edges,
                                long nbyte)
  {
    double edge_bw = std::max(node_router_bw_, router_router_bw_);
    for (int i=0; i<edges.size(); i++) {
      sstmac::edge_index ei(edges[i]);
      const sstmac::lwedge &e = net()->g().index_to_edge(ei);
      double edge_i_bw;
      if (net()->is_node(e.from()) || net()->is_node(e.to())) {
        edge_i_bw = node_router_bw_;
      }
      else {
        edge_i_bw = router_router_bw_;
      }
      edge_bw = std::max(edge_i_bw, edge_bw);
    }

    timestamp tstop(tstart.sec() + nbyte/edge_bw);

    flow_id_t f = fid_++;

    stop_time_.insert(std::make_pair(tstop, f));

    t_ = tstart;

    if (debug) std::cout << "overcommit_flower::start_flow: started " << f
                         << " from " << tstart
                         << " to " << tstop << std::endl;

    return f;
  }

  void
  overcommit_flower::stop_flow(timestamp t, flow_id_t f)
  {
    if (t < t_) {
      throw illformederror("overcommit_flower:stop_flow: moving back in time");
    }
    // find f (it is probably first)
    for (std::multimap<timestamp, flow_id_t>::iterator iter = stop_time_.begin();
         iter != stop_time_.end(); iter++) {
      if (iter->second == f) {
        if (iter->first < t) {
          throw illformederror("overcommit_flower:stop_flow: stopping after flow stop time");
        }
        if (debug) std::cout << "overcommit_flower:stop_flow("
                             << t << ", " << f << ")" << std::endl;
        stop_time_.erase(iter);
        t_ = t;
        return;
      }
    }
    throw illformederror("overcommit_flower:stop_flow: flow not found");
  }

  timestamp
  overcommit_flower::next_stopping_time()
  {
    require_one_flow();
    return stop_time_.begin()->first;
  }

  flow_id_t
  overcommit_flower::next_stopping_flow()
  {
    flow_id_t stopping_flow;
    if (stop_time_.begin() == stop_time_.end()) stopping_flow = 0;
    else stopping_flow = stop_time_.begin()->second;
    if (debug) std::cout << "overcommit_flower::next_stopping_flow: "
                         <<  stopping_flow << std::endl;
    return stopping_flow;
  }

} // end of namespace sstmac
