/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SIMULATOR_SHAREDCIRCUIT_STOPPER_H
#define SSTMAC_SIMULATOR_SHAREDCIRCUIT_STOPPER_H

#include <sstmac/simulator/flower.h>
#include <sstmac/simulator/sharedcircuit.h>
#include <sstmac/simulator/eventhandler.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/util/errors.h>
#include <map>
#include <set>

namespace sstmac {

  /**
   * The sharedcircuit::stopper takes care of stopping flows that
   * have completed and signaling the sender and receiver.
   */
  class sharedcircuit::stopper : public eventhandler {
    /// Private nested type to handle requests.
    struct registration {
      timestamp recv_delay;
      boost::shared_ptr<eventhandler> sender, recver;
      void init(const timestamp &rdelay,
                const boost::shared_ptr<eventhandler> &ss,
                const boost::shared_ptr<eventhandler> &rr)
      {
        recv_delay = rdelay;
        sender = ss;
        recver = rr;
      }
      void notify(const timestamp &now, const boost::shared_ptr<eventmanager> &mgr) {
        sender->handle(now);
        mgr->schedule(now + recv_delay, recver);
      }
    };

    /// Make sure we don't stop the same id twice.
    std::set<flow_id_t> stopped_;

    /// This object.  Needed so we can re-schedule ourselves.
    boost::weak_ptr<stopper> self_;

    /// The flow management object.
    boost::shared_ptr<flower> flower_;

    /// The event manager -- needed to reschedule ourselves.
    boost::shared_ptr<eventmanager> mgr_;
    /// The next flow to stop.  This may get reassigned during update(...).
    flow_id_t next_;
    /// The time at which the next flow should stop.
    /// Any incoming event notifications that precede this time get ignored.
    timestamp stoptime_;

    /// Keep track of message requests.
    typedef std::map<flow_id_t, registration> reg_t;
    reg_t reg_;

  protected:
    /// Sliding down the razor blade of life.
    stopper(const boost::shared_ptr<flower> &flower) :
      flower_(flower), next_(-1), stoptime_(-1)
    {}

    /// Here you are.
    void set_self(const boost::shared_ptr<stopper> &self) {
      if(self.get() != this)
        throw valueerror("sharedcircuit::stopper:  'self' pointer "
                         "does not point to 'this' object.");
      self_ = self;
    }

  public:
    /// Get a stopper.
    static boost::shared_ptr<stopper>
    construct(const boost::shared_ptr<flower> &flower) {
      boost::shared_ptr<stopper> self(new stopper(flower));
      self->set_self(self);
      return self;
    }

    /// Unstoppable.
    virtual ~stopper() throw() {}

    /// This is where we get our event manager.
    void set_eventmanager(const boost::shared_ptr<eventmanager> &mgr) {
      mgr_ = mgr;
    }

    /// We may have a new "first flow to finish" -- if so reschedule this event.
    void update(flow_id_t flow, const timestamp &newstop) {
      if(! ((flow == next_) && (newstop == stoptime_))) {
        next_ = flow;
        stoptime_ = newstop;
        mgr_->schedule(stoptime_, self_.lock());
      }
    }

    /// Register a new flow with the stopper.
    void add(flow_id_t flow,
             const timestamp &recv_delay,
             const boost::shared_ptr<eventhandler> &sendhandler,
             const boost::shared_ptr<eventhandler> &recvhandler)
    {
      if(reg_.find(flow) != reg_.end())
        throw illformederror("sharedcircuit::stopper::add:  Attempt to "
                             "register a pre-existing flow id.");
      reg_[flow].init(recv_delay, sendhandler, recvhandler);
    }

    /// Our time has come.
    void handle(const timestamp &now) {
      if(now != stoptime_) {
        // Ignore this event -- it has been rescheduled
        // (this is how we get singleton behavior).
        return;
      }
      if(stopped_.find(next_) == stopped_.end()) {
        flower_->stop_flow(now, next_);
        stopped_.insert(next_);
        // Notify the sender and receiver.
        reg_t::iterator it = reg_.find(next_);
        if(it == reg_.end()) {
          throw illformederror("sharedcircuit::stopper::handle:  Stopping a "
                               "flow that was not previously registered.");
        }
        it->second.notify(now, mgr_);
        reg_.erase(it);
      }
      // Reschedule
      if (flower_->next_stopping_flow() != 0) {
        next_ = -1; // we don't have a flow to stop right now.
        this->update(flower_->next_stopping_flow(),
                     flower_->next_stopping_time());
      }
    }
  };

} // end of namespace sstmac

#endif // ! SSTMAC_SIMULATOR_SHAREDCIRCUIT_STOPPER_H
