/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/sharedcircuit.h>
#include <sstmac/simulator/sharedcircuit-stopper.h>
#include <sstmac/simulator/sharedcircuit-starter.h>
#include <sstmac/simulator/constantlatency.h>
#include <sstmac/simulator/data.h>
#include <sstmac/network/network.h>
#include <sstmac/network/graph.h>
#include <sstmac/network/walk.h>
#include <sstmac/util/errors.h>

namespace sstmac {

  using boost::shared_ptr;

  //
  // Construct a new circuit.
  //
  sharedcircuit::sharedcircuit(const shared_ptr<latencymodel> &latency,
                               bandwidth bw,
                               const shared_ptr<sstmac::network> &net,
                               const shared_ptr<flower> &flow) :
    flower_(flow), stopper_(stopper::construct(flower_)),
    latency_(latency), bandwidth_(bw), net_(net)
  {}

  //
  // Construct a new circuit.
  //
  shared_ptr<sharedcircuit>
  sharedcircuit::construct(const timestamp latency, double bw,
                           const boost::shared_ptr<sstmac::network> &net,
                           const boost::shared_ptr<flower> &flow)
  {
    return construct(constantlatency::construct(latency),
                     bandwidth(bw, bandwidth::Bps), net, flow);
  }

  //
  // Construct a new circuit.
  //
  shared_ptr<sharedcircuit>
  sharedcircuit::construct(const shared_ptr<latencymodel> &latency,
                           bandwidth bw,
                           const shared_ptr<network> &net,
                           const shared_ptr<flower> &flow)
  {
    return shared_ptr<sharedcircuit>
      (new sharedcircuit(latency, bw, net, flow));
  }

  //
  // So long, Mr. Despair.
  //
  sharedcircuit::~sharedcircuit() throw() {
  }

  //
  // Required method from parent.
  //
  int sharedcircuit::slots() const {
    return net_->n();
  }

  //
  // My event manager.
  //
  void sharedcircuit::set_eventmanager(const boost::shared_ptr<eventmanager> &mgr) {
    mgr_ = mgr;
    stopper_->set_eventmanager(mgr);
  }

  //
  // Send the given message at the given time.
  //
  void sharedcircuit::send(const timestamp &start_time,
                           int64_t bytes,
                           nodeid sender,
                           nodeid recver,
                           const boost::shared_ptr<eventhandler> &sendhandler,
                           const boost::shared_ptr<eventhandler> &recvhandler)
  {
    // Decide what the latency will be for this transfer.
    data payload(bytes, data::Bytes);
    timestamp latency = latency_->latency(payload);
    timestamp halflat = latency / 2;
    timestamp start_at(start_time + halflat);
    sstmac::walk the_walk;
    the_walk.ev().reserve(net_->diameter());
    net_->route(sender.id, recver.id, the_walk);

    for(int i = 0; i < the_walk.length(); ++i)
      if(i > 0)
        if(the_walk.e(i).from() != the_walk.e(i-1).to())
          throw illformederror("sharedcircuit::send:  Non-contigous walk.");

    // Get the edge indices.
    const std::vector<sstmac::lwedge> &ev = the_walk.ev();
    std::vector<edge_id_t> edge_id(ev.size());
    for(size_t i = 0; i < ev.size(); ++i) {
      //edge_id.at(i) = ev.at(i).id();  // THIS YIELDS GARBAGE.
      edge_id.at(i) = net_->g().edge_to_index(ev.at(i)).value();
    }
    shared_ptr<starter> trans =
      starter::construct(flower_, stopper_, bytes, halflat,
                         edge_id, sendhandler, recvhandler);

    mgr_->schedule(start_at, trans);
  }

} // end of namespace sstmac
