/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/simulator/simplenode.h>
#include <sstmac/simulator/simplenode-netcall.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/simulator/networkcall.h>
#include <sstmac/processor/instructions.h>
#include <sstmac/simulator/networkdata.h>
#include <sstmac/simulator/nodeinterface.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/process/server.h>
#include <sstmac/processor/procscheduler.h>
#include <sstmac/util/errors.h>
#include <sstmac/util/debug.h>
#include <iostream>
#include <sstream>

namespace sstmac {

  using boost::shared_ptr;
  using boost::weak_ptr;

  //
  // Construct a new node.
  //
  simplenode::simplenode(const timestamp &latency, double memory_bandwidth, int n_core) :
    memlatency_(latency), membandwidth_(memory_bandwidth), n_core_(n_core)
  {
    scheduler_ = procscheduler::construct(n_core);
  }

  //
  // Construct a new node.
  //
  simplenode::simplenode(double memory_bandwidth, int n_core) :
    memlatency_(0), membandwidth_(memory_bandwidth), n_core_(n_core)
  {
    scheduler_ = procscheduler::construct(n_core);
  }

  //
  // Who am I?
  //
  void simplenode::set_self(const boost::shared_ptr<simplenode> &self) {
    if(self.get() != this) {
      throw valueerror("simplenode::set_self:  'self' pointer does not refer "
                       "to 'this' object.");
    }
    self_ = self;
  }

  //
  // This is who I am.
  //
  boost::shared_ptr<simplenode> simplenode::get_self() {
    return self_.lock();
  }

  //
  // Construct a new node with the given memory latency and bandwidth.
  //
  boost::shared_ptr<simplenode>
  simplenode::construct(const timestamp &latency, double bandwidth, int n_core) {
    shared_ptr<simplenode> retval(new simplenode(latency, bandwidth, n_core));
    retval->set_self(retval);
    return retval;
  }

  //
  // Construct a new node with the given memory bandwidth.
  //
  boost::shared_ptr<simplenode> simplenode::construct(double memory_bandwidth,
                                                      int n_core) {
    shared_ptr<simplenode> retval(new simplenode(memory_bandwidth,n_core));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  simplenode::~simplenode() throw() {
  }

  //
  // Get a new node of the same type as the current one.
  //
  boost::shared_ptr<node> simplenode::clone() const {
    if(debugger::active()) debug() << "simplenode::clone()\n";
    shared_ptr<simplenode> retval(new simplenode(this->membandwidth_,
                                                 this->n_core_));
    retval->server_.clear();
    //retval->iface_.clear();
    retval->set_self(retval);
    return retval;
  }

  //
  // Setup.
  //
  void simplenode::initialize(nodeid my_id,
                              const boost::shared_ptr<networkcall> &net,
                              const boost::shared_ptr<eventmanager> &mgr)
  {
    if(debugger::active()) debug() << "simplenode::initialized(" << my_id << ", " << net << ", "
            << mgr << ")\n";
    if(net == 0) {
      throw nullerror("simplenode::initialize:  null network pointer.");
    }
    if(mgr == 0) {
      throw nullerror("simplenode::initialize:  null eventmanager pointer.");
    }
    id_ = my_id;
    net_ = net;
    mgr_ = mgr;
    scheduler_->set_manager(mgr);
  }

  //
  // Current time.
  //
  const timestamp& simplenode::now() const {
    return mgr_->now();
  }

  //
  // Access a node interface for the given ppid.
  //
  shared_ptr<nodeinterface>
  simplenode::get_nodeinterface(const ppid &pid, const taskid &tid) {
    return nodeinterface::construct(self_.lock(), pid, tid);
  }

  //
  // Test whether the given ppid has a server.
  //
  bool simplenode::has_server(const ppid &caller) const {
    return (server_.find(caller) != server_.end());
  }

  //
  // Define a server for the given ppid.
  //
  void simplenode::set_server(ppid caller, const shared_ptr<server> &serv) {
    if(this->has_server(caller)) {
      throw illformederror("simplenode::set_server:  Caller already has "
                           "a server defined.");
    }
    serv->initialize(this->get_nodeinterface(caller, taskid(-1)),
                     netcall::construct(self_.lock(), caller));
    server_[caller] = serv;
  }

  //
  // Access the server for the given ppid.
  //
  shared_ptr<server> simplenode::get_server(const ppid &caller) {
    server_t::iterator it = server_.find(caller);
    if(it == server_.end()) {
      throw illformederror("simplenode::get_server:  No server defined "
                           "for the given ppid.");
    }
    return it->second;
  }

  //
  // Test whether this node has a processor.
  //
  bool simplenode::has_processor() const {
    return scheduler_->has_processor();
  }

  //
  // Define a processor.
  //
  void simplenode::set_processor(const proc_t &proc) {
    scheduler_->set_processor(proc);
  }

  //
  // Access the processor.
  //
  const shared_ptr<processor> simplenode::get_processor() const {
    return scheduler_->get_processor();
  }

  //
  // Set a simple trigger (wake-up call).
  //
  void simplenode::schedule(const timestamp &start_time,
                            ppid caller,
                            const boost::shared_ptr<eventhandler> &completion)
  {
    if(debugger::active()) debug() << "simplenode::schedule(" << start_time
                                   << ", " << caller << ", " << completion
                                   << ")\n";
    mgr_->schedule(start_time, completion);
  }

  //
  // Pretend to compute.
  //
  void simplenode::compute(const timestamp &start_time,
                           ppid caller,
                           const timestamp &cpu_time,
                           const boost::shared_ptr<eventhandler> &completion)
  {
    compute(start_time, caller, cpu_time, 1, completion);
  }

  //
  // Pretend to compute.
  //
  void simplenode::compute(const timestamp &start_time,
                           ppid caller,
                           const timestamp &cpu_time_per_thread,
                           int threads,
                           const boost::shared_ptr<eventhandler> &completion)
  {
    static bool first_entry = true;
    if(debugger::active()) debug() << "simplenode::compute(" << start_time
                                   <<", " <<caller << ", "
                                   << cpu_time_per_thread << ", "
                                   << threads << ", " << completion <<")\n";
    if(threads < 1)
      throw valueerror("simplenode::compute:  Can't compute on less than "
                       "one thread.");
    // FIXME:   Implement load monitor.
    if(first_entry) {
      std::cerr << "WARNING:  simplenode::compute:  "
                << "Load monitor not implemented\n";
      first_entry = false;
    }
    mgr_->schedule(start_time+cpu_time_per_thread, completion);
  }

  //
  // Compute request
  //
  void simplenode::compute(const timestamp  &start_time,
                           ppid caller,
                           const boost::shared_ptr<instructions> &instructions,
                           const boost::shared_ptr<eventhandler> &completion )
  {
    if(debugger::active()) debug() << "simplenode::compute(" << start_time << ", " 
      << caller << ", " << instructions << ", " << completion << ")\n";
    scheduler_->compute(start_time,instructions,completion);

    // old stuff
    //timestamp ctime = 
    //  processor_->compute(instructions);
    //mgr_->schedule( start_time + ctime, completion );
  }

  //
  // Pretend to copy data.
  //
  void simplenode::copy(const timestamp &start_time,
                        ppid p,
                        int64_t bytes,
                        const boost::shared_ptr<eventhandler> &completion)
  {
    if(debugger::active()) debug() << "simplenode::copy(" << start_time << ", " << p << ", "
            << bytes << ", " << completion << ")\n";
    if(bytes < 0) {
      throw valueerror("simplenode::copy:  negative number of bytes.");
    }
    timestamp copylag = memlatency_ + this->copycost(bytes);
    mgr_->schedule(start_time + copylag, completion);
  }

  //
  // Complete a receive.
  //
  void simplenode::incoming_message(const timestamp &current_time,
                                    ppid caller,
                                    nodeid sender,
                                    const boost::shared_ptr<networkdata> &load)
  {
    if(debugger::active()) debug() << "simplenode::incoming_message(" << current_time << ", "
            << caller << ", " << sender << ", " << load << ")\n";
    server_t::iterator it = server_.find(caller);
    if(it == server_.end()) {
      std::ostringstream buf;
      buf << "simplenode::incoming_message:  No server defined for " << caller;
      throw valueerror(buf.str());
    }
    it->second->incoming_message(current_time, sender, load);
  }

  //
  // Schedule a send operation to happen at the given time.
  //
  void simplenode::send(const timestamp &start_time,
                        ppid process,
                        nodeid recver,
                        const boost::shared_ptr<networkdata> &payload,
                        const boost::shared_ptr<eventhandler> &completion)
  {
    if(debugger::active()) debug() << "simplenode::send(" << start_time << ", " << process << ", "
            << recver << ", " << payload << ", " << completion << ")\n";
    if(payload == 0) {
      throw nullerror("simplenode::send:  null payload pointer.");
    }
    timestamp copylag = this->copycost(payload->bytes());
    net_->send(start_time+copylag, process, recver, payload, completion);
  }

  //
  // Private method to estimate copy cost.
  //
  timestamp simplenode::copycost(int64_t bytes) const {
    if(debugger::active()) debug() << "simplenode::copycost(" << bytes<< ")\n";
    if(bytes < 0) {
      throw valueerror("simplenode::copycost:  negative number of bytes.");
    }
    if(membandwidth_ > 0) {
      timestamp copytime(bytes / membandwidth_);
      return copytime;
    }
    else {
      return timestamp(0);
    }
  }

} // end of namespace sstmac.
