/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#define _ISOC99_SOURCE // llabs was added in C99
#include <sstmac/simulator/timestamp.h>
#include <sstmac/util/errors.h>
#include <iostream>
#include <iomanip>
#include <limits>
#include <stdio.h>
#include <math.h>

namespace sstmac {

  //
  // Private method to get around the fact that GMP doesn't support
  // assignment/construction from long long.
  //
  inline timestamp::tick_t to_tick(int64_t value) {
#ifdef SSTMAC_USE_GMPXX
    if(sizeof(int64_t) > sizeof(long)) {
      // Mask with low 31 bits set.
      static const int64_t mask = (uint64_t(1) << 31) - 1;
      const int32_t lowbits = int32_t(value & mask);
      const int32_t highbits = value >> 31;
      timestamp::tick_t retval(highbits);
      retval = retval << 31;
      retval = retval + lowbits;
      return retval;
    }
    else {
      return timestamp::tick_t(long(value));
    }
#else
    return timestamp::tick_t(value);
#endif
  }

  //
  // Static variables.
  //
  timestamp::tick_t *timestamp::psec_tick_spacing_ = NULL;;

  std::string *timestamp::tick_spacing_string_ = NULL;

  // Inline static members.
  inline void timestamp::init_stamps() {
    if(! psec_tick_spacing_) {
      psec_tick_spacing_ = new tick_t(1);
      tick_spacing_string_ = new std::string("1 ps");
    }
  }

  // Utility func. to avoid endless ifndef SSTMAC_USE_GMPXX calls.
#ifndef SSTMAC_USE_GMPXX
#define TOD(D) double(D)
#else
  inline double TOD(const timestamp::tick_t &t) { return t.get_d(); }
#endif
  
  //
  // Round the given time value to the nearest representable internal time.
  //
  timestamp::timestamp(double t)  {
    init_stamps();
#ifndef SSTMAC_USE_GMPXX
    ticks_ = int64_t((t * 1e12) / (*psec_tick_spacing_));
    if((t >= max_time()) || (t <= min_time())) {
      std::cerr << "timestamp():  Time value " << t << " is outside the range "
                << max_time() << ".." << min_time() << " available with "
                << *psec_tick_spacing_ << " picosecond clock resolution.\n";
      throw valueerror("timestamp(double):  Time value out of bounds.");
    }
#else
    ticks_ = t * 1e12;
    ticks_ /= *psec_tick_spacing_;
#endif
  }

  //
  // Assignment.
  //
  //timestamp& timestamp::operator=(double v) {
  //  ticks_ = (int64_t((v * 1e12) / *psec_tick_spacing_));
  //  return *this;
  //}

  //
  // Return the current time in seconds.
  //
  double timestamp::sec() const {
    init_stamps();
    return (TOD(ticks_) * 1e-12) * TOD(*psec_tick_spacing_);
  }

  //
  // Return the current time in milliseconds.
  //
  double timestamp::msec() const {
    init_stamps();
    return (TOD(ticks_) * 1e-9) * TOD(*psec_tick_spacing_);
  }

  //
  // Return the current time in microseconds.
  //
  double timestamp::usec() const {
    init_stamps();
    return (TOD(ticks_) * 1e-6) * TOD(*psec_tick_spacing_);
  }

  //
  // Return the current time in nanoseconds.
  //
  double timestamp::nsec() const {
    init_stamps();
    return (TOD(ticks_) * 1e-3) * TOD(*psec_tick_spacing_);
  }

  //
  // Return the current time in picoseconds.
  //
  double timestamp::psec() const {
    init_stamps();
    return TOD(ticks_) * TOD(*psec_tick_spacing_);
  }

  //
  // static:  Get the tick interval.
  //
  timestamp::tick_t timestamp::tick_interval() {
    init_stamps();
    return *psec_tick_spacing_;
  }

  //
  // Get the tick interval in string form (for example, "1ps").
  //
  const std::string& timestamp::tick_interval_string() {
    init_stamps();
    return *tick_spacing_string_;
  }

  //
  // static:  Get the number of ticks per second (1/tick_interval()).
  //
  timestamp::tick_t timestamp::frequency() {
    init_stamps();
    return (tick_t(1e12) / *psec_tick_spacing_);
  }

  //
  // static:  Get the largest time value possible (in seconds).
  //
  double timestamp::max_time() {
    init_stamps();
#ifndef SSTMAC_USE_GMPXX
    return (std::numeric_limits<int64_t>::max() * 1e-12) * *psec_tick_spacing_;
#else
    return INFINITY;
#endif
  }

  //
  // static:  Get the smallest (most negative) time value possible (in seconds).
  //
  double timestamp::min_time() {
#ifndef SSTMAC_USE_GMPXX
    return (std::numeric_limits<int64_t>::min() * 1e-12) * *psec_tick_spacing_;
#else
    return -INFINITY;
#endif
  }

  //
  // static:  Get a time value corresponding exactly to the given number of 
  // picoseconds.
  //
  timestamp timestamp::exact_psec(int64_t psec) {
    return scaled_time(psec, 1, "timestamp::exact_psec", "picoseconds");
  }

  //
  // static:  Get a time value corresponding exactly to the given number of 
  // nanoseconds.
  //
  timestamp timestamp::exact_nsec(int64_t nsec) {
    return scaled_time(nsec, 1e3, "timestamp::exact_nsec", "nanoseconds");
  }

  //
  // static:  Get a time value corresponding exactly to the given number of 
  // microseconds.
  //
  timestamp timestamp::exact_usec(int64_t usec) {
    return scaled_time(usec, 1e6, "timestamp::exact_usec", "microseconds");
  }

  //
  // static:  Get a time value corresponding exactly to the given number of 
  // milliseconds.
  //
  timestamp timestamp::exact_msec(int64_t msec) {
    return scaled_time(msec, 1e9, "timestamp::exact_msec", "milliseconds");
  }

  //
  // static:: Get a time value corresponding exactly to the given number of 
  // seconds.
  //
  timestamp timestamp::exact_sec(int64_t sec) {
    return scaled_time(sec, 1e12, "timestamp::exact_sec", "seconds");
  }

  //
  // Static:  Get scaled time.
  //
  timestamp timestamp::scaled_time(int64_t value, int64_t scaling,
                                   const char *caller, const char *units)
  {
    init_stamps();
#ifndef SSTMAC_USE_GMPXX
    static const int64_t max = std::numeric_limits<int64_t>::max();
    static const int64_t min = std::numeric_limits<int64_t>::min();
    if(((value/(*psec_tick_spacing_)) > (max/scaling)) ||
       ((value/(*psec_tick_spacing_)) < (min/scaling)))
      {
        std::cerr << caller << ":  The time value " << value << " " << units
                  << " is outside the range of "
                  << min << " through " << max << " picoseconds\n";
          throw valueerror("timestamp::scaled_time:  Value out of range.");
      }
#endif // ! SSTMAC_USE_GMPXX
    timestamp retval(0);
    retval.ticks_ =
      ((to_tick(value) * to_tick(scaling)) / *psec_tick_spacing_);
    return retval;
  }

  //
  // Add.
  //
  timestamp& timestamp::operator+=(const timestamp &other) {
#ifndef SSTMAC_USE_GMPXX
    double approximate = this->sec() + other.sec();
    if(approximate < min_time() || approximate > max_time()) {
      std::cerr << "operator+=:  Result " << approximate
                << " is out of range " << min_time() << ".."
                << max_time() << "\n";
      throw timeerror("timestamp::operator+=():  Result out of range.");
    }
#endif
    ticks_ += other.ticks_;
    return *this;
  }

  //
  // Subtract.
  //
  timestamp& timestamp::operator-=(const timestamp &other) {
#ifndef SSTMAC_USE_GMPXX
    double approximate = this->sec() - other.sec();
    if(approximate < min_time() || approximate > max_time()) {
      std::cerr << "operator-=:  Result " << approximate
                << " is out of range " << min_time() << ".."
                << max_time() << "\n";
      throw timeerror("timestamp::operator-=():  Result out of range.");
    }
#endif
    ticks_ -= other.ticks_;
    return *this;
  }

  //
  // Multiply.
  //
  timestamp& timestamp::operator*=(double scale) {
#ifndef SSTMAC_USE_GMPXX
    double approximate = this->sec() * scale;
    if(approximate < min_time() || approximate > max_time()) {
      std::cerr << "operator*=:  Result " << approximate
                << " is out of range " << min_time() << ".."
                << max_time() << "\n";
      throw timeerror("timestamp::operator*=():  Result out of range.");
    }
    ticks_ *= scale;
#else
    // gmpxx has stupid promotion rules; using operator*= will
    // promote (truncate) the floating point value to an integer
    // and then perform an integer multiply (unfortunately, this can come
    // with a major rounding error).
    ticks_ = ticks_.get_d() * scale;
#endif
    return *this;
  }

  //
  // Divide.
  //
  timestamp& timestamp::operator/=(double scale) {
#ifndef SSTMAC_USE_GMPXX
    double approximate = this->sec() / scale;
    if(approximate < min_time() || approximate > max_time()) {
      std::cerr << "operator/=:  Result " << approximate
                << " is out of range " << min_time() << ".."
                << max_time() << "\n";
      throw timeerror("timestamp::operator/=():  Result out of range.");
    }
    ticks_ /= scale;
#else
    // Same stupid gmpixx rounding rules as with multiply.
    ticks_ = ticks_.get_d() / scale;
#endif
    return *this;
  }

  timestamp operator+(const timestamp &a, const timestamp &b) {
    timestamp rv(a);
    rv += b;
    return rv;
  }

  timestamp operator-(const timestamp &a, const timestamp &b) {
    timestamp rv(a);
    rv -= b;
    return rv;
  }

  timestamp operator*(const timestamp &t, double scaling) {
    timestamp rv(t);
    rv *= scaling;
    return rv;
  }

  timestamp operator*(double scaling, const timestamp &t) {
    return (t * scaling);
  }
    
  timestamp operator/(const timestamp &t, double scaling) {
    timestamp rv(t);
    rv /= scaling;
    return rv;
  }

  std::ostream& operator<<(std::ostream &os, const timestamp &t) {
    timestamp::tick_t psec = t.ticks() / t.tick_interval();
    timestamp::tick_t frac = psec % timestamp::tick_t(1e12);
    timestamp::tick_t secs = psec / timestamp::tick_t(1e12);
    os << "timestamp(" << secs << "." 
       << std::setw(12) << std::setfill('0') << frac << " sec)";
    return os;
  }


} // end of namespace sstmac.
