/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SKELETON_ACTORMATMUL_ACTORMATMUL_MESSAGES_H
#define SSTMAC_SKELETON_ACTORMATMUL_ACTORMATMUL_MESSAGES_H

#include <sstmac/skeleton/actormatmul/actormatmul.h>
#include <sstmac/process/actorpattern.h>
#include <sstmac/process/actormsg.h>

namespace sstmac {

  /**
   * Private message base for all actormatmul messages.
   */
  class actormatmul::base : public actormsg {
  public:
    virtual ~base() {}
    using actormsg::actorclone;
    virtual void handle(actorid caller, const boost::shared_ptr<actormatmul>& owner) = 0;
  };

  /**
   * Private message type to indicate that a local computor has completed work.
   */
  class actormatmul::localstore : public base {
    int iteration_;
    localstore(int iteration) : iteration_(iteration) {}
  public:
    static boost::shared_ptr<localstore> construct(int iteration) {
      return boost::shared_ptr<localstore>(new localstore(iteration));
    }
    virtual ~localstore() {}
    virtual boost::shared_ptr<actormsg> actorclone() const {
      return construct(iteration_);
    }
    virtual void handle(actorid id, const boost::shared_ptr<actormatmul>& owner) {
      owner->local_computor_done(id, iteration_);
    }
  };

  /**
   * Private message type to indicate that a remote computor has completed work.
   */
  class actormatmul::remotestore : public base {
    int iteration_;
    remotestore(int iteration) : iteration_(iteration) {}
  public:
    static boost::shared_ptr<remotestore> construct(int iteration) {
      return boost::shared_ptr<remotestore>(new remotestore(iteration));
    }
    virtual ~remotestore() {}
    virtual boost::shared_ptr<actormsg> actorclone() const {
      return construct(iteration_);
    }
    virtual void handle(actorid id, const boost::shared_ptr<actormatmul>& owner) {
      owner->remote_computor_done(id, iteration_);
    }
  };

  /**
   * Private message type to indicate arrival of work for a computor.
   */
  class actormatmul::work : public actormsg {
    int64_t matdim_;
    int iteration_;
    actorid store_to_;
    work(int64_t matdim, int iteration, const actorid &store_to) :
      matdim_(matdim), iteration_(iteration), store_to_(store_to)
    {}
  public:
    static boost::shared_ptr<work>
    construct(int64_t matdim, int iteration, const actorid &store_to) {
      return boost::shared_ptr<work>(new work(matdim, iteration, store_to));
    }
    virtual ~work() {}
    virtual boost::shared_ptr<actormsg> actorclone() const { 
      return construct(matdim_, iteration_, store_to_);
    }
    int64_t matdim() const { return matdim_; }
    int iteration() const { return iteration_; }
    const actorid& store_to() const { return store_to_; }
  };

  /**
   * Private message type to indicate request to fetch data.
   */
  class actormatmul::getdata : public actormatmul::base {
    int iteration_;
    getdata(int iteration) : iteration_(iteration) {}
  public:
    static boost::shared_ptr<getdata> construct(int iteration) {
      return boost::shared_ptr<getdata>(new getdata(iteration));
    }
    virtual ~getdata() {}
    virtual boost::shared_ptr<actormsg> actorclone() const {
      return construct(iteration_);
    }
    virtual void handle(actorid caller, const boost::shared_ptr<actormatmul>& owner) {
      owner->get_data(caller, iteration_);
    }
  };

  /**
   * Private message type to indicate arrival of data.
   */
  class actormatmul::putdata : public actormatmul::base {
    int iteration_;
    putdata(int iteration) : iteration_(iteration) {}
  public:
    static boost::shared_ptr<putdata> construct(int iteration) {
      return boost::shared_ptr<putdata>(new putdata(iteration));
    }
    virtual ~putdata() {}
    virtual boost::shared_ptr<actormsg> actorclone() const {
      return construct(iteration_);
    }
    virtual void handle(actorid caller, const boost::shared_ptr<actormatmul>& owner) {
      owner->put_data(caller, iteration_);
    }
  };

  /**
   * Private message type to update lower neighbor index.
   */
  class actormatmul::lowernodeupdate : public actormatmul::base {
    int new_flat_lower_;
    lowernodeupdate(int flatid) : new_flat_lower_(flatid) {}
  public:
    static boost::shared_ptr<lowernodeupdate> construct(int flatid) {
      return boost::shared_ptr<lowernodeupdate>(new lowernodeupdate(flatid));
    }
    virtual ~lowernodeupdate() {}
    virtual boost::shared_ptr<actormsg> actorclone() const {
      return construct(new_flat_lower_);
    }
    virtual void handle(actorid, const boost::shared_ptr<actormatmul>& owner) {
      owner->update_lower_neighbor(new_flat_lower_);
    }
  };

  /**
   * Private message type to update upper neighbor index.
   */
  class actormatmul::uppernodeupdate : public actormatmul::base {
    int new_flat_upper_;
    uppernodeupdate(int flatid) : new_flat_upper_(flatid) {}
  public:
    static boost::shared_ptr<uppernodeupdate> construct(int flatid) {
      return boost::shared_ptr<uppernodeupdate>(new uppernodeupdate(flatid));
    }
    virtual ~uppernodeupdate() {}
    virtual boost::shared_ptr<actormsg> actorclone() const {
      return construct(new_flat_upper_);
    }
    virtual void handle(actorid, const boost::shared_ptr<actormatmul>& owner) {
      owner->update_upper_neighbor(new_flat_upper_);
    }
  };

  /**
   * Private message type to donate a computor.
   */
  class actormatmul::donatecomputor : public actormatmul::base {
    actorid thread_;
    donatecomputor(actorid thread) : thread_(thread) {}
  public:
    static boost::shared_ptr<donatecomputor> construct(actorid thread) {
      return boost::shared_ptr<donatecomputor>(new donatecomputor(thread));
    }
    virtual ~donatecomputor() {}
    virtual boost::shared_ptr<actormsg> actorclone() const {
      return construct(thread_);
    }
    virtual void handle(actorid, const boost::shared_ptr<actormatmul>& owner) {
      owner->remote_computor_available(thread_);
    }
  };

  /**
   * Private message type to indicate that we're all done.
   */
  class actormatmul::done : public actormatmul::base {
    done() {}
  public:
    static boost::shared_ptr<done> construct() {
      return boost::shared_ptr<done>(new done);
    }
    virtual ~done() {}
    virtual boost::shared_ptr<actormsg> actorclone() const { return construct(); }
    virtual void handle(actorid caller, const boost::shared_ptr<actormatmul>& owner) {
      owner->all_done(caller);
    }
  };


} // end of namespace sstmac

#endif // ! SSTMAC_SKELETON_ACTORMATMUL_ACTORMATMUL_MESSAGES_H
