/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SKELETON_ACTORMATMUL_ACTORMATMUL_H
#define SSTMAC_SKELETON_ACTORMATMUL_ACTORMATMUL_H

#include <sstmac/process/simplemboxapp.h>
#include <string>
#include <vector>
#include <map>
#include <stdint.h>

namespace sstmac {

  /**
   * Partition a matrix-matrix multiply using the actor model.
   *
   * Variation on the actormult2 implementation to allow prefetching
   * instead of requiring all threads to run out of work.
   */
  class actormatmul : public simplemboxapp {
    /// The label used for our global counter.
    static const std::string label_;

    /// Private message types.
    class base;
    class getdata;
    class putdata;
    class localstore;
    class remotestore;
    class lowernodeupdate;
    class uppernodeupdate;
    class donatecomputor;
    class done;
    class work;

    /// Private class types.
    class computor;

    /// This object.
    boost::weak_ptr<actormatmul> self_;

    /// The constructor values
    /// Are we prefetching matrix subblocks (you probably should).
    bool use_prefetch_;
    /// The total matrix dimension.
    int64_t fullmatrix_dimension_;
    /// The number of computors (each manages one or more threads)
    /// owned by this actormatmul.
    int64_t computors_;
    /// Control debugging verbosity.
    int chatty_;

    /// Guesstimate of how long each multiply-add takes (wall time elapsed).
    timestamp time_per_FMA_;
    /// Option for slowing down a single node (blunt instrument).
    int64_t degraded_id_;
    /// The minimum number of work-units-per-thread for each submatrix
    /// multiply.
    double granularity_factor_;
    /// The number of threads managed by each computor.
    int threads_per_computor_;

    /// The dimension of the submatrix on each node (rows or columns).
    int64_t submatrix_dimension_;
    /// The dimension of each work-unit submatrix (we might want to make this
    /// a more adaptive value).
    int64_t workmatrix_dimension_;
    /// The number of splits needed to divide submatrix_dimension_ into
    /// workmatrix_dimension_ subblocks.
    int64_t workmatrix_splits_;

    /// Information on our location in the world and the nodes we
    /// will communicate with.
    /// Gets queried from the queue when run() starts and then never change.
    /// The flat rank (similar to MPI rank) for this node.
    int flat_rank_;
    /// The flat rank of our "upper" neighbor.  When the upper neighbor runs
    /// out of work, we update this to be their upper neighbor.
    int flat_upper_;
    /// The flat rank of our "lower" neighbor.  When the lower neighbor runs
    /// out of work, we update this to be their lower neighbor.
    int flat_lower_;
    /// The total number of actormatmul players in our world.
    int nodes_;
    /// The square root of nodes_.
    int sqrtnodes_;

    /// Try to make managing the prefetch queues a little more sane.
    struct evaluation {
      /// Have we tried to fetch these data blocks yet?
      bool active;
      /// How many data blocks do we need to receive before we can start?
      int pending_arrivals;
      /// Once data is fetched, how many more blocks are left to process?
      int64_t remaining;
      evaluation() : active(false), pending_arrivals(2), remaining(-1) {}
    };

    /// Mutable state
    /// The current local iteration.
    int current_iteration_;
    /// The iteration we are fetching for.
    int prefetch_for_iteration_;
    /// The identifiers for local computors.
    /// A local computor that has been "donated" to another node is still
    /// stored here (so we can terminate it when all work is done).
    /// This is also useful for detecting "donation loops' -- if a computor
    /// that we donated has somehow managed to make it all the way back here.
    typedef std::set<actorid> local_computors_t;
    local_computors_t local_computors_;
    /// The local computors that are not working (pending prefetch).
    std::vector<actorid> idle_local_computors_;
    /// Remote computors that we haven't been able to feed yet (e.g.
    /// in case of a pending prefetch).  When we donate a remote computor on
    /// (or drop it), it is simply erased from this list (the "owner" matmul
    /// is responsible for terminating the computors when all work is done).
    /// If all local work is done, these computors should all be donated
    /// and cleared out.
    std::vector<actorid> idle_remote_computors_;
    /// The local evaluations remaining.  Stored on a per-evaluation-block basis
    std::vector<evaluation> remaining_;
    /// Total number of remaining stores
    int64_t remaining_stores_;
    /// Local flag to indicate whether local work is done.
    /// Mainly serves to know whether this node is the "last one out"
    /// in a distributed run.
    bool done_;
    
  private:
    /// Called by 'getdata' message when somebody requests data.
    void get_data(actorid requestor, int iteration);

    /// Called by 'putdata' message when somebody delivers data.
    void put_data(actorid donor, int iteration);
   
    /// Test whether we've hit the watermark where we no longer
    /// accept remote worker computors.
    void test_watermark();

    /// Called by 'localstore' message when a computor completes a unit of work.
    /// The first call to this method is a "check-in" by a computor
    /// indicating that it is ready to start working.
    void local_computor_done(actorid caller, int iteration);

    /// Called by 'remotestore' message when a remote computor completes work.
    /// Any call to this method indicates that a unit of work is done.
    void remote_computor_done(actorid caller, int iteration);

    /// Called by 'computordonate' when we are being given control over
    /// remote worker computors.
    void remote_computor_available(actorid caller);

    /// Called by 'lowernodeupdate' when a neighbor node has run out of
    /// local work.
    void update_lower_neighbor(int new_flat_lower);

    /// Called by 'uppernodeupdate' when a neighbor node has run out of
    /// local work.
    void update_upper_neighbor(int new_flat_upper);

    /// Called by 'done' message to indicate all work is over.
    void all_done(actorid sender);

    /// Called to start any idle computors (work has arrived).
    void alert_computors();

    /// Called to request next block of data.
    void request_data();

    /// Called at the top of run() to initialize data.
    void init();

    /// Real constructor.
    actormatmul(bool prefetch, int64_t matrix, int64_t computors, int chatty,
                const timestamp &time_per_FMA, int64_t degraded_id,
                double granularity_factor, int threads_per_computor);

  public:
    /// Construct a new object.
    /// \param matrix    matrix sizes (rows or columns)
    /// \param computors   evaluation computors per node.
    /// \param chatty:  If 0 no debug output, if > 0 only node 0 chats,
    ///                 and if < 0 all nodes chat
    /// \param time_per_FMA:  How much CPU time we assume each FMA takes.
    /// \param degraded_id:   Any process running on a node with the given
    ///                       flat rank will run at half performance
    /// \param granularity_factor:  Ensure that the number of matrix-matrix
    ///                 multiplies on each node is at least this times
    ///                 as great as the number of local computors.
    ///                 It may make sense to make this number reasonably high
    ///                 (4 or greater) to allow for efficient load balancing.
    /// \param threads_per_computor:  Each computor will manage the given
    ///                 number of (implicit) threads.
    static boost::shared_ptr<actormatmul>
    construct(bool prefetch, int64_t matrix, int64_t computors, int chatty = 0,
              const timestamp &time_per_FMA = timestamp::exact_psec(250),
              int64_t degraded_id = -1,
              double granularity_factor = 4.0,
              int threads_per_computor = 1);

    virtual ~actormatmul();

    /// Get a fresh clone of this object.  Does not include the state
    /// of any runtime variables.
    boost::shared_ptr<simplemboxapp> clone() const;

    /// Start the evaluation.
    void run();
  };

} // end of namespace sstmac

#endif
