/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/skeleton/mpiparsers/rsqallocation.h>
#include <sstmac/skeleton/mpiparsers/rsqmetatrace.h>
#include <sstmac/network/torus.h>
#include <sstmac/util/errors.h>
#include <boost/lexical_cast.hpp>
#include <map>

namespace sstmac {

  using boost::shared_ptr;
  using boost::lexical_cast;
  using sstmac::torus;
  using sstmac::vertex;

  //
  // A new allocation strategy.
  //
  rsqallocation::rsqallocation(const shared_ptr<rsqmetatrace>& meta,
                               const shared_ptr<torus>& tor) :
    meta_(meta), torus_(tor)
  {
    if(meta_.get() == NULL)
      throw nullerror("rsqallocation(rsqmetatrace, torus):  "
                      "NULL rsqmetatrace pointer.");
    if(torus_.get() == NULL)
      throw nullerror("rsqallocation(rsqmetatrace, torus):  "
                      "NULL torus pointer.");
    if(torus_->dimensions().size() < 2)
      throw valueerror("rsqallocation(rsqmetatrace, torus):  "
                       "Torus dimension less than 2");
  }

  //
  // Hello.
  //
  shared_ptr<rsqallocation>
  rsqallocation::construct(const shared_ptr<rsqmetatrace>& meta,
                           const shared_ptr<torus>& tor)
  {
    return shared_ptr<rsqallocation>(new rsqallocation(meta, tor));
  }

  //
  // Goodbye.
  //
  rsqallocation::~rsqallocation() throw() {
  }

  //
  // The main allocate call.
  //
  void rsqallocation::allocate(int nnode,
                               std::vector<nodeid> &allocation,
                               std::set<nodeid> &available,
                               std::set<nodeid> &allocated)
  {
    static const std::string here("rsqallocation::allocate");
    /// Check our torus.
    const std::vector<int> &tdim = torus_->dimensions();

    // DEBUG:
#ifdef DEBUG
    do {
      std::set<nodeid>::iterator it, end = available.end();
      for(it = available.begin(); it != end; ++it) {
        std::cerr << *it << " is available at vertex index "
                  << torus_->index_to_node(it->id) << "\n";
      }
    } while(0);
#endif // ! DEBUG

    vertex crud = torus_->index_to_node(available.begin()->id);
    const size_t vertdim = crud.v().size();
    /// Number of leading dimensions that we set to zero.
    size_t blankdims = tdim.size() - 2;
    /// Figure out which nodes we need from the torus.
    nodeid nodecnt = meta_->nodes();
    typedef std::map<rsqcrd, nodeid> need_t;
    need_t need;
    for(nodeid id = nodeid(0); id < nodecnt; ++id) {
      vertex vtx(vertdim);
      // I am completely confused on the node indexing in a 2D torus --
      // this is a horrible hack.
      rsqcrd crd = meta_->node_crd(id);
      vtx[2*blankdims] = crd.y;
      vtx[2*(blankdims+1)] = crd.z;
      /* Next need to convert this vertex to node index */
#ifdef DEBUG
      std::cerr << "Trying to find node index for vertex [" << vtx[0] << ", "
                << vtx[1] << ", " << vtx[2] << ", " << vtx[3] << "]\n";
#endif // ! DEBUG
      int id = torus_->node_to_index(vtx);
#ifdef DEBUG
      std::cerr << "  -- found at index " << id << "\n";
#endif // ! DEBUG
      nodeid nid(id);
#define SSTMAC_LS(VAL) lexical_cast<std::string>(VAL)
#define SSTMAC_LC(VAL) (std::string("[") + SSTMAC_LS(VAL.x) + ", " + SSTMAC_LS(VAL.y) + ", " + SSTMAC_LS(VAL.z) + "]")
#ifdef DEBUG
      std::cerr << "need[" << SSTMAC_LC(crd) << "] = " << nid << "\n";
#endif // ! DEBUG
      need[crd] = nid;
      /* get the node from the 'available' array (die horribly if not avail.) */
      std::set<nodeid>::iterator it = available.find(nid);
      if(it == available.end())
        throw valueerror(here + ":  Node " + lexical_cast<std::string>(nid)
                         + " is not in the available list.");
      allocated.insert(*it);      
      available.erase(it);
      /* finally place the node in the allocation and allocated arrays */
    }
    // Now populate the allocation array in the same order as the
    // index order to allow rsqindexing to do its work very easily
    mpiid peers = meta_->peers();
    allocation.clear();
    allocation.resize(peers.id);
    for(mpiid id = mpiid(0); id < peers; ++id) {
      const rsqcrd &crd = meta_->peer_crd(id);
      need_t::iterator it = need.find(crd);
      if(it == need.end())
        throw illformederror(here + ":  Peer " + lexical_cast<std::string>(id) +
                             " requires a node at " + SSTMAC_LC(crd) +
                             " which is not in meta->node_crd.");
      allocation.at(id.id) = it->second;
    }
  }

} // end of namespace sstmac
