/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/util/debug.h>
#include <sstmac/util/errors.h>
#include <fstream>
#include <iterator>
#include <stdio.h>
#include <errno.h>
#include <string.h>

namespace sstmac {

  // Useful stuff :)
  bool debugger::initialized = false;
  std::ofstream *debugger::nullfile = NULL;
  std::ofstream *debugger::ofile = NULL;
  std::ostream *debugger::ostr = NULL; // hack to get to_stream working
  std::ostream *debugger::nullstr = NULL; // hack to get to_stream working
  debugger *debugger::debug_ = NULL;

  //
  // Initialize shared data.
  //
  void debugger::init_debug() {
    if(! initialized) {
      initialized = true;
      if(! nullfile) {
        nullfile = new std::ofstream("/dev/null");
      }
      if(! ofile) {
        ofile = nullfile;
      }
      if(! debug_) {
        debug_ = new debugger;
      }
      nullstr = nullfile;
    }
  }

  //
  // Singleton for a debugger.
  //
  debugger& debug() {
    debugger::init_debug();
    return *debugger::debug_; 
  }

  //
  // Create a debugger.
  //
  debugger::debugger() : stream_(nullfile) {
    init_debug();
  }

  //
  // Goodbye.
  debugger::~debugger() {
  }

  //
  // Clear any calltree or preamble information.
  //
  void debugger::clear() {
    calltree_.clear();
    preamble_.clear();
  }

  //
  // Request that output be discarded.
  //
  debugger& debugger::no_debug() {
    if(stream_) *stream_ << std::flush;
    stream_ = nullfile;
    ostr = stream_;
    return *this;
  }

  //
  // Request that output be written to the given stream.
  //
  debugger& debugger::to_stream(std::ostream &os) {
    if(stream_) *stream_ << std::flush;
    stream_ = &os;
    ostr = stream_;
    return *this;
  }

  //
  // Send output to stderr.
  //
  void debugger::to_stderr() {
    this->to_stream(std::cerr);
  }

  //
  // Request that the file with the given name be created and used.
  //
  debugger& debugger::to_file(const std::string &name) {
    if(stream_) *stream_ << std::flush;
    ofile = new std::ofstream(name.c_str());
    if(! *ofile) {
      delete ofile;
      ofile = nullfile;
      // Let people know what's going on.
      FILE *fp = fopen(name.c_str(), "w");
      if(fp) {
        // Serious trouble -- ofstream::open failed and fopen worked.
        throw ioerror(std::string("SERIOUS PROBLEM:  debugger::to_file:  ") +
                      "Trying to open \"" + name +
                      "\" -- ofstream::open failed but fopen succeeded.");
      }
      throw ioerror(std::string("debugger::to_file:  Failed to open file: ") +
                    name + ":  " + strerror(errno));
    }
    stream_ = ofile;
    ostr = ofile;
    return *this;
  }

  //
  // Access the stream used by this debugger.
  //
  std::ostream& debugger::stream() {
    std::copy(preamble_.begin(), preamble_.end(),
              std::ostream_iterator<std::string>(*stream_, ": "));
    return *stream_;
  }

  //
  // Define a preamble that gets printed when stream() is called.
  //
  void debugger::push_preamble(const std::string& str) {
    preamble_.push_back(str);
  }

  //
  // Remove the topmost preamble.
  //
  void debugger::pop_preamble() {
    if(! preamble_.empty()) preamble_.pop_back();
  }

  //
  // Add an entry on the calltree graph.
  //
  void debugger::push_call(const std::string& str) {
    calltree_.push_back(str);
  }

  //
  // Remove an entry from the calltree graph.
  void debugger::pop_call() {
    if(! calltree_.empty()) calltree_.pop_back();
  }

  // Print the current calltree to the output stream.
  void debugger::dump_calltree() {
    *stream_ << "=================== Begin calltree report ===================\n";
    std::copy(calltree_.begin(), calltree_.end(),
              std::ostream_iterator<std::string>(*stream_, "\n"));
    *stream_ << "===================  End calltree  report ===================\n";
  }

  //
  // Get the calltree for your viewing pleasure.
  //
  const std::vector<std::string>& debugger::calltree() const {
    return calltree_;
  }

} // end of namespace sstmac
