/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_UTIL_DEBUG_H
#define SSTMAC_UTIL_DEBUG_H

#include <vector>
#include <iostream>

/// The SST/macro macroscale architecture simulator
namespace sstmac {

  class debugger;

  /**
   * Utility class to permit the caller to turn on and off debugging
   * and to collect a call graph (or location).
   */
  class debugger {
#ifndef SWIG // Suppress a swig warning.
    friend debugger& debug();
#endif
  private:
    /// Streams that we use for the global (static) debugger.
    static bool initialized;
    static std::ofstream *nullfile;
    static std::ofstream *ofile;
    static std::ostream *ostr; // hack to get to_stream working
    static std::ostream *nullstr; // hack to get to_stream working
    static debugger *debug_;

    /// Initialize debugger.
    static void init_debug();

    /// The stream we write to when we call stream().
    /// By default, the output goes to /dev/null.
    std::ostream *stream_;

    /// The current calltree (if you are interested).
    /// Modify this by calling push_call and pop_call.
    std::vector<std::string> calltree_;

    /// All "preambles" printed first when stream() is called.
    /// The premables are printed first, separated by ": "
    /// Modify this by calling push_preamble and pop_preamble.
    std::vector<std::string> preamble_;
    
  public:
    /// Test whether the static debugger is active.
    static bool active() {
      return (ostr != nullstr); // need ostreams for to_stream
    }

    /// Create a debugger.
    debugger();

    /// You can, of course, override this class.
    virtual ~debugger();

    /// Clear any calltree or preamble information.
    void clear();

    /// Request that output be discarded.
    debugger& no_debug();

    /// Request that output be written to the given stream.
    debugger& to_stream(std::ostream &os);

    /// Send output to stderr (python can't provide a c++ stream).
    void to_stderr();

    /// Request that the file with the given name be created
    /// and used for output.
    debugger& to_file(const std::string &name);

    /// Access the stream used by this debugger.
    std::ostream& stream();

    /// Define a preamble that gets printed when stream() is called.
    void push_preamble(const std::string&);

    /// Remove the topmost preamble.
    void pop_preamble();

    /// Add an entry on the calltree graph.
    void push_call(const std::string&);

    /// Remove an entry from the calltree graph.
    void pop_call();

    /// Print the current calltree to the output stream.
    void dump_calltree();

    /// Get the calltree for your viewing pleasure.
    const std::vector<std::string>& calltree() const;
  };

  template <typename Type>
  std::ostream& operator<<(debugger &deb, const Type &value) {
    std::ostream &str = deb.stream();
    str << value;
    str.flush(); // hack to get stream flushed when doing file io from C++
    return str;
  }


  /// The "master" debugger used to print entry information for functions
  /// and do other useful stuff.
  debugger& debug();

} // end of namespace sstmac

#endif // ! SSTMAC_UTIL_DEBUG_H
