/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


#include <sstmac/util/rng.h>
#include <string>

namespace sstmac {

  namespace RNG {

    ////////////////////////////////////////////////////////////////////
    // default seeds

    static const rngint_t defaultz = 362436069;
    static const rngint_t defaultw = 521288629;
    static const rngint_t defaultjsr=123456789;
    static const rngint_t defaultjcong=380116160;

    ////////////////////////////////////////////////////////////////////
    // UniformInteger

    UniformInteger::~UniformInteger()
    {
    }

    void
    UniformInteger::reseed()
    {
      std::vector<rngint_t> s;
      vec_reseed(s);
    }

    void
    UniformInteger::reseed(rngint_t s0)
    {
      std::vector<rngint_t> s(1);
      s[0] = s0;
      vec_reseed(s);
    }

    void
    UniformInteger::reseed(rngint_t s0,rngint_t s1)
    {
      std::vector<rngint_t> s(2);
      s[0] = s0;
      s[1] = s1;
      vec_reseed(s);
    }

    void
    UniformInteger::reseed(rngint_t s0,rngint_t s1,rngint_t s2)
    {
      std::vector<rngint_t> s(3);
      s[0] = s0;
      s[1] = s1;
      s[2] = s2;
      vec_reseed(s);
    }

    void
    UniformInteger::reseed(rngint_t s0,rngint_t s1,
                           rngint_t s2,rngint_t s3)
    {
      std::vector<rngint_t> s(4);
      s[0] = s0;
      s[1] = s1;
      s[2] = s2;
      s[3] = s3;
      vec_reseed(s);
    }

    ////////////////////////////////////////////////////////////////////
    // MWC

    MWC::MWC()
    {
    }

    boost::shared_ptr<MWC>
    MWC::construct()
    {
      boost::shared_ptr<MWC> result(new MWC);
      result->reseed();
      return result;
    }

    boost::shared_ptr<MWC>
    MWC::construct(const std::vector<rngint_t> &seeds)
    {
      boost::shared_ptr<MWC> result(new MWC);
      result->vec_reseed(seeds);
      return result;
    }

    boost::shared_ptr<MWC>
    MWC::construct(rngint_t zarg)
    {
      boost::shared_ptr<MWC> result(new MWC);
      result->reseed(zarg);
      return result;
    }

    boost::shared_ptr<MWC>
    MWC::construct(rngint_t zarg,
                   rngint_t warg)
    {
      boost::shared_ptr<MWC> result(new MWC);
      result->reseed(zarg,warg);
      return result;
    }

    MWC::~MWC()
    {
    }

    void
    MWC::vec_reseed(const std::vector<rngint_t> &seeds)
    {
      if (seeds.size() > 0) z = seeds[0];
      else                  z = defaultz;
      if (seeds.size() > 1) w = seeds[1];
      else                  w = defaultw;
    }

    int
    MWC::nseed()
    {
      return 2;
    }

    ////////////////////////////////////////////////////////////////////
    // SHR3

    SHR3::SHR3()
    {
    }

    boost::shared_ptr<SHR3>
    SHR3::construct()
    {
      boost::shared_ptr<SHR3> result(new SHR3);
      result->reseed();
      return result;
    }

    boost::shared_ptr<SHR3>
    SHR3::construct(const std::vector<rngint_t> &seeds)
    {
      boost::shared_ptr<SHR3> result(new SHR3);
      result->vec_reseed(seeds);
      return result;
    }

    boost::shared_ptr<SHR3>
    SHR3::construct(rngint_t jsrarg)
    {
      boost::shared_ptr<SHR3> result(new SHR3);
      result->reseed(jsrarg);
      return result;
    }

    SHR3::~SHR3()
    {
    }

    void
    SHR3::vec_reseed(const std::vector<rngint_t> &seeds)
    {
      if (seeds.size() > 0) jsr = seeds[0];
      else                  jsr = defaultjsr;
    }

    int
    SHR3::nseed()
    {
      return 1;
    }

    ////////////////////////////////////////////////////////////////////
    // CONG

    CONG::CONG()
    {
    }

    boost::shared_ptr<CONG>
    CONG::construct()
    {
      boost::shared_ptr<CONG> result(new CONG);
      result->reseed();
      return result;
    }

    boost::shared_ptr<CONG>
    CONG::construct(const std::vector<rngint_t> &seeds)
    {
      boost::shared_ptr<CONG> result(new CONG);
      result->vec_reseed(seeds);
      return result;
    }

    boost::shared_ptr<CONG>
    CONG::construct(rngint_t jcongarg)
    {
      boost::shared_ptr<CONG> result(new CONG);
      result->reseed(jcongarg);
      return result;
    }

    CONG::~CONG()
    {
    }

    void
    CONG::vec_reseed(const std::vector<rngint_t> &seeds)
    {
      if (seeds.size() > 0) jcong = seeds[0];
      else                  jcong = defaultjcong;
    }

    int
    CONG::nseed()
    {
      return 1;
    }

    ////////////////////////////////////////////////////////////////////
    // SimpleCombo

    SimpleCombo::SimpleCombo()
    {
      mwc_ = MWC::construct();
      cong_ = CONG::construct();
      shr3_ = SHR3::construct();
    }

    boost::shared_ptr<SimpleCombo>
    SimpleCombo::construct()
    {
      boost::shared_ptr<SimpleCombo> result(new SimpleCombo);
      result->reseed();
      return result;
    }

    boost::shared_ptr<SimpleCombo>
    SimpleCombo::construct(const std::vector<rngint_t> &seeds)
    {
      boost::shared_ptr<SimpleCombo> result(new SimpleCombo);
      result->vec_reseed(seeds);
      return result;
    }

    boost::shared_ptr<SimpleCombo>
    SimpleCombo::construct(rngint_t zarg)
    {
      boost::shared_ptr<SimpleCombo> result(new SimpleCombo);
      result->reseed(zarg);
      return result;
    }

    boost::shared_ptr<SimpleCombo>
    SimpleCombo::construct(rngint_t zarg,
                           rngint_t warg)
    {
      boost::shared_ptr<SimpleCombo> result(new SimpleCombo);
      result->reseed(zarg, warg);
      return result;
    }

    boost::shared_ptr<SimpleCombo>
    SimpleCombo::construct(rngint_t zarg,
                           rngint_t warg,
                           rngint_t jsrarg)
    {
      boost::shared_ptr<SimpleCombo> result(new SimpleCombo);
      result->reseed(zarg, warg, jsrarg);
      return result;
    }

    boost::shared_ptr<SimpleCombo>
    SimpleCombo::construct(rngint_t zarg,
                           rngint_t warg,
                           rngint_t jsrarg,
                           rngint_t jcongarg)
    {
      boost::shared_ptr<SimpleCombo> result(new SimpleCombo);
      result->reseed(zarg, warg, jsrarg, jcongarg);
      return result;
    }

    SimpleCombo::~SimpleCombo()
    {
    }

    void
    SimpleCombo::vec_reseed(const std::vector<rngint_t> &seeds)
    {
      std::vector<rngint_t> MWCseeds, SHR3seeds, CONGseeds;

      int i=0;
      for (int j=0; j<mwc_->nseed() && i < seeds.size(); j++,i++) {
        MWCseeds.push_back(seeds[i]);
      }
      for (int j=0; j<shr3_->nseed() && i < seeds.size(); j++,i++) {
        SHR3seeds.push_back(seeds[i]);
      }
      for (int j=0; j<cong_->nseed() && i < seeds.size(); j++,i++) {
        CONGseeds.push_back(seeds[i]);
      }

      mwc_->vec_reseed(MWCseeds);
      shr3_->vec_reseed(SHR3seeds);
      cong_->vec_reseed(CONGseeds);
    }

    int
    SimpleCombo::nseed()
    {
      return mwc_->nseed()+shr3_->nseed()+cong_->nseed();
    }

    ////////////////////////////////////////////////////////////////////
    // Table256

    Table256::Table256()
    {
      seeder_ = SimpleCombo::construct();
    }

    void
    Table256::vec_reseed(const std::vector<rngint_t> &seeds)
    {
      c = 0;
      seeder_->vec_reseed(seeds);
      for(int i=0;i<256;i++) t[i]=seeder_->value();
    }

    int
    Table256::nseed()
    {
      return seeder_->nseed();
    }

    ////////////////////////////////////////////////////////////////////
    // LFIB4

    LFIB4::LFIB4()
    {
    }

    boost::shared_ptr<LFIB4>
    LFIB4::construct()
    {
      boost::shared_ptr<LFIB4> result(new LFIB4);
      result->reseed();
      return result;
    }

    boost::shared_ptr<LFIB4>
    LFIB4::construct(const std::vector<rngint_t> &seeds)
    {
      boost::shared_ptr<LFIB4> result(new LFIB4);
      result->vec_reseed(seeds);
      return result;
    }

    LFIB4::~LFIB4()
    {
    }

    void
    LFIB4::vec_reseed(const std::vector<rngint_t> &seeds)
    {
      Table256::vec_reseed(seeds);
    }

    int
    LFIB4::nseed()
    {
      return Table256::nseed();
    }

    ////////////////////////////////////////////////////////////////////
    // SWB

    SWB::SWB()
    {
    }

    boost::shared_ptr<SWB>
    SWB::construct()
    {
      boost::shared_ptr<SWB> result(new SWB);
      result->reseed();
      return result;
    }

    boost::shared_ptr<SWB>
    SWB::construct(const std::vector<rngint_t> &seeds)
    {
      boost::shared_ptr<SWB> result(new SWB);
      result->vec_reseed(seeds);
      return result;
    }

    SWB::~SWB()
    {
    }

    void
    SWB::vec_reseed(const std::vector<rngint_t> &seeds)
    {
      x = 0;
      y = 0;
      Table256::vec_reseed(seeds);
    }

    int
    SWB::nseed()
    {
      return Table256::nseed();
    }

    ////////////////////////////////////////////////////////////////////
    // Combo

    Combo::Combo()
    {
      swb_ = SWB::construct();
      simplecombo_ = SimpleCombo::construct();
    }

    boost::shared_ptr<Combo>
    Combo::construct()
    {
      boost::shared_ptr<Combo> result(new Combo);
      result->reseed();
      return result;
    }

    boost::shared_ptr<Combo>
    Combo::construct(const std::vector<rngint_t> &seeds)
    {
      boost::shared_ptr<Combo> result(new Combo);
      result->vec_reseed(seeds);
      return result;
    }

    Combo::~Combo()
    {
    }

    void
    Combo::vec_reseed(const std::vector<rngint_t> &seeds)
    {
      std::vector<rngint_t> SWBseeds, SimpleComboseeds;

      int i=0;
      for (int j=0; j<swb_->nseed() && i < seeds.size(); j++,i++) {
        SWBseeds.push_back(seeds[i]);
      }
      for (int j=0; j<simplecombo_->nseed() && i < seeds.size(); j++,i++) {
        SimpleComboseeds.push_back(seeds[i]);
      }

      swb_->vec_reseed(SWBseeds);
      simplecombo_->vec_reseed(SimpleComboseeds);

      // This advances the simplecombo_ RNG to mimic the behavior of
      // the original RNG code, which used the same simplecombo
      // as the one used to seed the table.
      for (int i=0; i<256; i++) simplecombo_->value();
    }

    int
    Combo::nseed()
    {
      return swb_->nseed() + simplecombo_->nseed();
    }

  }

}
