/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <math.h> // pow
#include <boost/shared_ptr.hpp>
#include <time.h>
#include <sstream>

#include <sstmac/skeleton/mpipingpong/mpipingpong.h>
#include <sstmac/process/mpilaunch.h>
#include <sstmac/simulator/manager.h>
#include <sstmac/simulator/fullnetwork.h>
#include <sstmac/simulator/circuit.h>
#include <sstmac/simulator/randomindexing.h>
#include <sstmac/simulator/blockindexing.h>
#include <sstmac/simulator/roundrobinindexing.h>
#include <sstmac/simulator/allocationstrategy.h>
#include <sstmac/network/line.h>
#include <sstmac/network/ring.h>
#include <sstmac/network/hdtorus.h>
#include <sstmac/network/composite.h>
#include <sstmac/simulator/eventmap.h>
#include <sstmac/simulator/simplenode.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/network/fattree.h>
#include <sstmac/network/gamma.h>
#include <sstmac/network/crossbar.h>
#include <sstmac/network/torus.h>
#include <sstmac/skeleton/mpiparsers/parsedumpi.h>
#include <sstmac/skeleton/mpiparsers/dumpimeta.h>

// These have to be included after the sstmac headers
// because the dummies at Trolltech defined 'slots' to be a preprocessor token
#include <ui/simtabs.h>
#include <QtGui>
#include <QMessageBox>

using boost::shared_ptr;

using namespace sstmac;

qapplaunch::qapplaunch(int &argc, char *argv[]) : QApplication(argc, argv) {

}

bool qapplaunch::notify(QObject *receiver, QEvent *event) {
  std::string what;

  try {
    return QApplication::notify(receiver, event);
  } catch(sstmac::ssterror &err) {
    what = err.what();
    std::cout << "ssterror: " << what << std::endl;
  } catch(std::exception &err) {
    what = err.what();
    std::cout << "exception: " << what << std::endl;
  } catch(const char *err) {
    what = err;
    std::cout << "err: " << what << std::endl;;
  } catch(...) {
    what =  "Unknown error";
  };
    
  if(what != "")
    QMessageBox::information(0,tr("Error"), tr(what.c_str()));

  return false;
}

SimDialog::SimDialog(QWidget *parent)
   : QMainWindow(parent)
{
  //appl_ = appl;
  Qt::WindowFlags flags = 0;
  flags |= Qt::WindowMinimizeButtonHint;
  flags |= Qt::WindowMaximizeButtonHint;
  flags |= Qt::WindowCloseButtonHint;

  setWindowFlags(flags);
         
  //mdi
  mdiArea = new QMdiArea;
  setCentralWidget(mdiArea);
         
  connect(mdiArea, SIGNAL(subWindowActivated(QMdiSubWindow *)),
          this, SLOT(updateMenus()));
  windowMapper = new QSignalMapper(this);
  connect(windowMapper, SIGNAL(mapped(QWidget *)),
          this, SLOT(setActiveSubWindow(QWidget *)));
         
  tabWidget = new QTabWidget;
  QWidget *child = new parametertab(tabWidget, mdiArea);

  setWindowIcon(QIcon(":./peertopeer.png"));


  mdiArea->addSubWindow(child, Qt::WindowMinMaxButtonsHint | Qt::WindowTitleHint);
  child->setWindowTitle("Input Parameters");
   
  createMenus();
  updateMenus();
         
  setWindowTitle(tr("SST Macroscale Simulator"));
}

void SimDialog::updateWindowMenu()
{
  windowMenu->clear();
  windowMenu->addAction(closeAct);
  windowMenu->addAction(closeAllAct);
  windowMenu->addSeparator();
  windowMenu->addAction(tileAct);
  windowMenu->addAction(cascadeAct);
  windowMenu->addSeparator();

  QList<QMdiSubWindow *> windows = mdiArea->subWindowList();
  separatorAct->setVisible(!windows.isEmpty());
    
  if(!windows.isEmpty()) {
    QWidget *child = qobject_cast<QWidget *>(windows.at(0)->widget());
    QString text = child->windowTitle();//tr("Input Parameters");
    QAction *action  = windowMenu->addAction(text);
    action->setCheckable(true);
    action->setChecked(child == activeMdiChild());
    connect(action, SIGNAL(triggered()), windowMapper, SLOT(map()));
    windowMapper->setMapping(action, windows.at(0));
                
    for (int i = 1; i < windows.size(); ++i) {
      QWidget *child = qobject_cast<QWidget *>(windows.at(i)->widget());

      //text = tr("Results %1").arg(i);
      text = child->windowTitle();
      QAction *action  = windowMenu->addAction(text);
      action->setCheckable(true);
      action->setChecked(child == activeMdiChild());
      connect(action, SIGNAL(triggered()), windowMapper, SLOT(map()));
      windowMapper->setMapping(action, windows.at(i));
    }
  }
}

void SimDialog::closeallandcheck()
{
        
  QList<QMdiSubWindow *> windows = mdiArea->subWindowList();
        
  for (int i = 1; i < windows.size(); ++i) {

    QMdiSubWindow *existing = findMdiChild();
    if (existing) {

      mdiArea->setActiveSubWindow(existing);
      mdiArea->closeActiveSubWindow();
    }
  }
}

QMdiSubWindow *SimDialog::findMdiChild()
{
  QString title = "Input Parameters";

  foreach (QMdiSubWindow *window, mdiArea->subWindowList()) {
                
    QWidget *mdiChild = qobject_cast<QWidget *>(window->widget());
    if (mdiChild->windowTitle() != title) {
      return window;
    }
  }
  return 0;
}

void SimDialog::closeandcheck()
{
  //std::cout << "closeandcheck\n";
  QList<QMdiSubWindow *> windows = mdiArea->subWindowList();
                
  if(!windows.isEmpty()) {
    //mdiArea->closeAllSubWindows();
    QWidget *child = mdiArea->activeSubWindow();
    QString text = child->windowTitle();
    //std::cout << "text = " << text << std::endl;
    if (text == "Input Parameters") {
      //std::cout << "input par\n";
      return;
    } else {
      //std::cout << "no input par\n";
      mdiArea->closeActiveSubWindow();
    }
  }
}

void SimDialog::setActiveSubWindow(QWidget *window)
{
  if (!window)
    return;
  mdiArea->setActiveSubWindow(qobject_cast<QMdiSubWindow *>(window));
}

void SimDialog::createMenus()
{
  exitAct = new QAction(tr("E&xit"), this);
  exitAct->setShortcut(tr("Ctrl+Q"));
  exitAct->setStatusTip(tr("Exit the application"));
  connect(exitAct, SIGNAL(triggered()), qApp, SLOT(closeAllWindows()));
  
  fileMenu = menuBar()->addMenu(tr("&File"));
  fileMenu->addAction(exitAct);

  closeAct = new QAction(tr("Cl&ose"), this);
  closeAct->setShortcut(tr("Ctrl+F4"));
  closeAct->setStatusTip(tr("Close the active window"));
  //connect(closeAct, SIGNAL(triggered()),
  //        mdiArea, SLOT(closeActiveSubWindow()));
  connect(closeAct, SIGNAL(triggered()),
          this, SLOT(closeandcheck()));

  closeAllAct = new QAction(tr("Close &All Results"), this);
  closeAllAct->setStatusTip(tr("Close all results windows"));
  //connect(closeAllAct, SIGNAL(triggered()),
  //      mdiArea, SLOT(closeAllSubWindows()));
  connect(closeAllAct, SIGNAL(triggered()),
          this, SLOT(closeallandcheck()));
  
  tileAct = new QAction(tr("&Tile"), this);
  tileAct->setStatusTip(tr("Tile the windows"));
  connect(tileAct, SIGNAL(triggered()), mdiArea, SLOT(tileSubWindows()));

  cascadeAct = new QAction(tr("&Cascade"), this);
  cascadeAct->setStatusTip(tr("Cascade the windows"));
  connect(cascadeAct, SIGNAL(triggered()), this, SLOT(customcascadeSubWindows()));
  
  separatorAct = new QAction(this);
  separatorAct->setSeparator(true);
        
  windowMenu = menuBar()->addMenu(tr("&Window"));
  windowMenu->addAction(closeAct);
  windowMenu->addAction(closeAllAct);
  windowMenu->addAction(tileAct);
  windowMenu->addAction(cascadeAct);
  windowMenu->addAction(cascadeAct);
        
  updateWindowMenu();
  connect(windowMenu, SIGNAL(aboutToShow()), this, SLOT(updateWindowMenu()));
  
  menuBar()->addSeparator();
  mdiArea->setActivationOrder(QMdiArea::CreationOrder);
}

void SimDialog::updateMenus()
{
  bool hasMdiChild = true;//(activeMdiChild() != 0);
  closeAct->setEnabled(hasMdiChild);
  closeAllAct->setEnabled(hasMdiChild);
  tileAct->setEnabled(hasMdiChild);
  cascadeAct->setEnabled(hasMdiChild);
  separatorAct->setVisible(hasMdiChild);
}

void SimDialog::customcascadeSubWindows()
{
  mdiArea->cascadeSubWindows();
        
  QList<QMdiSubWindow *> windows = mdiArea->subWindowList();
  
  for(int i = windows.size()-1; i >= 0; --i) 
    {
      windows.at(i)->lower();
    }
}


QWidget *SimDialog::activeMdiChild()
{
  if (QMdiSubWindow *activeSubWindow = mdiArea->activeSubWindow())
    return qobject_cast<QWidget *>(activeSubWindow->widget());
  return 0;
}

void SimDialog::closethetab(int i)
{
  tabWidget->removeTab(i);
}

runmetrics::runmetrics(double time, char *metafilename)
{
  time_ = time;
  metafilename_ = metafilename;
}

runmetrics&
runmetrics::operator=(const runmetrics& that)
{
  if (this == &that) return *this;   // Gracefully handle self assignment
        
  time_ = that.time_;

  return *this;
}

void resultstab::tokenize(const std::string& str,
                      std::vector<std::string>& tokens,
                      const std::string& delimiters = " ")
{
  // Skip delimiters at beginning.
  std::string::size_type lastPos = str.find_first_not_of(delimiters, 0);
  // Find first "non-delimiter".
  std::string::size_type pos     = str.find_first_of(delimiters, lastPos);

  while (std::string::npos != pos || std::string::npos != lastPos)
    {
      // Found a token, add it to the vector.
      tokens.push_back(str.substr(lastPos, pos - lastPos));
      // Skip delimiters.  Note the "not_of"
      lastPos = str.find_first_not_of(delimiters, pos);
      // Find next "non-delimiter"
      pos = str.find_first_of(delimiters, lastPos);
    }
}


resultstab::resultstab(RMPtr metrics, QWidget *parent)
     : QWidget(parent)
{
  assert(metrics != NULL);
  metrics_ = metrics;
  //QLabel *rtimeLabel = new QLabel(tr("Run time = "));
  QLabel *timeLabel = new QLabel();
  QLabel *paramsLabel = new QLabel();
  QString holdtime = QString("Runtime = %1").arg(metrics_->time_ - 1, 0, 'g', 10);
  //hold.sprintf("%.f13", metrics_->time_);
  std::ifstream filestr;
  filestr.open (metrics_->metafilename_);
  
  std::string line, params;
  std::vector<std::string> tokens;
        
  params += "INPUT PARAMETERS\n\n";
        
  while(getline(filestr, line)) {
    //std::cout << "line = " << line << std::endl;
                
    tokenize(line, tokens, " .\n,");
        
    for(unsigned int i = 0; i < tokens.size(); i++ )
      {
        if(tokens[0] == "import" || tokens[0] == "mgr" || tokens[0] == "starttime" || tokens[0] == "runtime" || tokens[0] == "print" || tokens[0] == "alloc_unit" || tokens[0] == "stack" || tokens[0] == "nnodejob" || tokens[0] == "meta" || tokens[0] == "parser")
          break;
        else if(tokens[0] == "net") {
          params += "network = ";
          params += tokens[3];
          params += "\n";
          break;
        }
        else if(tokens[0] == "interconnect") {
          params += "interconnect = ";
          params += tokens[3];
          params += "\n\n";
          break;
        }
        else if(tokens[0] == "job") {
          break;
        }
        else if(tokens[0] == "indexing") {
          params += "indexing = ";
          params += tokens[3];
          params += "\n";
          break;
        }
        else if(tokens[0] == "allocation") {
          params += "allocation = ";
          params += tokens[3];
          params += "\n";
          break;
        }
        else if(tokens[0] == "#Python") {
          params += "\nPython script: ";
          params += tokens[2] + "." + tokens[3] + "." + tokens[4] + "." + tokens[5] + "." + tokens[6] + "." + tokens[7] + "." + tokens[8] + "\n";
          break;
        }
        else {
          params += line;
          params += "\n";
          break;
        }
      }
    tokens.clear();
  }
        
  QString qparams = QString(params.c_str());
        
  timeLabel->setText(holdtime);
  paramsLabel->setText(qparams);
  timeLabel->setFrameStyle(QFrame::Panel | QFrame::Sunken);
  paramsLabel->setFrameStyle(QFrame::Panel | QFrame::Sunken);

  QVBoxLayout *mainLayout = new QVBoxLayout();
  //      mainLayout->addWidget(rtimeLabel);
  mainLayout->addWidget(timeLabel);
  mainLayout->addWidget(paramsLabel);
  mainLayout->addStretch(1);
        
  setLayout(mainLayout);

}

void parametertab::setActiveSubWindow(QWidget *window)
{
  if (!window)
    return;
  mdiArea->setActiveSubWindow(qobject_cast<QMdiSubWindow *>(window));
}

void parametertab::construct_indexing(shared_ptr<sstmac::indexstrategy> &ind) {

  if(indexComboBox->currentText() == "Random") {
    fs << "indexing = sstmac.randomindexing.construct()\n";
    if(scriptonly != true) 
      ind = sstmac::randomindexing::construct();
  }
  else if(indexComboBox->currentText() == "Block") {
    fs << "indexing = sstmac.blockindexing.construct()\n";
    if(scriptonly != true) 
      ind = sstmac::blockindexing::construct();
  }
  else if(indexComboBox->currentText() == "Round robin") {
    fs << "indexing = sstmac.roundrobinindexing.construct()\n";
    if(scriptonly != true) 
      ind = sstmac::roundrobinindexing::construct();
  }
}

void parametertab::construct_allocation(shared_ptr<sstmac::allocationstrategy> &allocation) {

  if(allocationComboBox->currentText() == "Random") {
    fs << "allocation = sstmac.randomallocation.construct()\n";
    if(scriptonly != true) 
      allocation = sstmac::randomallocation::construct();
  }
  else if(allocationComboBox->currentText() == "First available") {
    fs << "allocation = sstmac.firstavailableallocation.construct()\n";
    if(scriptonly != true) 
      allocation = sstmac::firstavailableallocation::construct();
  }
}

void parametertab::construct_manager(shared_ptr<manager> &mgr, shared_ptr<interconnect> &fab, int nnodes)
{
  if(scriptonly != true) { 
    mgr->set_interconnect(fab);
    mgr->set_eventmanager(eventmap::construct());
    mgr->add_nodes(simplenode::construct(mem_band), nnodes);
  }
  
  fs << "mgr = sstmac.manager.construct()\n" <<
    "mgr.set_interconnect(interconnect)\n" <<
    "mgr.set_eventmanager(sstmac.eventmap.construct())\n" <<
    "nnodes = " << nnodes << "\n" <<
    "mgr.add_nodes(sstmac.simplenode.construct(mem_band), nnodes)\n";
}

bool parametertab::setupsim()
{
  int nnodes = 0, ppn = 0, iter = 10, arrlen = 4;
  double timescale = 1.;
  size_t stack = 81920;
  size_t alloc_unit = 10240000;
  std::string filename;
        
  shared_ptr<manager> mgr = manager::construct();
  shared_ptr<interconnect> fab;
  shared_ptr<network> net;
  shared_ptr<mpilaunch> job;
  shared_ptr<parsedumpi> pars;
  shared_ptr<dumpimeta> meta;
  
  if(!validatefields(fab,net,nnodes,ppn)) {
    std::cout << "validatefields failed\n";
    return false; //validation failed.
  }
                
  if(typeComboBox->currentIndex() == 0)
    {
      if(!validatepingpong(iter, arrlen))
        return false;
                        
      job = mpilaunch::construct(mpipingpong::construct(iter, arrlen));
    }
  else
    {
      if(!validatetracedriven(filename, timescale))
        return false;
                
      meta = dumpimeta::construct(filename);
                
      //Check that number of procs in meta file matches number of procs input in gui
      if(meta->getnumprocs() != (ppn*nnodes))
        {
          std::stringstream out;
          std::string errmsg;
          
          out << "Number nodes specified (NNodes*PPN = " << ppn*nnodes << ") does not match trace meta file (" << meta->getnumprocs() << ").\n";
          errmsg = out.str();
          std::cerr << errmsg << "\n";
          QMessageBox::information(this,"Mismatch between meta file and input", tr(errmsg.c_str()));
          return false;
        }
                
      pars = parsedumpi::construct(meta,timescale);
      job = mpilaunch::construct(pars);
                
      fs << "tscale = " << timescale << "\n"
         << "inputfile = \"" << filename << "\"\n"
         << "meta = sstmac.dumpimeta.construct(inputfile)\n"
         << "parser = sstmac.parsedumpi.construct(meta,tscale)\n"
         << "job = sstmac.mpilaunch.construct(parser)\n";
    }
  
  shared_ptr<sstmac::indexstrategy> index;
  construct_indexing(index);
  shared_ptr<sstmac::allocationstrategy> allocation;
  construct_allocation(allocation);
  construct_manager(mgr,fab,nnodes);
                
  if(scriptonly != true) {
    job->set_scheduler(scheduler::construct(stack, alloc_unit, false));
                
    mgr->add_launcher(timestamp(1), job, nnodes, ppn, index, allocation);
  }
  
  fs << "nnodejob = " << nnodes << "\n" << "nprocpernode = " << ppn << "\n" << "stack = " << stack << "\nalloc_unit = " << alloc_unit << "\n";
  fs << "job.set_scheduler(sstmac.scheduler.construct(stack, alloc_unit, False))\n";
  fs << "mgr.add_launcher(sstmac.timestamp(1), job, nnodejob, nprocpernode, indexing, allocation)\n" <<
    "starttime = mgr.run()\n" 
     << "runtime = starttime.sec()\n"
     << "print ('simulated run time for %d nodes with %d processes per node = %12.9f seconds') % (nnodejob, nprocpernode, runtime)\n";
  
  //Don't run simulation, just generate the python script 
  if(scriptonly != true) {
  
    timestamp t = mgr->run();
    std::cout << "end time is " << t << std::endl;

    rmetrics_ = shared_ptr<runmetrics>(new runmetrics(t.sec(), metafilename_));
  }
  
  return true;
}

void parametertab::runsim()
{

  initpythonscr(); //writes python script to rerun sim
  if(!setupsim()){
    std::cout << "Simulation initialization failed\n";
    fs.close();
    std::remove(metafilename_);
    return;
  }
        
  if(scriptonly)
    QMessageBox::information(this, "Python script generated: ", metafilename_);
  else
    std::cout << "Python script generated: " << metafilename_ << std::endl;
        
  fs << "#Python script: " << metafilename_ << std::endl;
  fs.close(); //close python script file
        
  //mdi
  if(scriptonly != true) {
    QWidget *child = new resultstab(rmetrics_);
                
    //child->setWindowTitle(tr("Results %1").arg(sequencenum++));
    child->setWindowTitle(lineEditOutput->text());
    mdiArea->addSubWindow(child);
    child->setVisible(true);
                
    //windowMapper = new QSignalMapper(child);
    //connect(windowMapper, SIGNAL(mapped(QWidget *)),
    //        this, SLOT(setActiveSubWindow(QWidget *)));
  }
        
  lineEditOutput->setText(tr("Results %1").arg(++sequencenum));
  
}

parametertab::parametertab(QTabWidget *tW, QMdiArea *ma, QWidget *parent)
   : QWidget(parent)
//parametertab::parametertab(QMdiArea *tW, QWidget *parent)
//         : QMdiArea(parent)

{
  //initpythonscr();
         
  tabWidget = tW;
  mdiArea = ma;
         
  menuBar = new QMenuBar;
  fileMenu = new QMenu(tr("&File"), this);
  exitAction = fileMenu->addAction(tr("E&xit"));
  menuBar->addMenu(fileMenu);
        
        
  mainWin = new QMainWindow;
  sequencenum = 1;
  //createActions();
  //createMenus();
  //updateMenus();
         
  //createMenu();
  createFullIndexGroupBox();
  createFullAllocationGroupBox();
  createCircuitIndexGroupBox();
  createCircuitAllocationGroupBox();
  createParameterGroupBox();
  //createProcsGroupBox();
  createNetworkGroupBox();
  createControlGroupBox();
  createSimTypeGroupBox();
  createNodesGroupBox();
  createOutputLabelGroupBox();
   
  QGridLayout *mainLayout = new QGridLayout();
  mainLayout->addWidget(outputGroupBox,0,0,1,1);
  mainLayout->addWidget(fileGroupBox,1,0,1,3);
  mainLayout->addWidget(nodesGroupBox,2,0,1,3);
  mainLayout->addWidget(interTabWidget,3,0,1,3);
  //mainLayout->addWidget(procsGroupBox,2,0,1,1);
  //mainLayout->addWidget(indexGroupBox,2,1,1,1);
  //mainLayout->addWidget(allocationGroupBox,2,2,1,1);
   
  //mainLayout->addWidget(controlGroupBox,4,0,1,3);

  setLayout(mainLayout);
}

void parametertab::initpythonscr()
{
  //char *scratch;
  struct tm loct;
  time_t ltime;
  long fsize = 50;

  metafilename_ = (char*)malloc(fsize); 
  metafilename_[fsize-1] = '\0';
        
  ltime=time(&ltime);
  localtime_r(&ltime, &loct);
  snprintf(metafilename_, fsize-1, "sim-%04d.%02d.%02d.%02d.%02d.%02d.py",
           loct.tm_year+1900, loct.tm_mon+1, loct.tm_mday,
           loct.tm_hour, loct.tm_min, loct.tm_sec);

  fs.open( metafilename_, std::ios::out );
  text_start_ = fs.tellg();
  fs << "import sys\n" << "import sstmac\n";
}
 
void parametertab::createOutputLabelGroupBox()
{
  outputGroupBox = new QGroupBox(tr("Output Window Label"));
        
  //QLabel *labelOutput = new QLabel(tr("Label output:"));
  lineEditOutput = new QLineEdit;
  lineEditOutput->setFrame(false);
  lineEditOutput->setText("Results 1");
  QLabel *labelfiller = new QLabel("                                       ");
  QLabel *labelfiller1 = new QLabel("                                       ");
  QLabel *labelfiller2 = new QLabel("                                       ");
        
  QHBoxLayout *layout = new QHBoxLayout;
        
  //layout->addWidget(labelOutput);
  layout->addWidget(lineEditOutput);
  layout->addWidget(labelfiller);
  layout->addWidget(labelfiller1);
  layout->addWidget(labelfiller2);
        
  outputGroupBox->setLayout(layout);

}

void parametertab::createSimTypeGroupBox()
{
  typeComboBox = new QComboBox;

  typeComboBox->addItem(tr("Ping-pong test"));
  typeComboBox->addItem(tr("MPI trace driven"));

  connect(typeComboBox, SIGNAL(activated(int)),
          this, SLOT(typeChanged(int)));                  

  fileGroupBox = new QGroupBox(tr("Simulation Type"));
  
  labelTimeScale = new QLabel(tr("Time scale:"));
  labelTimeScale->hide();
  labelTimeScale->setAlignment(Qt::AlignRight);

  lineEditTimeScale = new QLineEdit;
  lineEditTimeScale->hide();
  lineEditTimeScale->setFrame(false);
  lineEditTimeScale->setText("1");
        
  lineeditFileName = new QLineEdit();
  lineeditFileName->hide();
  lineeditFileName->setFrame(false);
  buttonBrowse = new QPushButton(tr("&Browse trace files"));
  buttonBrowse->hide();
        
  connect(buttonBrowse, SIGNAL(clicked()),
          this, SLOT(setOpenFileName()));
  
  labelIteration = new QLabel(tr("Iterations:"));
  labelIteration->setAlignment(Qt::AlignRight);
  labeldatalength = new QLabel("Packet size:");
  labeldatalength->setAlignment(Qt::AlignRight);
  lineEditIter = new QLineEdit;
  lineEditIter->setFrame(false);
  lineEditIter->setText("10");
  lineEditPacket = new QLineEdit;
  lineEditPacket->setFrame(false);
  lineEditPacket->setText("4");
  QLabel *filler = new QLabel("               ");
        
  QGridLayout *layout = new QGridLayout;
  layout->addWidget(typeComboBox,0,0);
  //      layout->addWidget(radiotrace,1,0);
  layout->addWidget(labelTimeScale,0,1);
  layout->addWidget(lineEditTimeScale,0,2);
  layout->addWidget(filler,0,3);
  //      layout->addWidget(labelFiller,0,2);
  //      layout->setColumnMinimumWidth(2,200);
  layout->addWidget(lineeditFileName,1,0,1,3);
  layout->addWidget(buttonBrowse,1,3,1,1);
        
  //      layout->addWidget(radioping,0,0);
  layout->addWidget(labelIteration,0,1);
  layout->addWidget(lineEditIter,0,2);
  //      layout->addWidget(filler,0,2);
  //      layout->setColumnMinimumWidth(2,200);
  layout->addWidget(labeldatalength,0,3);
  layout->addWidget(lineEditPacket,0,4);
  //      layout->addWidget(filler,1,2);
        
  fileGroupBox->setLayout(layout);
}

void parametertab::setOpenFileName()
{
  QFileDialog::Options options;
    
  QString selectedFilter;
  QString fileName = QFileDialog::getOpenFileName(this,
                                                  tr("Browse for trace files"),
                                                  lineeditFileName->text(),
                                                  tr("All Files (*);;Text Files (*.txt)"),
                                                  &selectedFilter,
                                                  options);
  if (!fileName.isEmpty())
    lineeditFileName->setText(fileName);
        
  /*QString qfile = lineeditFileName->text();
    std::string stdstring = std::string(qfile.toAscii());
    std::cout << "Filename = " << stdstring << std::endl;*/
}

void parametertab::createNodesGroupBox()
{
  nodesGroupBox = new QGroupBox(tr("Nodes"));
  QLabel *labelnnodes = new QLabel(tr("NNodes:"));
  labelnnodes->setAlignment(Qt::AlignRight);
  lineeditnnodes = new QLineEdit();
  lineeditnnodes->setText("2");
  QLabel *labelprocspernode = new QLabel(tr("PPN:"));
  labelprocspernode->setAlignment(Qt::AlignRight);
  lineeditprocspernode = new QLineEdit();
  lineeditprocspernode->setText("2");
        
  QGridLayout *procsbox = new QGridLayout;
  procsbox->addWidget(labelnnodes,2,1);
  procsbox->addWidget(lineeditnnodes,2,2);
  procsbox->addWidget(labelprocspernode,2,3);
  procsbox->addWidget(lineeditprocspernode,2,4);
        
  QLabel *labelindexCB = new QLabel(tr("Indexing:"));
  labelindexCB->setAlignment(Qt::AlignRight);
  procsbox->addWidget(labelindexCB,3,1);
  procsbox->addWidget(indexComboBox,3,2);
  QLabel *labelallocCB = new QLabel(tr("Allocation:"));
  labelallocCB->setAlignment(Qt::AlignRight);
  procsbox->addWidget(labelallocCB,3,3);
  procsbox->addWidget(allocationComboBox,3,4);
        
  nodesGroupBox->setLayout(procsbox);
}
void parametertab::createControlGroupBox()
{
  controlGroupBox = new QGroupBox();
  QHBoxLayout *layout = new QHBoxLayout;

  QPushButton *buttonRun = new QPushButton(tr("&Run"));
  QPushButton *buttonClear = new QPushButton(tr("C&lear"));
  QLabel *filler = new QLabel();
        
  connect(buttonRun, SIGNAL(clicked()),
          this, SLOT(runsim()));
        
  layout->addWidget(filler);
  layout->addWidget(buttonClear);
  layout->addWidget(buttonRun);
  controlGroupBox->setLayout(layout);
}

bool parametertab::validatedouble(QString text, double min = 0.0, double max = 1e15, double dec = 12)
{
  int pos = 0;

  QDoubleValidator vdouble(min, max, dec, this);
  vdouble.setNotation(QDoubleValidator::ScientificNotation);
        
  if (vdouble.validate(text , pos ) == QValidator::Invalid ||
      vdouble.validate(text , pos ) == QValidator::Intermediate) {
    return false;
  } 
        
  return true;
}

bool parametertab::validateint(QString text, int min = 0, int max = 1000000000)
{
  int pos = 0;
  QIntValidator vint( min, max, this );
        
  if (vint.validate(text, pos ) == QValidator::Invalid ||
      vint.validate(text, pos ) == QValidator::Intermediate) {
    return false;
  } 
        
  return true;
}

bool parametertab::validatepingpong(int &iter, int &arrlen) {
        
  if(!validateint(lineEditIter->text(),1)) {
    /*if(lineEditIter->text() == "")
      lineEditIter->setText("   ");
      lineEditIter->selectAll();*/
    QMessageBox::information(this, "Invalid Input", "Interations requires an integer value.");
    return false;
  }
  if(!validateint(lineEditPacket->text(),1)) {
    //lineEditPacket->selectAll();
    QMessageBox::information(this, "Invalid Input", "Packet size requires an integer value.");
    return false;
  }
        
  iter = lineEditIter->text().toInt();
  arrlen = lineEditPacket->text().toInt();
        
  fs << "iter = " << iter << "\n"
     << "arrlen = " << arrlen << "\n"
     << "job = sstmac.mpilaunch.construct(sstmac.mpipingpong.construct(iter, arrlen))\n";
        
  return true;
}

bool parametertab::validatetracedriven(std::string &fname, double &tscale) {
        
  if(!validatedouble(lineEditTimeScale->text())) {
    /*if(lineEditTimeScale->text() == "")
      lineEditTimeScale->setText("   ");
      lineEditTimeScale->selectAll();*/
    QMessageBox::information(this, "Invalid Input", "Time scale requires a floating point value.");
    return false;
  }
        
  tscale = lineEditTimeScale->text().toDouble();
        
  if(lineeditFileName->text() == "")
    {
      QMessageBox::information(this, "Invalid Input", "Trace file not specified.");
      //lineeditFileName->setText("   ");
      //lineEditTimeScale->selectAll();
      return false;
    }
        
  QString qfile = lineeditFileName->text();
  fname = std::string(qfile.toAscii());
  
  return true;
}

bool parametertab::validatennodes(int &nnodes) {

  if(!validateint(lineeditnnodes->text(),1)) {
    /*if(lineeditnnodes->text() == "") {
      QMessageBox::information(this, "Invalid Input", "NNodes requires a integer value.");
      //lineeditnnodes->setText("      ");
      //lineeditnnodes->selectAll();
      }*/
    //lineeditnnodes->selectAll();
    QMessageBox::information(this, "Invalid Input", "NNodes requires an integer value.");
    return false;
  }
        
  nnodes = lineeditnnodes->text().toInt();
  return true;
}

bool parametertab::validateppn(int &ppn) {

  if(!validateint(lineeditprocspernode->text())) {
    //lineeditprocspernode->selectAll();
    QMessageBox::information(this, "Invalid Input", "PPN requires an integer value.");
    return false;
  }
        
  ppn = lineeditprocspernode->text().toInt();
  return true;
}

bool parametertab::validateLatBandMem(timestamp &lat, double &band) {
        
  if(!validatedouble(lineEditLat->text())) {
    //lineEditLat->selectAll();
    QMessageBox::information(this, "Invalid Input", "Latency (sec) requires a floating point value.");
    return false;
  }
  if(!validatedouble(lineEditBand->text())) {
    //lineEditBand->selectAll();
    QMessageBox::information(this, "Invalid Input", "Bandwidth (Bps) requires a floating point value.");
    return false;
  }
  if(!validatedouble(lineEditMemBand->text())) {
    //lineEditMemBand->selectAll();
    QMessageBox::information(this, "Invalid Input", "Memory Bandwidth (Bps) requires a floating point value.");
    return false;
  }
  lat = timestamp(lineEditLat->text().toDouble());
  band = lineEditBand->text().toDouble();
        
  fs << "lat = " << lat.sec() << "\nband = " << band << "\n";
        
  //fab = sstmac::circuit::construct(lat, band, net);
        
  return true;
}

bool parametertab::validateLatBand(shared_ptr<interconnect> &fab, shared_ptr<network> &net)
{
  if(!validatedouble(lineEditIntraLat->text())) {
    lineEditIntraLat->selectAll();
    return false;
  }
  if(!validatedouble(lineEditInterLat->text())) {
    lineEditInterLat->selectAll();
    return false;
  }
  if(!validatedouble(lineEditIntraBand->text())) {
    lineEditIntraBand->selectAll();
    return false;
  }
  if(!validatedouble(lineEditInterBand->text())) {
    lineEditInterBand->selectAll();
    return false;
  }
  timestamp lat;
  double band = 0.0;
        
  if(comboboxLat->currentText() == "Seconds")
    lat = timestamp(lineEditInterLat->text().toDouble());
  else if(comboboxLat->currentText() == "Milliseconds")
    lat = timestamp(lineEditInterLat->text().toDouble()*pow(10,-3));
  else if(comboboxLat->currentText() == "Microseconds")
    lat = timestamp(lineEditInterLat->text().toDouble()*pow(10,-6));
  else if(comboboxLat->currentText() == "Nanoseconds")
    lat = timestamp(lineEditInterLat->text().toDouble()*pow(10,-9));
        
  if(comboboxBand->currentText() == "Bytes/s")
    band = lineEditInterBand->text().toDouble();
  else if(comboboxBand->currentText() == "KB/s")
    band = lineEditInterBand->text().toDouble()*pow(2,10);
  else if(comboboxBand->currentText() == "MB/s")
    band = lineEditInterBand->text().toDouble()*pow(2,20);
  else if(comboboxBand->currentText() == "GB/s")
    band = lineEditInterBand->text().toDouble()*pow(2,30);
        
  fab = sstmac::circuit::construct(lat, band, net);

  return true;
}

bool parametertab::validateN1andN2()
{
  if(!validateint(lineEditn1->text())) {
    //lineEditn1->selectAll();
    QMessageBox::information(this, "Invalid Input", "N1 requires an integer value.");
    return false;
  }
  if(!validateint(lineEditn2->text())) {
    //lineEditn2->selectAll();
    QMessageBox::information(this, "Invalid Input", "N2 requires an integer value.");
    return false;
  }
        
  return true;
}

bool parametertab::validateLevels()
{
  if(!validateint(lineEditFatLevels->text())) {
    //lineEditFatLevels->selectAll();
    QMessageBox::information(this, "Invalid Input", "Levels requires an integer value.");
    return false;
  }
        
  return true;
}

bool parametertab::validateLinks()
{
  if(!validateint(lineEditFatLinks->text())) {
    //lineEditFatLinks->selectAll();
    QMessageBox::information(this, "Invalid Input", "Links requires an integer value.");
    return false;
  }
        
  return true;
}

bool parametertab::validateDegree()
{
  if(!validateint(lineEditGammaDegree->text())) {
    //lineEditGammaDegree->selectAll();
    QMessageBox::information(this, "Invalid Input", "Degree requires an integer value.");
    return false;
  }
        
  return true;
}

bool parametertab::validateDiameter()
{
  if(!validateint(lineEditGammaDiameter->text())) {
    //lineEditGammaDiameter->selectAll();
    QMessageBox::information(this, "Invalid Input", "Diameter requires an integer value.");
    return false;
  }
        
  return true;
}

bool parametertab::validateN1()
{
  if(!validateint(lineEditn1->text())) {
    //lineEditn1->selectAll();
    QMessageBox::information(this, "Invalid Input", "N1 requires an integer value.");
    return false;
  }
        
  return true;
}

bool parametertab::validateN3()
{ 
  if(!validateint(lineEditn3->text())) {
    //lineEditn3->selectAll();
    QMessageBox::information(this, "Invalid Input", "N3 requires an integer value.");
    return false;
  }
  return true;
}

bool parametertab::validateNx()
{ 
  if(!validateint(lineEditnx->text())) {
    //lineEditnx->selectAll();
    QMessageBox::information(this, "Invalid Input", "Nx requires an integer value.");
    return false;
  }
  return true;
}

bool parametertab::validateMemBand()
{ 
  if(!validatedouble(lineEditMemBand->text())) {
    //lineEditMemBand->selectAll();
    QMessageBox::information(this, "Invalid Input", "Memory Bandwidth (Bps) requires a floating point value.");
    return false;
  }
  mem_band = lineEditMemBand->text().toDouble();
  fs << "mem_band = " << lineEditMemBand->text().toDouble() << "\n";
        
  return true;
}

bool parametertab::validateFullMemBand()
{ 
  if(!validatedouble(lineEditFullMemBand->text())) {
    //lineEditFullMemBand->selectAll();
    QMessageBox::information(this, "Invalid Input", "Memory Bandwidth (Bps) requires a floating point value.");
    return false;
  }
  //fs << "mem_band = " << lineEditFullMemBand->text().toDouble() << "\n";
  return true;
}


bool parametertab::validatewidth()
{ 
  if(!validateint(lineEditwidth->text())) {
    //lineEditwidth->selectAll();
    QMessageBox::information(this, "Invalid Input", "Width requires an integer value.");
    return false;
  }
  return true;
}

bool parametertab::validateswitch()
{ 
  if(!validateint(lineEditnnodeperswitch->text())) {
    //lineEditnnodeperswitch->selectAll();
    QMessageBox::information(this, "Invalid Input", "Nodes per switch requires an integer value.");
    return false;
  }
  return true;
}

bool parametertab::validateFullnetwork(shared_ptr<interconnect> &fab)
{
        
  if(!validatedouble(lineEditLat->text())) {
    //lineEditInterLat->selectAll();
    QMessageBox::information(this, "Invalid Input", "Latency (sec) requires a floating point value.");
    return false;
  }
                
  if(!validatedouble(lineEditBand->text())) {
    //lineEditInterBand->selectAll();
    QMessageBox::information(this, "Invalid Input", "Bandwidth (Bps) requires a floating point value.");
    return false;
  }
        
  if(!validatedouble(lineEditMemBand->text())) {
    //lineEditMemBand->selectAll();
    QMessageBox::information(this, "Invalid Input", "Memory Bandwidth (Bps) requires a floating point value.");
    return false;
  }
  mem_band = lineEditMemBand->text().toDouble();
        
        
  timestamp lat;
  lat = timestamp(lineEditLat->text().toDouble());
  double band = 0.0;
  band = lineEditBand->text().toDouble();
        
  /*
    if(comboboxLat1->currentText() == "Seconds")
    lat = timestamp(lineEditLat->text().toDouble());
    else if(comboboxLat1->currentText() == "Milliseconds")
    lat = timestamp(lineEditLat->text().toDouble()*pow(10,-3));
    else if(comboboxLat1->currentText() == "Microseconds")
    lat = timestamp(lineEditLat->text().toDouble()*pow(10,-6));
    else if(comboboxLat1->currentText() == "Nanoseconds")
    lat = timestamp(lineEditLat->text().toDouble()*pow(10,-9));
        
    if(comboboxBand1->currentText() == "Bytes/s")
    band = lineEditBand->text().toDouble();
    else if(comboboxBand1->currentText() == "KB/s")
    band = lineEditBand->text().toDouble()*pow(2,10);
    else if(comboboxBand1->currentText() == "MB/s")
    band = lineEditBand->text().toDouble()*pow(2,20);
    else if(comboboxBand1->currentText() == "GB/s")
    band = lineEditBand->text().toDouble()*pow(2,30);*/
        
  if(scriptonly != true)
    fab = sstmac::fullnetwork::construct(lat, band);
  //fs << "ts = sstmac.timestamp(" << lat.sec() << ")\n";
  fs << "lat = " << lat.sec() << "\n";
  fs << "band = " << band << "\n";
  fs << "mem_band = " << mem_band << "\n";
  fs << "interconnect = sstmac.fullnetwork.construct(lat, band)\n";

  return true; //everything's good
}

bool parametertab::validateCrossbar(shared_ptr<sstmac::network> &net)
{
  if(!validateint(lineEditcrossnnodes->text())) {
    //lineEditcrossnnodes->selectAll();
    QMessageBox::information(this, "Invalid Input", "Num of Nodes requires an integer value.");
    return false;
  }
  int nnodes(lineEditcrossnnodes->text().toInt());
  net = crossbar::construct(nnodes);
                
  return true;
}

bool parametertab::validateFattree(shared_ptr<sstmac::network> &net)
{
  if(!validateLevels()) {
    return false;
  }
  if(!validateLinks()) {
    return false;
  }
  if(!validateswitch()) {
    return false;
  }
  if(!validatewidth()) {
    return false;
  }
  
  int levels(lineEditFatLevels->text().toInt());
  int links(lineEditFatLinks->text().toInt());
  int width(lineEditwidth->text().toInt());
  int nnps(lineEditnnodeperswitch->text().toInt());
  if(nnps < 0)
    nnps = 0;
  net = fattree::construct(levels, links, nnps, width);
  fs << "levels = " << levels << "\n"
     << "links = " << links << "\n"
     << "nnps = " << nnps << "\n"
     << "width = " << width << "\n"
     << "net = sstmac.fattree.construct(levels, links, nnps, width)\n";
        
  return true;
}

bool parametertab::validateGamma(shared_ptr<sstmac::network> &net)
{
  if(!validateDegree()) {
    return false;
  }
  if(!validateDiameter()) {
    return false;
  }
  if(!validateswitch()) {
    return false;
  }
  if(!validatewidth()) {
    return false;
  }
        
  int degree(lineEditGammaDegree->text().toInt());
  int diameter(lineEditGammaDiameter->text().toInt());
  int width(lineEditwidth->text().toInt());
  int nnps(lineEditnnodeperswitch->text().toInt());
  if(nnps < 0)
    nnps = 0;
  net = sstmac::gamma::construct(degree, diameter, nnps, width);
  
  fs << "degree = " << degree << "\n"
     << "diameter = " << diameter << "\n"
     << "nnps = " << nnps << "\n"
     << "width = " << width << "\n"
     << "net = sstmac.gamma.construct(levels, links, nnps, width)\n";
        
  return true;
}

bool parametertab::validateMesh2D(shared_ptr<sstmac::network> &net)
{
  if(!validateN1andN2()) {
    return false;
  }
  /*if(!validateswitch()) {
    return false;
    }
    if(!validatewidth()) {
    return false;
    }*/
        
  int n1(lineEditn1->text().toInt());
  int n2(lineEditn2->text().toInt());
  /*int width(lineEditwidth->text().toInt());
    int nnps(lineEditnnodeperswitch->text().toInt());
    if(nnps < 0)
    nnps = 0;*/
  net = composite::construct(
                             line::construct(n1),
                             line::construct(n2));
            
  fs << "n1 = " << n1 << "\n"
     << "n2 = " << n2 << "\n"
     << "net = sstmac.composite.construct(sstmac.line.construct(n1), sstmac.line.construct(n2))\n";
        
  return true;
}

bool parametertab::validateMesh3D(shared_ptr<sstmac::network> &net)
{
  if(!validateN1andN2()) {
    return false;
  }
  if(!validateN3()) {
    return false;
  }
        
  int n1(lineEditn1->text().toInt());
  int n2(lineEditn2->text().toInt());
  int n3(lineEditn3->text().toInt());

  net =   composite::construct(
                               line::construct(n1),
                               composite::construct(line::construct(n2),
                                                    ring::construct(n3)));
            
  fs << "n1 = " << n1 << "\n"
     << "n2 = " << n2 << "\n"
     << "n3 = " << n3 << "\n"
     << "net = sstmac.composite.construct(sstmac.line.construct(n1),"
     << " sstmac.composite.construct(sstmac.line.construct(n2),"
     << " sstmac.ring.construct(n3)))\n";
        
  return true;
}

bool parametertab::validateRing(shared_ptr<sstmac::network> &net)
{
  if(!validateN1()) {
    return false;
  }
  if(!validateswitch()) {
    return false;
  }
  if(!validatewidth()) {
    return false;
  }
        
  int n1(lineEditn1->text().toInt());
  int width(lineEditwidth->text().toInt());
  int nnps(lineEditnnodeperswitch->text().toInt());
  if(nnps < 0)
    nnps = 0;
  net = ring::construct(n1, nnps, width);
  
  fs << "n1 = " << n1 << "\n"
     << "nnps = " << nnps << "\n"
     << "width = " << width << "\n"
     << "net = sstmac.ring.construct(n1, nnps, width)\n";
                
  return true;
}

bool parametertab::validateTorus2D(shared_ptr<sstmac::network> &net)
{
  if(!validateN1andN2()) {
    return false;
  }
  if(!validateswitch()) {
    return false;
  }
  if(!validatewidth()) {
    return false;
  }
        
  int n1(lineEditn1->text().toInt());
  int n2(lineEditn2->text().toInt());
  int width(lineEditwidth->text().toInt());
  int nnps(lineEditnnodeperswitch->text().toInt());
  if(nnps < 0)
    nnps = 0;
  net = hdtorus::construct_2D(n1, n2, nnps, width);
  
  fs << "n1 = " << n1 << "\n"
     << "n2 = " << n2 << "\n"
     << "nnps = " << nnps << "\n"
     << "width = " << width << "\n"
     << "net = sstmac.hdtorus.construct_2D(n1, n2, nnps, width)\n";
                
  return true;
}

bool parametertab::validateTorus3D(shared_ptr<sstmac::network> &net)
{
  if(!validateN1andN2()) {
    return false;
  }
  if(!validateN3()) {
    return false;
  }
  if(!validateswitch()) {
    return false;
  }
  if(!validatewidth()) {
    return false;
  }
        
  int n1(lineEditn1->text().toInt());
  int n2(lineEditn2->text().toInt());
  int n3(lineEditn3->text().toInt());
  int width(lineEditwidth->text().toInt());
  int nnps(lineEditnnodeperswitch->text().toInt());
  if(nnps < 0)
    nnps = 0;
  net = hdtorus::construct_3D(n1, n2, n3, nnps, width);
  
  fs << "n1 = " << n1 << "\n"
     << "n2 = " << n2 << "\n"
     << "n3 = " << n3 << "\n"
     << "nnps = " << nnps << "\n"
     << "width = " << width << "\n"
     << "net = sstmac.hdtorus.construct_3D(n1, n2, n3, nnps, width)\n";
        
  return true;
}

bool parametertab::validateTorus4D(shared_ptr<sstmac::network> &net)
{
  if(!validateN1andN2()) {
    return false;
  }
  if(!validateN3()) {
    return false;
  }
  if(!validateNx()) {
    return false;
  }
  if(!validateswitch()) {
    return false;
  }
  if(!validatewidth()) {
    return false;
  }
        
  int n1(lineEditn1->text().toInt());
  int n2(lineEditn2->text().toInt());
  int n3(lineEditn3->text().toInt());
  int nx(lineEditnx->text().toInt());
  int width(lineEditwidth->text().toInt());
  int nnps(lineEditnnodeperswitch->text().toInt());
  if(nnps < 0)
    nnps = 0;
  net = hdtorus::construct_4D(n1, n2, n3, nx, nnps, width);
  
  fs << "n1 = " << n1 << "\n"
     << "n2 = " << n2 << "\n"
     << "n3 = " << n3 << "\n"
     << "nx = " << nx << "\n"
     << "nnps = " << nnps << "\n"
     << "width = " << width << "\n"
     << "net = sstmac.hdtorus.construct_4D(n1, n2, n3, nx, nnps, width)\n";
        
  return true;
}


bool parametertab::validatefields(shared_ptr<interconnect> &fab, shared_ptr<network> &net, int &nnodes, int &ppn)
{
  double band;
  timestamp lat;

  if(interTabWidget->currentIndex() == 1)
    {
      if(!validateFullnetwork(fab)){
        std::cout << "validateFullnetwork failed\n";
        return false;
      }
                
      if(!validateFullMemBand()){
        std::cout << "validateFullMemBand failed\n";
        return false; 
      }
                
    }
  else //if(interconnectComboBox->currentText() == "Circuit")
    {
      if(!validateLatBandMem(lat, band)){
        std::cout << "validateLatBandMem failed\n";
        return false; 
      }
                
      if(!validateMemBand()){
        std::cout << "validateMemBand failed\n";
        return false; 
      }
                        
      if(topologyComboBox->currentText() == "Crossbar")
        {
          if( !validateCrossbar(net) )
            return false;
        }
      else if(topologyComboBox->currentText() == "Fat Tree")
        {
          if( !validateFattree(net)) {
            return false;
          }
        }
      else if(topologyComboBox->currentText() == "Gamma")
        {
          if(!validateGamma(net))
            return false;
        }
      else if(topologyComboBox->currentText() == "Mesh2D")
        {
          if(!validateMesh2D(net))
            return false;
        }
      else if(topologyComboBox->currentText() == "Mesh3D")
        {
          if(!validateMesh3D(net))
            return false;
        }
      else if(topologyComboBox->currentText() == "Ring")
        {
          if(!validateRing(net))
            return false;
        }
      else if(topologyComboBox->currentText() == "Torus2D")
        {
          if(!validateTorus2D(net))
            return false;
        }
      else if(topologyComboBox->currentText() == "Torus3D")
        {
          if(!validateTorus3D(net))
            return false;
        }       
      else if(topologyComboBox->currentText() == "Torus4D")
        {
          if(!validateTorus4D(net))
            return false;
        }       
                
      if(scriptonly != true) 
        fab = sstmac::circuit::construct(lat, band, net);
      fs << "interconnect = sstmac.circuit.construct(sstmac.timestamp(lat), band, net)\n";
                
    }
  if(!validatennodes(nnodes)){
    std::cout << "validatennodes failed\n";
    return false;
  }
  if(!validateppn(ppn)) {
    std::cout << "validateppn failed\n";
    return false;
  }
                        
  return true;
}

void parametertab::createParameterGroupBox()
{
  parameterGroupBox = new QGroupBox(tr("Network Parameters"));
        
  labelLatency = new QLabel(tr("Latency"));

  labelLatency->setAlignment(Qt::AlignRight);

  labelBandwidth = new QLabel(tr("Bandwidth"));
  labelBandwidth->setAlignment(Qt::AlignRight);
        
  labelIntraLat = new QLabel(tr("Intranode:"));
  labelIntraLat->setAlignment(Qt::AlignRight);
  labelInterLat = new QLabel(tr("Latency (sec):"));
  labelInterLat->setAlignment(Qt::AlignRight);
  labelIntraBand = new QLabel(tr("Intranode:"));
  labelIntraBand->setAlignment(Qt::AlignRight);
  labelInterBand = new QLabel(tr("Bandwidth (Bps):"));
  labelInterBand->setAlignment(Qt::AlignRight);
        
                
  lineEditIntraLat = new QLineEdit;
  lineEditInterLat = new QLineEdit;
  lineEditIntraBand = new QLineEdit;
  lineEditInterBand = new QLineEdit;
  //      lineEditNumNodes = new QLineEdit;
  //      lineEditPPN = new QLineEdit;
  lineEditIntraLat->setFrame(false);
  lineEditIntraLat->setText("1.0");
  lineEditInterLat->setFrame(false);
  lineEditInterLat->setText("1.3e-06");
  lineEditIntraBand->setFrame(false);
  lineEditIntraBand->setText("1.2");
  lineEditInterBand->setFrame(false);
  lineEditInterBand->setText("1.474e+09");
  //      lineEditNumNodes->setFrame(false);
  //      lineEditPPN->setFrame(false);
                
  comboboxLat = new QComboBox;
  comboboxLat->addItem(tr("Seconds"));
  comboboxLat->addItem(tr("Milliseconds"));
  comboboxLat->addItem(tr("Microseconds"));
  comboboxLat->addItem(tr("Nanoseconds"));
  comboboxLat->setCurrentIndex(2);
        
  comboboxLat1 = new QComboBox;
  comboboxLat1->addItem(tr("Seconds"));
  comboboxLat1->addItem(tr("Milliseconds"));
  comboboxLat1->addItem(tr("Microseconds"));
  comboboxLat1->addItem(tr("Nanoseconds"));
  comboboxLat1->setCurrentIndex(2);
        
  comboboxBand = new QComboBox;
  comboboxBand->addItem(tr("Bytes/s"));
  comboboxBand->addItem(tr("KB/s"));
  comboboxBand->addItem(tr("MB/s"));
  comboboxBand->addItem(tr("GB/s"));
  comboboxBand->setCurrentIndex(3);
        
  comboboxBand1 = new QComboBox;
  comboboxBand1->addItem(tr("Bytes/s"));
  comboboxBand1->addItem(tr("KB/s"));
  comboboxBand1->addItem(tr("MB/s"));
  comboboxBand1->addItem(tr("GB/s"));
  comboboxBand1->setCurrentIndex(3);

}

void parametertab::circuitview()
{
  labelLatency->show();
  labelBandwidth->show();
  labelIntraLat->show();
  labelInterLat->show();
  labelIntraBand->show();
  labelInterBand->show();
  lineEditInterLat->show();
  lineEditInterBand->show();
  lineEditIntraLat->show();
  lineEditIntraBand->show();
  comboboxLat->show();
  comboboxBand->show();
}

void parametertab::fullnetview()
{
  labelLatency->show();
  labelBandwidth->show();
  labelIntraLat->hide();
  labelInterLat->hide();
  labelIntraBand->hide();
  labelInterBand->hide();
  lineEditInterLat->show();
  lineEditInterBand->show();
  lineEditIntraLat->hide();
  lineEditIntraBand->hide();
  comboboxLat->show();
  comboboxBand->show();
}

void parametertab::test()
{
  child = new QTextEdit;
  mdiArea->addSubWindow(child);
  show();
}

void parametertab::enableFields(int box)
{
        
  if(box == 0) {
        
    topologyComboBox->show();       
    circuitview();
    if(topologyComboBox->currentText() == "Fat Tree")
      {
        labelFatLevels->show();
        labelFatLinks->show();
        lineEditFatLevels->show();
        lineEditFatLinks->show();       
        labelnnodeperswitch->show();
        labelwidth->show();
        lineEditnnodeperswitch->show();
        lineEditwidth->show();
      }
    else if(topologyComboBox->currentText() == "Mesh2D")
      {
        labeln1->show();
        lineEditn1->show();
        labeln2->show();
        lineEditn2->show();
        /*labelnnodeperswitch->show();
          labelwidth->show();
          lineEditnnodeperswitch->show();
          lineEditwidth->show();*/
      }
    else if(topologyComboBox->currentText() == "Mesh3D")
      {
        labeln1->show();
        lineEditn1->show();
        labeln2->show();
        lineEditn2->show();     
        labeln3->show();
        lineEditn3->show();
        /*labelnnodeperswitch->show();
          labelwidth->show();
          lineEditnnodeperswitch->show();
          lineEditwidth->show();*/
      }
    else if(topologyComboBox->currentText() == "Gamma")
      {
        labelGammaDegree->show();
        labelGammaDiameter->show();
        lineEditGammaDegree->show();
        lineEditGammaDiameter->show();
        labelnnodeperswitch->show();
        labelwidth->show();
        lineEditnnodeperswitch->show();
        lineEditwidth->show();
      }               
    else if(topologyComboBox->currentText() == "Torus2D")
      {
        labeln1->show();
        lineEditn1->show();
        labeln2->show();
        lineEditn2->show();
        labelnnodeperswitch->show();
        labelwidth->show();
        lineEditnnodeperswitch->show();
        lineEditwidth->show();
      }
    else if(topologyComboBox->currentText() == "Torus3D")
      {
        labeln1->show();
        lineEditn1->show();
        labeln2->show();
        lineEditn2->show();     
        labeln3->show();
        lineEditn3->show();
        labelnnodeperswitch->show();
        labelwidth->show();
        lineEditnnodeperswitch->show();
        lineEditwidth->show();
      }
    else if(topologyComboBox->currentText() == "Torus4D")
      {
        labeln1->show();
        lineEditn1->show();
        labeln2->show();
        lineEditn2->show();     
        labeln3->show();
        lineEditn3->show();
        labelnx->show();
        lineEditnx->show();
        labelnnodeperswitch->show();
        labelwidth->show();
        lineEditnnodeperswitch->show();
        lineEditwidth->show();
      }
    else if(topologyComboBox->currentText() == "Ring")
      {
        labeln1->show();
        lineEditn1->show();
        labelnnodeperswitch->show();
        labelwidth->show();
        lineEditnnodeperswitch->show();
        lineEditwidth->show();
      }
    else if(topologyComboBox->currentText() == "Crossbar")
      {
        labelcrossnnodes->show();
        lineEditcrossnnodes->show();    
      }
  }
  else if(box == 1){
                
    topologyComboBox->hide();
    fullnetview();
    if(topologyComboBox->currentText() == "Fat Tree")
      {
        labelFatLevels->hide();
        labelFatLinks->hide();
        lineEditFatLevels->hide();
        lineEditFatLinks->hide();
      }
    else if(topologyComboBox->currentText() == "Mesh2D")
      {
        labeln1->hide();
        lineEditn1->hide();
        labeln2->hide();
        lineEditn2->hide();     
      }
    else if(topologyComboBox->currentText() == "Mesh3D")
      {
        labeln1->hide();
        lineEditn1->hide();
        labeln2->hide();
        lineEditn2->hide();     
        labeln3->hide();
        lineEditn3->hide();
      }
    else if(topologyComboBox->currentText() == "Gamma")
      {
        labelGammaDegree->hide();
        labelGammaDiameter->hide();
        lineEditGammaDegree->hide();
        lineEditGammaDiameter->hide();
      }               
    else if(topologyComboBox->currentText() == "Torus2D")
      {
        labeln1->hide();
        lineEditn1->hide();
        labeln2->hide();
        lineEditn2->hide();
        labelnnodeperswitch->hide();
        labelwidth->hide();
        lineEditnnodeperswitch->hide();
        lineEditwidth->hide();  
      }
    else if(topologyComboBox->currentText() == "Torus3D")
      {
        labeln1->hide();
        lineEditn1->hide();
        labeln2->hide();
        lineEditn2->hide();     
        labeln3->hide();
        lineEditn3->hide();
        labelnnodeperswitch->hide();
        labelwidth->hide();
        lineEditnnodeperswitch->hide();
        lineEditwidth->hide();  
      }
    else if(topologyComboBox->currentText() == "Torus4D")
      {
        labeln1->hide();
        lineEditn1->hide();
        labeln2->hide();
        lineEditn2->hide();     
        labeln3->hide();
        lineEditn3->hide();
        labelnx->hide();
        lineEditnx->hide();
        labelnnodeperswitch->hide();
        labelwidth->hide();
        lineEditnnodeperswitch->hide();
        lineEditwidth->hide();  
      }
    else if(topologyComboBox->currentText() == "Ring")
      {
        labeln1->hide();
        lineEditn1->hide();
        labelnnodeperswitch->hide();
        labelwidth->hide();
        lineEditnnodeperswitch->hide();
        lineEditwidth->hide();  
      }
    else if(topologyComboBox->currentText() == "Crossbar")
      {
        labelcrossnnodes->hide();
        lineEditcrossnnodes->hide();    
      }
  }
}

void parametertab::typeChanged(int index)
{
  switch (index) {
  case 0:
    labelIteration->show();
    labeldatalength->show();
    lineEditIter->show();
    lineEditPacket->show();
    labelTimeScale->hide();
    lineEditTimeScale->hide();
    lineeditFileName->hide();
    buttonBrowse->hide();
    break;
  case 1:
    labelIteration->hide();
    labeldatalength->hide();
    lineEditIter->hide();
    lineEditPacket->hide();
    labelTimeScale->show();
    lineEditTimeScale->show();
    lineeditFileName->show();
    buttonBrowse->show();
    break;
  }
}

void parametertab::turnfieldson(NETTYPE n)
{
  if(n == FATTREE)
    {
      labelFatLevels->show();
      labelFatLinks->show();
      labelnnodeperswitch->show();
      labelwidth->show();
      lineEditFatLevels->show();
      lineEditFatLinks->show();       
      lineEditnnodeperswitch->show();
      lineEditwidth->show();  
    }
}       

void parametertab::turnfieldsoff()
{
  labelFatLevels->hide();
  labelFatLinks->hide();
  lineEditFatLevels->hide();
  lineEditFatLinks->hide();
  labelGammaDegree->hide();
  labelGammaDiameter->hide();
  lineEditGammaDegree->hide();
  lineEditGammaDiameter->hide();
  labeln1->hide();
  lineEditn1->hide();
  labeln2->hide();
  lineEditn2->hide();     
  labeln3->hide();
  lineEditn3->hide();
  labelnx->hide();
  lineEditnx->hide();
  labelnnodeperswitch->hide();
  labelwidth->hide();
  lineEditnnodeperswitch->hide();
  lineEditwidth->hide();
  labelcrossnnodes->hide();
  lineEditcrossnnodes->hide();
}

void parametertab::topologyChanged(int index)
{
  turnfieldsoff();
                
  switch (index) {
  case CROSSBAR:
    labelcrossnnodes->show();
    lineEditcrossnnodes->show();
    break;
  case FATTREE:
    labelFatLevels->show();
    labelFatLinks->show();
    labelnnodeperswitch->show();
    labelwidth->show();
    lineEditFatLevels->show();
    lineEditFatLinks->show();       
    lineEditnnodeperswitch->show();
    lineEditwidth->show();  
    break;
  case GAMMA:
    labelGammaDegree->show();
    labelGammaDiameter->show();
    labelnnodeperswitch->show();
    labelwidth->show();
    lineEditGammaDegree->show();
    lineEditGammaDiameter->show();
    lineEditnnodeperswitch->show();
    lineEditwidth->show();
    break;
  case MESH2D:
    labeln1->show();
    lineEditn1->show();
    labeln2->show();
    lineEditn2->show();     
    break;
  case MESH3D:
    labeln1->show();
    lineEditn1->show();
    labeln2->show();
    lineEditn2->show();     
    labeln3->show();
    lineEditn3->show();
    break;
  case RING:
    labeln1->show();
    lineEditn1->show();     
    labelnnodeperswitch->show();
    labelwidth->show();
    lineEditnnodeperswitch->show();
    lineEditwidth->show();
    break;
  case TORUS2D:
    labeln1->show();
    lineEditn1->show();
    labeln2->show();
    lineEditn2->show();     
    labelnnodeperswitch->show();
    labelwidth->show();
    lineEditnnodeperswitch->show();
    lineEditwidth->show();
    break;
  case TORUS3D:
    labeln1->show();
    lineEditn1->show();
    labeln2->show();
    lineEditn2->show();     
    labeln3->show();
    lineEditn3->show();
    labelnnodeperswitch->show();
    labelwidth->show();
    lineEditnnodeperswitch->show();
    lineEditwidth->show();
    break;
  case TORUS4D:
    labeln1->show();
    lineEditn1->show();
    labeln2->show();
    lineEditn2->show();     
    labeln3->show();
    lineEditn3->show();
    labelnx->show();
    lineEditnx->show();
    labelnnodeperswitch->show();
    labelwidth->show();
    lineEditnnodeperswitch->show();
    lineEditwidth->show();
    break;
  }
}

void parametertab::createFullIndexGroupBox()
{
  indexGroupBox = new QGroupBox(tr("Indexing"));
  indexComboBox = new QComboBox;
  indexComboBox->addItem(tr("Block"));
  indexComboBox->addItem(tr("Random"));
  indexComboBox->addItem(tr("Round robin"));
        
  /*QHBoxLayout *indexbox = new QHBoxLayout;
    indexbox->addWidget(indexComboBox);
        
    indexGroupBox->setLayout(indexbox);*/
}

void parametertab::createFullAllocationGroupBox()
{
  //allocationGroupBox = new QGroupBox(tr("Allocation"));
  allocationComboBox = new QComboBox;
        
  allocationComboBox->addItem(tr("First available"));
  allocationComboBox->addItem(tr("Random"));
  /*      
    QHBoxLayout *allocationbox = new QHBoxLayout;
    allocationbox->addWidget(allocationComboBox);
        
    allocationGroupBox->setLayout(allocationbox);*/
}

void parametertab::createCircuitIndexGroupBox()
{
  //indexGroupBox = new QGroupBox(tr("Indexing"));
  indexCircuitComboBox = new QComboBox;
  indexCircuitComboBox->addItem(tr("Block"));
  indexCircuitComboBox->addItem(tr("Random"));
  indexCircuitComboBox->addItem(tr("Round robin"));
        
  /*      QHBoxLayout *indexbox = new QHBoxLayout;
          indexbox->addWidget(indexComboBox);
        
          indexGroupBox->setLayout(indexbox);*/
}

void parametertab::createCircuitAllocationGroupBox()
{
  //allocationGroupBox = new QGroupBox(tr("Allocation"));
  allocationCircuitComboBox = new QComboBox;
  allocationCircuitComboBox->addItem(tr("Random"));
  allocationCircuitComboBox->addItem(tr("First available"));
  /*      
          QHBoxLayout *allocationbox = new QHBoxLayout;
          allocationbox->addWidget(allocationComboBox);
        
          allocationGroupBox->setLayout(allocationbox);*/
}

void parametertab::createProcsGroupBox()
{
  /*procsGroupBox = new QGroupBox();
    QLabel *labelnnodes = new QLabel(tr("Num of nodes:"));
    lineeditnnodes = new QLineEdit();
    QLabel *labelprocspernode = new QLabel(tr("Procs per node:"));
    lineeditprocspernode = new QLineEdit();
        
    QGridLayout *procsbox = new QGridLayout;
    procsbox->addWidget(labelnnodes,0,0);
    procsbox->addWidget(lineeditnnodes,0,1);
    procsbox->addWidget(labelprocspernode,1,0);
    procsbox->addWidget(lineeditprocspernode,1,1);
        
    procsGroupBox->setLayout(procsbox);*/
}

void parametertab::createNetworkGroupBox()
{       

  networkGroupBox = new QWidget;
  fullGroupBox = new QWidget;
  interTabWidget = new QTabWidget;
        
  interconnectComboBox = new QComboBox;
  interconnectComboBox->addItem(tr("Circuit"));
  interconnectComboBox->addItem(tr("Full network"));
  interconnectComboBox->setCurrentIndex(0);
                
  connect(interconnectComboBox, SIGNAL(activated(int)),
          this, SLOT(enableFields(int)));
                                
  topologyComboBox = new QComboBox;
  topologyComboBox->addItem(tr("Crossbar"));
  topologyComboBox->addItem(tr("Fat Tree"));
  topologyComboBox->addItem(tr("Gamma"));
  topologyComboBox->addItem(tr("Mesh2D"));
  topologyComboBox->addItem(tr("Mesh3D"));
  topologyComboBox->addItem(tr("Ring"));
  topologyComboBox->addItem(tr("Torus2D"));
  topologyComboBox->addItem(tr("Torus3D"));
  topologyComboBox->addItem(tr("Torus4D"));
  topologyComboBox->setCurrentIndex(1);
                
  connect(topologyComboBox, SIGNAL(activated(int)),
          this, SLOT(topologyChanged(int)));

  labelFatLevels = new QLabel(tr("Levels:"));
  labelFatLevels->setAlignment(Qt::AlignRight);
  labelFatLinks = new QLabel(tr("Links:"));
  labelFatLinks->setAlignment(Qt::AlignRight);
  lineEditFatLevels = new QLineEdit("4");
  lineEditFatLevels->setFrame(false);
  lineEditFatLinks = new QLineEdit("4");
  lineEditFatLinks->setFrame(false);

  labelGammaDegree = new QLabel(tr("Degree:"));
  labelGammaDegree->setAlignment(Qt::AlignRight);
  labelGammaDiameter = new QLabel(tr("Diameter:"));
  labelGammaDiameter->setAlignment(Qt::AlignRight);
  lineEditGammaDegree = new QLineEdit("4");
  lineEditGammaDegree->setFrame(false);
  lineEditGammaDiameter = new QLineEdit("3");
  lineEditGammaDiameter->setFrame(false);
                
  labeln1 = new QLabel(tr("N1:"));
  labeln1->setAlignment(Qt::AlignRight);
  lineEditn1 = new QLineEdit;
  lineEditn1->setFrame(false);
  lineEditn1->setText("2");
  labeln2 = new QLabel(tr("N2:"));
  labeln2->setAlignment(Qt::AlignRight);
  lineEditn2 = new QLineEdit;
  lineEditn2->setFrame(false);
  lineEditn2->setText("2");
  labeln3 = new QLabel(tr("N3:"));
  labeln3->setAlignment(Qt::AlignRight);
  lineEditn3 = new QLineEdit;
  lineEditn3->setFrame(false);
  lineEditn3->setText("2");
  labelnx = new QLabel(tr("Nx:"));
  labelnx->setAlignment(Qt::AlignRight);
  lineEditnx = new QLineEdit;
  lineEditnx->setFrame(false);
  lineEditnx->setText("2");
                
  labelcrossnnodes = new QLabel(tr("Num of Nodes:"));
  labelcrossnnodes->setAlignment(Qt::AlignRight);
  lineEditcrossnnodes = new QLineEdit;
  lineEditcrossnnodes->setFrame(false);
  lineEditcrossnnodes->setText("2");
                
  labelnnodeperswitch = new QLabel(tr("Nodes per switch:"));
  labelnnodeperswitch->setAlignment(Qt::AlignRight);
  lineEditnnodeperswitch = new QLineEdit;
  lineEditnnodeperswitch->setFrame(false);
  lineEditnnodeperswitch->setText("1");
  labelwidth = new QLabel(tr("Width:"));
  labelwidth->setAlignment(Qt::AlignRight);
  lineEditwidth = new QLineEdit;
  lineEditwidth->setFrame(false);
  lineEditwidth->setText("1");

  turnfieldsoff();
  turnfieldson(FATTREE);
                
  QGridLayout *fullnetbox = new QGridLayout;
  labelLat = new QLabel(tr("Latency (sec):"));
  labelLat->setAlignment(Qt::AlignRight);
  labelBand = new QLabel(tr("Bandwidth (Bps):"));
  labelBand->setAlignment(Qt::AlignRight);
  lineEditLat = new QLineEdit;
  lineEditLat->setFrame(false);
  lineEditLat->setText("1.3e-06");
  lineEditBand = new QLineEdit;
  lineEditBand->setFrame(false);
  lineEditBand->setText("1.474e+09");
                
  QLabel *labelnnodes = new QLabel(tr("Num of nodes:"));
  labelnnodes->setAlignment(Qt::AlignRight);
  lineeditnnodes = new QLineEdit();
  QLabel *labelprocspernode = new QLabel(tr("Procs per node:"));
  labelprocspernode->setAlignment(Qt::AlignRight);
  lineeditprocspernode = new QLineEdit();
        
  QLabel *labelFullMemBand = new QLabel(tr("Memory Bandwidth (Bps):"));
  labelFullMemBand->setAlignment(Qt::AlignRight);
  lineEditFullMemBand = new QLineEdit("1e+10");
  lineEditFullMemBand->setFrame(false);
                
  QPushButton *buttonRunFull = new QPushButton(tr("&Run Full Network Simulation"));
                
  fullnetbox->addWidget(labelInterLat,0,0);
  fullnetbox->addWidget(lineEditInterLat,0,1);
  fullnetbox->addWidget(labelInterBand,0,2);
  fullnetbox->addWidget(lineEditInterBand,0,3);
                
  fullnetbox->addWidget(labelFullMemBand,1,0);
  fullnetbox->addWidget(lineEditFullMemBand,1,1);
  fullnetbox->addWidget(buttonRunFull,2,0);
                
  QPushButton *buttonGenPython1 = new QPushButton(tr("&Just Generate Python Script"));
  fullnetbox->addWidget(buttonGenPython1,2,1);
                
  connect(buttonRunFull, SIGNAL(clicked()),
          this, SLOT(runsim()));
                
  QGridLayout *networkbox = new QGridLayout;
                
  networkbox->addWidget(topologyComboBox,1,0);

  networkbox->addWidget(labelFatLevels,1,1);
  networkbox->addWidget(lineEditFatLevels,1,2);
  networkbox->addWidget(labelFatLinks,1,3);
  networkbox->addWidget(lineEditFatLinks,1,4);

  networkbox->addWidget(labelGammaDegree,1,1);
  networkbox->addWidget(lineEditGammaDegree,1,2);
  networkbox->addWidget(labelGammaDiameter,1,3);
  networkbox->addWidget(lineEditGammaDiameter,1,4);

  networkbox->addWidget(labeln1,1,1);
  networkbox->addWidget(lineEditn1,1,2);
  networkbox->addWidget(labeln2,1,3);
  networkbox->addWidget(lineEditn2,1,4);
  networkbox->addWidget(labeln3,1,5);
  networkbox->addWidget(lineEditn3,1,6);
                
  networkbox->addWidget(labelcrossnnodes,1,1);
  networkbox->addWidget(lineEditcrossnnodes,1,2);
                
  networkbox->addWidget(labelnnodeperswitch,2,1);
  networkbox->addWidget(lineEditnnodeperswitch,2,2);
  networkbox->addWidget(labelwidth,2,3);
  networkbox->addWidget(lineEditwidth,2,4);
  networkbox->addWidget(labelnx,2,5);
  networkbox->addWidget(lineEditnx,2,6);
                
  networkbox->addWidget(labelLat,3,1);
  networkbox->addWidget(lineEditLat,3,2);
  //networkbox->addWidget(comboboxLat1,3,4);
  networkbox->addWidget(labelBand,3,3);
  networkbox->addWidget(lineEditBand,3,4);
  //networkbox->addWidget(comboboxBand1,4,4);
                
  QLabel *labelMemBand = new QLabel(tr("Memory Bandwidth (Bps):"));
  labelMemBand->setAlignment(Qt::AlignRight);
  lineEditMemBand = new QLineEdit("1e+10");
  lineEditMemBand->setFrame(false);
                
  networkbox->addWidget(labelMemBand,4,1);
  networkbox->addWidget(lineEditMemBand,4,2);
                
  QPushButton *buttonRunCircuit = new QPushButton(tr("&Run Circuit Simulation"));
  networkbox->addWidget(buttonRunCircuit,6,0);
                
                
  connect(buttonRunCircuit, SIGNAL(clicked()),
          this, SLOT(runsim()));
            
  QPushButton *buttonGenPython2 = new QPushButton(tr("&Just Generate Python Script"));
  networkbox->addWidget(buttonGenPython2,6,1);
                
  connect(buttonGenPython1, SIGNAL(clicked()),
          this, SLOT(genpythonscript()));
  connect(buttonGenPython2, SIGNAL(clicked()),
          this, SLOT(genpythonscript()));
                
  networkGroupBox->setLayout(networkbox);
  fullGroupBox->setLayout(fullnetbox);
  interTabWidget->addTab(networkGroupBox, tr("Circuit"));
  interTabWidget->addTab(fullGroupBox, tr("Full Network"));
}

void parametertab::genpythonscript()
{
  scriptonly = true;
  runsim();
  scriptonly = false;
}



