/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// gtest.cc: Driver code for testing networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#include <cstdlib>
#include <fstream>
#include <iomanip>
#include <algorithm>
#include <sstmac/network/gamma.h>
#include <sstmac/network/torus.h>
#include <sstmac/network/ring.h>
#include <sstmac/network/hypercube.h>
#include <sstmac/network/fattree.h>
#include <sstmac/network/clos.h>
#include <sstmac/network/tbird.h>
#include <sstmac/network/crossbar.h>
#include <sstmac/util/rng.h>
#include <cmath>

using std::ofstream;
using std::set;
using std::atoi;
using std::cout;
using std::endl;

void
help(const char *argv0)
{
  std::cout << "Usage: " << argv0
            << " [-n name] [-d] [-h]"
            << std::endl;
  std::cout
      << "The options are:" << std::endl
      << "  -h      Print help information and exit." << std::endl
      << "  -d      Write a dot file to net.dot." << std::endl
      << "  -n name Create the named network." << std::endl
      << "Examples of network names are:" << std::endl
      << "  gamma(degree,diameter[,nnodeperswitch[,width]])" << std::endl
      << "  hybercube(dimension)" << std::endl
      << "  ring(nx[,nnodeperswitch[,width]])" << std::endl
      << "  torus2D(nx,ny[,nnodeperswitch[,width]])" << std::endl
      << "  torus3D(nx,ny,nz[,nnodeperswitch[,width]])" << std::endl
      << "  crossbar(nnode)" << std::endl
      << "  fattree(nlevel,half_degree[,nnodeperswitch[,width]])" << std::endl
      << "  clos(half_degree)" << std::endl
      << "  tbird" << std::endl
      << "If no networks are specified, then all except rings will be used."
      << std::endl;
}

static double
average(const std::vector<int> &d)
{
  double a = 0.0;
  for (int i=0; i<d.size(); i++) a+=d[i];
  return a/d.size();
}

static double
stddev(const std::vector<int> &d)
{
  double a = average(d);
  double s = 0.0;
  for (int i=0; i<d.size(); i++) s+=(a-d[i])*(a-d[i]);
  return std::sqrt(s/(d.size()-1));
}

std::string
get_token(std::string &tokens, int *narg = 0)
{
  std::string result;

  if (tokens.size() != 0) {
    size_t pos = tokens.find_first_of("_(),");

    if (pos == std::string::npos) {
      result = tokens;
      tokens.clear();
    }
    else {
      result = tokens.substr(0,pos);
      tokens = tokens.substr(pos+1,std::string::npos);
    }
  }

  if (result.size() > 0 && narg) *narg++;

  return result;
}

boost::shared_ptr<sstmac::network>
construct_network(const char *argv0, const std::string &netname)
{
  std::string remaining_tokens = netname;
  std::string netbase = get_token(remaining_tokens);
  int narg = 0;
  std::string token;
  std::vector<int> args;
  while ((token = get_token(remaining_tokens, &narg)).size() > 0) {
    args.push_back(atoi(token.c_str()));
  }

  boost::shared_ptr<sstmac::network> net;
  if (netbase == "gamma") {
    if (args.size() == 2) {
      net = sstmac::gamma::construct(args[0],args[1]);
    }
    else if (args.size() == 3) {
      net = sstmac::gamma::construct(args[0],args[1],args[2]);
    }
    else if (args.size() == 4) {
      net = sstmac::gamma::construct(args[0],args[1],args[2],args[3]);
    }
  }
  else if (netbase == "hypercube") {
    if (args.size() == 1) {
      net = sstmac::hypercube::construct(args[0]);
    }
  }
  else if (netbase == "torus2D") {
    if (args.size() == 2) {
      net = sstmac::torus::construct_2D(args[0],args[1]);
    }
    else if (args.size() == 3) {
      net = sstmac::torus::construct_2D(args[0],args[1],args[2]);
    }
    else if (args.size() == 4) {
      net = sstmac::torus::construct_2D(args[0],args[1],args[2],args[3]);
    }
  }
  else if (netbase == "torus3D") {
    if (args.size() == 3) {
      net = sstmac::torus::construct_3D(args[0],args[1],args[2]);
    }
    else if (args.size() == 4) {
      net = sstmac::torus::construct_3D(args[0],args[1],args[2],args[3]);
    }
    else if (args.size() == 5) {
      net = sstmac::torus::construct_3D(args[0],args[1],args[2],args[3],args[4]);
    }
  }
  else if (netbase == "ring") {
    if (args.size() == 1) {
      net = sstmac::ring::construct(args[0]);
    }
    else if (args.size() == 2) {
      net = sstmac::ring::construct(args[0],args[1]);
    }
    else if (args.size() == 3) {
      net = sstmac::ring::construct(args[0],args[1],args[2]);
    }
  }
  else if (netbase == "fattree") {
    if (args.size() == 2) {
      net = sstmac::fattree::construct(args[0],args[1]);
    }
    if (args.size() == 3) {
      net = sstmac::fattree::construct(args[0],args[1],args[2]);
    }
    if (args.size() == 4) {
      net = sstmac::fattree::construct(args[0],args[1],args[2],args[3]);
    }
  }
  else if (netbase == "crossbar") {
    if (args.size() == 1) {
      net = sstmac::crossbar::construct(args[0]);
    }
  }
  else if (netbase == "clos") {
    if (args.size() == 1) {
      net = sstmac::clos::construct(args[0]);
    }
  }
  else if (netbase == "tbird") {
    if (args.size() == 0) {
      net = sstmac::tbird::construct();
    }
  }

  if (net.get() == 0) {
    help(argv0);
    exit(1);
  }

  return net;
}

void
follow_hops(int step, const boost::shared_ptr<sstmac::network>& net,
            sstmac::vertex_index loc, sstmac::vertex_index dst)
{
  if (loc == dst) return;
  std::vector<sstmac::lwedge> nxt;
  try {
    net->next_hop(loc,dst,nxt);
  }
  catch (std::logic_error &e) {
    std::cout << "  next hop failed: " << e.what() << std::endl;
    return;
  }
  std::cout << "  step " << std::setw(4) << step
            << ": next hop from " << std::setw(4) << loc.value()
            << " to " << std::setw(4) << dst.value()
            << ":";
  for (int i=0; i<nxt.size(); i++) {
    std::cout << " " << std::setw(4) << nxt[i].to();
    if (nxt[i].from() != loc) {
      std::cout << "(bad loc in edge)";
    }
    try {
      net->g().edge_to_index(nxt[i]);
    }
    catch (...) {
      std::cout << "(bad edge)";
    }
  }
  std::cout << std::endl;
  for (int i=0; i<nxt.size(); i++) {
    follow_hops(step+1, net, nxt[i].to(), dst);
  }
}

int
main(int argc, char **argv)
{
  std::string netname = "fattree(3,4)";
  if (argc > 1) {
    netname = argv[1];
  }

  bool verbose = false;
  bool print_dot = false;
  bool show_help = false;
  int rc = 0;

  while(1) {
    int c = getopt(argc, argv, "vhn:d");
    if (c < 0) break;
    switch (c) {
    case 'h':
      show_help = true;
      break;
    case 'v':
      verbose = true;
      break;
    case 'd':
      print_dot = true;
      break;
    case 'n':
      netname = optarg;
      break;
    default:
      show_help = true;
      rc = 1;
    }
  }

  if (show_help) {
    help(argv[0]);
    exit(rc);
  }

  boost::shared_ptr<sstmac::network> net;
  net = construct_network(argv[0],netname);

  net->g().print();
  net->g().print_all(std::cout, verbose);

  ofstream dot("net.dot");
  net->g().print_dot(dot);

  sstmac::walk w;
  const set<sstmac::vertex> &v = net->nodes();
  for (set<sstmac::vertex>::const_iterator j = v.begin();
       j != v.end();
       j++) {
      const sstmac::vertex &v0 = *j;
      for (set<sstmac::vertex>::const_iterator i = v.begin();
           i != v.end();
           i++) {
          net->route(v0,*i,w);
          cout << "route from " << v0 << " to " << *i << ": ";
          if (w.ev().size()) {
            cout << w.ev()[0].from();
            for (int k=0; k<w.ev().size(); k++) {
              const sstmac::lwedge &e = w.ev()[k];
              cout << " (" << e.id();
              try {
                sstmac::edge_index ei = net->g().edge_to_index(e);
                if (net->edge_width(ei) != 1) {
                  cout << "," << net->edge_width(ei);
                }
              }
              catch (...) {
                cout << ",BAD";
              }
              cout << ")-> " << e.to();
            }
          }
          cout << endl;
          net->g().check_walk(v0,*i,w);
        }
    }

  boost::shared_ptr<sstmac::RNG::UniformInteger>
    rng = sstmac::RNG::SimpleCombo::construct();
  sstmac::RNG::UniformInteger_functor rngf(rng);
  int order = net->g().order();
  if (order > 0) {
    for (int i=0; i<10; i++) {
      sstmac::vertex_index loc(rngf(order)), dst(rngf(order));
      std::cout << "hopping from " << loc << " to " << dst << ":"
                << std::endl;
      follow_hops(1, net,loc,dst);
    }
  }

  if (verbose) {
    bool tight_packing = false;
    int maxb = 50;
    int maxp = -1;
    std::vector<int> sc(maxb);
    net->stochastic_max_link_load(sc, maxp, tight_packing);
    int minsc = *std::min_element(sc.begin(), sc.end());
    int maxsc = *std::max_element(sc.begin(), sc.end());
    double avesc = average(sc);
    double devsc = stddev(sc);
    int maxp_p = (maxp==-1?net->n():maxp);
    if (maxp_p > net->n()) maxp_p = net->n();
    std::cout.setf(std::ios_base::fixed, std::ios_base::floatfield);
    std::cout << std::setw(20) << net->name()
              << std::setw(6) << maxp_p
              << std::setw(1) << (tight_packing?'p':' ')
              << std::setw(6) << net->n()
              << std::setw(8) << net->nswitch()
              << std::setw(6) << net->nlink()
              << std::setw(4) << net->d()
              << std::setw(4) << net->diameter()
              << std::setw(8) << std::setprecision(4) << net->a()
              << std::setw(4) << maxsc
              << std::setw(8) << std::setprecision(4) << avesc
              << std::setw(8) << std::setprecision(4) << devsc
              << std::endl;
  }

  return 0;
}

