#
#   This file is part of SST/macroscale:
#                The macroscale architecture simulator from the SST suite.
#   Copyright (c) 2009 Sandia Corporation.
#   This software is distributed under the BSD License.
#   Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#   the U.S. Government retains certain rights in this software.
#   For more information, see the LICENSE file in the top
#   SST/macroscale directory.
#

import sys
import getopt
import sstmac
import math
import os

options = { "nnode": 0,
            "nnodejob": 0,
            "nprocpernode": 1,
            "memory_bandwidth": 1e10,
            "bandwidth": 147400000000,
            "latency": 0.0000013,
            "verbose": False,
            "help": False,
            "nrepeat": 1,
            "nconcurrent": 1,
            "indexing": "block",
            "allocation": "firstavailable",
            "network": "fattree",
            "interconnect": "circuit",
            "timing_model": "original",
            "nlevel": 4,
            "half_degree": 4,
            "degree": 4,
            "diameter": 3,
            "niter": 10,
            "nbyte": 100000,
            "timescale": 1.0,
            "inputfile": "./",
            "n1": 2,
            "n2": 2,
            "n3": 2,
            "nx": 2,
            "ny": 2,
            "nz": 2,
            "nnodeperswitch": -1,
            "width": 1,
            "app": "mpipingpong",
            "noiselevel": 0.0,
            "debug": "off"
            }

used_options = []

def get_option(opt):
    if not opt in used_options:
        used_options.append(opt)
    return options[opt]

def print_used_options():
    print "Examined options:"
    for i in used_options:
        if i != "help" and i != "verbose":
            optstr = str(options[i])
            if i == "inputfile":
                optstr = ".../" \
                    + os.path.basename(os.path.dirname(optstr)) \
                    + "/" + os.path.basename(optstr)
            print ("  %15s = %s") % ( i, optstr )

def usage():
    print "Usage: sys.argv[0] [options]"
    print "  Where [options] can be"
    for i in options.keys():
        if i == "verbose" or i == "help":
            print "  --" + str(i)
            
        else:
            print "  --" + str(i) + " value (current setting is " + str(options[i]) + ")"

def construct_network():
    netname = get_option("network")
    if netname == "mesh":
        net = sstmac.composite.construct(
            sstmac.line.construct(int(get_option("n1"))),
            sstmac.composite.construct(
                sstmac.line.construct(int(get_option("n2"))),
                sstmac.ring.construct(int(get_option("n3")))))
    elif netname == "fattree":
        nnps = int(get_option("nnodeperswitch"))
        if nnps < 0:
            nnps = 0
        width = int(get_option("width"))
        net = sstmac.fattree.construct(int(get_option("nlevel")),
                                       int(get_option("half_degree")),
                                       nnps, width)
    elif netname == "crossbar":
        nnode = int(get_option("nnode"))
        net = sstmac.crossbar.construct(nnode)
    elif netname == "gamma":
        nnps = int(get_option("nnodeperswitch"))
        if nnps < 0:
            nnps = 0
        width = int(get_option("width"))
        net = sstmac.gamma.construct(int(get_option("diameter")),
                                     int(get_option("degree")),
                                     nnps, width)
    elif netname == "ring":
        nnps = int(get_option("nnodeperswitch"))
        if nnps < 0:
            nnps = 1
        net = sstmac.ring.construct(int(get_option("n1")),
                                    nnps,
                                    int(get_option("width")))
    elif netname == "torus2D":
        nnps = int(get_option("nnodeperswitch"))
        if nnps < 0:
            nnps = 1
        net = sstmac.torus.construct_2D(int(get_option("n1")),
                                        int(get_option("n2")),
                                        nnps,
                                        int(get_option("width")))
    elif netname == "torus3D":
        nnps = int(get_option("nnodeperswitch"))
        if nnps < 0:
            nnps = 1
        net = sstmac.torus.construct_3D(int(get_option("n1")),
                                        int(get_option("n2")),
                                        int(get_option("n3")),
                                        nnps,
                                        int(get_option("width")))
    elif netname == "2HDtorus":
        nnps = int(get_option("nnodeperswitch"))
        if nnps < 0:
            nnps = 1
        net = sstmac.hdtorus.construct_2D(int(get_option("n1")),
                                        int(get_option("n2")),
                                        nnps,
                                        int(get_option("width")))
    elif netname == "3HDtorus":
        nnps = int(get_option("nnodeperswitch"))
        if nnps < 0:
            nnps = 1
        net = sstmac.hdtorus.construct_3D(int(get_option("n1")),
                                        int(get_option("n2")),
                                        int(get_option("n3")),
                                        nnps,
                                        int(get_option("width")))
    elif netname == "4HDtorus":
        nnps = int(get_option("nnodeperswitch"))
        if nnps < 0:
            nnps = 1
        net = sstmac.hdtorus.construct_4D(int(get_option("n1")),
                                        int(get_option("n2")),
                                        int(get_option("n3")),
                                        int(get_option("nx")), 
                                        nnps,
                                        int(get_option("width")))
    elif netname == "tbird":
        net = sstmac.tbird.construct()
    else:
        print 'construct_network: unknown network name: ' + netname
        raise NotImplementedError('unknown network: ' + netname)
    return net

def construct_interconnect():
    interconnectname = get_option("interconnect")
    if interconnectname == "circuit":
        net = construct_network()
        latency = sstmac.timestamp(float(get_option("latency")))
        bandwidth = float(get_option("bandwidth"))
        interconnect = sstmac.circuit.construct(latency, bandwidth, net)
        timing_model = get_option("timing_model")
        if timing_model == "original":
            interconnect.original_timing_model()
        elif timing_model == "split":
            interconnect.split_timing_model()
        else:
            print 'unknown timing_model: ' + timing_model
            raise NotImplementedError('unknown timing_model: ' + timing_model)
    elif interconnectname == "fullnetwork":
        latency = sstmac.timestamp(float(get_option("latency")))
        bandwidth = float(get_option("bandwidth"))
        interconnect = sstmac.fullnetwork.construct(latency, bandwidth)
    else:
        print 'construct_interconnect: unknown interconnect name: ' \
            + interconnect
        raise NotImplementedError('unknown interconnect: ' + interconnect)
    return interconnect

def construct_indexing():
    indexingname = get_option("indexing")
    if indexingname == "random":
        indexing = sstmac.randomindexing.construct()
    elif indexingname == "block":
        indexing = sstmac.blockindexing.construct()
    elif indexingname == "roundrobin":
        indexing = sstmac.roundrobinindexing.construct()
    else:
        print 'construct_indexing: unknown indexing name: ' \
            + indexingname
        raise NotImplementedError('unknown indexing: ' + indexingname)
    return indexing

def construct_allocation():
    allocationname = get_option("allocation")
    if allocationname == "random":
        allocation = sstmac.randomallocation.construct()
    elif allocationname == "firstavailable":
        allocation = sstmac.firstavailableallocation.construct()
    else:
        print 'construct_allocation: unknown allocation name: ' \
            + allocationname
        raise NotImplementedError('unknown allocation: ' + allocationname)
    return allocation

def construct_parser():
    timescale = float(get_option("timescale"))
    inputfile = get_option("inputfile")
    parser = sstmac.parsedumpi.construct(inputfile,timescale)
    return parser

def construct_app():
    appname = get_option("app")
    if appname == "mpipingpong":
        app = sstmac.mpipingpong.construct(get_option("niter"),
                                           get_option("nbyte"))
    elif appname == "mpi3d":
        noiselevel = float(get_option("noiselevel"))
        app = sstmac.mpi3d.construct(int(get_option("nx")),
                                     int(get_option("ny")),
                                     int(get_option("nz"))
                                     )
        app.noiselevel = noiselevel
    elif appname == "mpitrace":
        app = construct_parser()
    else:
        print 'construct_app: unknown app name: ' + appname
        raise NotImplementedError('unknown app: ' + app)
    return app

def main(args):
    """Build and run a simulator."""

    # convert the global options variable into a list of longopts for getopt
    longopts = ["help", "verbose"]
    for i in options.keys():
        if i != "help" and i != "verbose":
            longopts.append(i + "=")

    opts = getopt.getopt(args, "", longopts)

    print_help = False
    for i in opts[0]:
        opname = i[0][2:]
        if opname == "help":
            options["help"] = True
        elif opname == "verbose":
            options["verbose"] = True
        else:
            options[opname] = i[1]

    if options["help"] == True:
        usage()
        return

    if options["debug"] == "stderr":
        sstmac.debug().to_stderr()
#    elif options["debug"] == "stdout":
#        sstmac.debug().to_stdout()
    elif options["debug"] != "off":
        raise Exception("invalid debug option:" + options["debug"])

    app = construct_app()
    indexing = construct_indexing()
    allocation = construct_allocation()
    interconnect = construct_interconnect()

    mgr = sstmac.manager.construct()
    mgr.set_interconnect(interconnect)
    mgr.set_eventmanager(sstmac.eventmap.construct())

    nnode = int(get_option("nnode"))
    if nnode == 0:
        nnode = interconnect.slots()
        if nnode <= 0:
            raise Exception("must specify --nnode for the requested network")

    memory_bandwidth = float(get_option("memory_bandwidth"))
    mgr.add_nodes(sstmac.simplenode.construct(memory_bandwidth), nnode)

    job = sstmac.mpilaunch.construct(app)
    job.set_scheduler(sstmac.scheduler.construct(32768, 10240000, False))

    nprocpernode = int(get_option("nprocpernode"))

    nnodejob = int(get_option("nnodejob"))
    if nnodejob == 0:
        nnodejob = nnode

    starttime = sstmac.timestamp(0.)

    nrepeat = int(get_option("nrepeat"))

    nconcurrent = int(get_option("nconcurrent"))

    print_used_options()

    for i in range(nrepeat):
        for j in range(nconcurrent):
            mgr.add_launcher(starttime, job, nnodejob, nprocpernode, \
                                 indexing, allocation)
        starttime = mgr.run()
        runtime = starttime.sec();
        print ("simulated run time for %d nodes with %d processes per node = %12.9f seconds") % (nnode, nprocpernode, runtime)

if __name__ == "__main__":
    main(sys.argv[1:])
