/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/process/simplemboxapp.h>
#include <sstmac/process/simplemboxlaunch.h>
#include <sstmac/process/actormsg.h>
#include <sstmac/process/scheduler.h>
#include <sstmac/network/fattree.h>
#include <sstmac/simulator/circuit.h>
#include <sstmac/simulator/simplenode.h>
#include <sstmac/simulator/manager.h>
#include <sstmac/util/debug.h>
#include <boost/lexical_cast.hpp>
#include <string>
#include <cmath>

using namespace sstmac;

using boost::shared_ptr;
using boost::dynamic_pointer_cast;
using boost::lexical_cast;

/**
 * An example message type for mailbox delivery.
 */
class postcard : public sstmac::actormsg {
  /// Our value
  std::string msg_;

  /// Hi.
  postcard(const std::string &msg) :
    msg_(msg)
  {}

public:
  /// Build.
  static shared_ptr<postcard> construct(const std::string &msg = "") {
    return shared_ptr<postcard>(new postcard(msg));
  }

  /// Go away.
  virtual ~postcard() {
  }

  /// Create a copy.
  shared_ptr<actormsg> actorclone() const {
    return construct(msg_);
  }
  
  /// Access and update our data.
  std::string& value() { return msg_; }
};

/**
 * A test application for mailbox-type message delivery.
 */
class mboxtester : public sstmac::simplemboxapp {
  /// How many iterations are we planning to do?
  int iterations_;

  /// The real constructor.
  mboxtester(int iterations) : iterations_(iterations) {
  }

public:
  /// Get a heap-allocated object.
  static shared_ptr<mboxtester> construct(int iterations) {
    shared_ptr<mboxtester> rv(new mboxtester(iterations));
    rv->set_self(rv);
    return rv;
  }
  
  /// Goodbye.
  virtual ~mboxtester() {
  }

  /// Clone this object.
  virtual shared_ptr<simplemboxapp> clone() const {
    return construct(iterations_);
  }

  /// Make magic.
  void run() {
    timestamp start = mbox()->init();
    actorid me = mbox()->mypid();
    //std::cout << "Process " << me << " started at " << start << "\n";
    int sent = 0;
    shared_ptr<postcard> msg = postcard::construct();
    std::string hellostr("Hello number ");
    std::string mestr(" from node number ");
    mestr += lexical_cast<std::string>(me);
    const std::set<actorid> &world = mbox()->world();
    // Send a message to everybody
    for(std::set<actorid>::const_iterator it=world.begin(); it!=world.end(); ++it){
      ++sent;
      msg->value() = hellostr + lexical_cast<std::string>(sent) + mestr;
      mbox()->send(*it, msg, 64);
    }
    // Receive a message from everybody
    for(int id = 0; id < sent; ++id) {
      shared_ptr<postcard> recvd =
	dynamic_pointer_cast<postcard>(mbox()->recv(actorid::any(),
						    actorpattern::any()).second);
      if(recvd)
	std::cout << "Message received on " << me << ": " << recvd->value()
		  << "\n";
    }
    timestamp stop = mbox()->finalize();
    //std::cout << "Process " << me << " finalized at " << stop << "\n";
  }
};

int main(int argc, char **argv) {
  //sstmac::debug().to_stream(std::cerr);
  static const int fatness[] = {2, 8};
  const int maxnodes = int(pow(fatness[1], fatness[0]));
  int nodes = 2;
  if(argc > 1) nodes = atoi(argv[1]);
  if(nodes <= 0 || nodes > maxnodes) nodes = maxnodes;
  static const timestamp latency = timestamp::exact_usec(10);
  static const double bandwidth = 2.5e9;
  shared_ptr<fattree> fatso = fattree::construct(fatness[0], fatness[1], 0, 1);
  shared_ptr<manager> mgr = manager::construct();
  mgr->set_interconnect(circuit::construct(latency, bandwidth, fatso));
  std::cerr << "Slots: " << mgr->open_slots() << "\n";
  std::cerr << "Target nodes: " << nodes << "\n";
  mgr->add_nodes(simplenode::construct(timestamp::exact_usec(1), 1e10, 1), nodes);
  shared_ptr<simplemboxlaunch> launch =
    simplemboxlaunch::construct(mboxtester::construct(1));
  launch->set_scheduler(scheduler::construct(1<<16, 1<<20, false));
  mgr->add_launcher(timestamp(0), launch, nodes, 1);
  mgr->run();
}
